/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: common.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "fullengine.h"
#include "common.h"
#include "engine.h"
#include "discover.h"
#include "handlemgr.h"
#include "volume.h"
#include "memman.h"
#include "internalAPI.h"
#include "message.h"
#include "cluster.h"
#include "remote.h"


boolean is_top_object(storage_object_t * obj) {

	boolean result = FALSE;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Examining object %s.\n", obj->name);

	/* The object must not be part of a volume. */
	if (obj->volume == NULL) {

		/* The object must not be part of a container. */
		if (obj->consuming_container == NULL) {

			/* The object must have no parent objects. */
			if (ListEmpty(obj->parent_objects)) {
				result = TRUE;

			} else {
				LOG_DEBUG("Object %s has parent object(s).\n", obj->name);
			}

		} else {
			LOG_DEBUG("Object %s is part of container %s.\n", obj->name, obj->consuming_container->name);
		}

	} else {
		LOG_DEBUG("Object %s is part of volume %s.\n", obj->name, obj->volume->name);
	}

	LOG_PROC_EXIT_BOOLEAN(result);
	return result;
}


int evms_can_delete(object_handle_t thing) {
	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			switch (type) {
			case VOLUME:
				{
					logical_volume_t * volume = (logical_volume_t *) object;

					/* Check if the volume is opened. */
					if (!is_volume_opened(volume)) {

						if ((volume->file_system_manager != NULL) &&
						    !(volume->flags & VOLFLAG_UNMKFS)) {
							rc = volume->file_system_manager->functions.fsim->can_unmkfs(volume);

							if (rc != 0) {
								LOG_DETAILS("Volume %s cannot be deleted because the file system %s cannot be removed.  Reason code is %d.\n", volume->name, volume->file_system_manager->short_name, rc);
							}
						}

						if (rc == 0) {
							rc = volume->object->plugin->functions.plugin->can_set_volume(volume->object, FALSE);
						}

					} else {
						LOG_DETAILS("Volume %s cannot be deleted because it is opened.\n", volume->name);
						if (volume->mount_point) {
							LOG_DETAILS("Volume %s is mounted on %s.\n", volume->name, volume->mount_point);
						}
						rc = EBUSY;
					}
				}
				break;

			case CONTAINER:
				{
					storage_container_t * container = (storage_container_t *) object;

					rc = container->plugin->container_functions->can_delete_container(container);
				}
				break;

			case EVMS_OBJECT:
			case REGION:
			case SEGMENT:
			case DISK:
				{
					storage_object_t * storage_object = (storage_object_t *) object;

					/*
					 * Make sure the object is a top level object.  We
					 * can't delete objects in the middle of a feature
					 * stack.  A top level object has no parent.
					 */
					if (is_top_object(storage_object)) {

						/*
						 * Ask the feature if the object can be
						 * deleted.
						 */
						rc = storage_object->plugin->functions.plugin->can_delete(storage_object);

					} else {
						/* The object is not a top level object. */
						LOG_DETAILS("Object %s cannot be deleted because it is not a top level object.\n", storage_object->name);
						rc = EINVAL;
					}
				}
				break;

			default:
				/* We can't delete a thing of this type. */
				LOG_DETAILS("A thing of type %#x cannot be deleted.\n", type);
				rc = EINVAL;
				break;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int can_destroy_object(ADDRESS Object,
			      TAG     ObjectTag,
			      ADDRESS ObjectHandle,
			      ADDRESS Parameters) {
	int rc = 0;

	storage_object_t * obj = (storage_object_t *) Object;

	LOG_PROC_ENTRY();

	/* Disks cannot be destroyed, they must be deleted.  Return 0 if this is a
	 * DISK so that the can_destroy_object() will not fail if this is a DISK.
	 * The disk will not be destroyed on a evms_destroy(), but anything stacked
	 * on top of the disk will.
	 */

	if (obj->object_type != DISK) {
		rc = obj->plugin->functions.plugin->can_delete(obj);

		if (rc == 0) {
			/*
			 * If the object is not produced by a container
			 * do a recursive check to see if all
			 * of its children can be deleted.
			 */
			if (obj->producing_container == NULL) {
				rc = ForEachItem(obj->child_objects,
						 can_destroy_object,
						 NULL,
						 TRUE);
			}
		}
	}

	/* If the object is a DISK, we return 0 so that the recursive check doesn't
	 * fail and the recursion ends.
	 * can_destroy_object() might be called initially on a DISK (e.g., an EVMS
	 * volume is made from a DISK and the volume is being destroyed).  In that
	 * case, can_destroy_object() will return 0, but a subsequent call to
	 * evms_destroy() will destroy the volume but leave the disk.
	 */

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_can_destroy(object_handle_t thing) {
	int rc;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			switch (type) {
			case VOLUME:
				{
					logical_volume_t * volume = (logical_volume_t *) object;

					/* Check if the volume is opened. */
					if (!is_volume_opened(volume)) {

						if ((volume->file_system_manager != NULL) &&
						    !(volume->flags & VOLFLAG_UNMKFS)) {
							rc = volume->file_system_manager->functions.fsim->can_unmkfs(volume);

							if (rc != 0) {
								LOG_DETAILS("Volume %s cannot be destroyed because the file system %s cannot be removed.  Reason code is %d.\n", volume->name, volume->file_system_manager->short_name, rc);
							}
						}

						if (rc == 0) {
							rc = can_destroy_object(volume->object,
										get_tag_for_object(volume->object),
										NULL,
										NULL);
						}

					} else {
						LOG_DETAILS("Volume %s cannot be destroyed because it is opened.\n", volume->name);
						if (volume->mount_point != NULL) {
							LOG_DETAILS("Volume %s is mounted on %s.\n", volume->name, volume->mount_point);
						}
						rc = EBUSY;
					}
				}
				break;

			case CONTAINER:
				{
					storage_container_t * container = (storage_container_t *) object;

					rc = container->plugin->container_functions->can_delete_container(container);

					if (rc == 0) {
						rc = ForEachItem(container->objects_consumed,
								 can_destroy_object,
								 NULL,
								 TRUE);
					}
				}
				break;

			case EVMS_OBJECT:
			case REGION:
			case SEGMENT:
				{
					storage_object_t * obj = object;

					if (is_top_object(obj)) {

						rc = can_destroy_object(obj,
									get_tag_for_object(obj),
									NULL,
									NULL);

					} else {
						LOG_DETAILS("Object %s cannot be destroyed because it is not a top level object.\n", obj->name);
						rc = EINVAL;
					}
				}
				break;

			case DISK:
				{
					storage_object_t * obj = object;

					LOG_DETAILS("Disk %s cannot be destroyed because disks cannot be destroyed.  Disks must be deleted.\n", obj->name);
					rc = EINVAL;
				}
				break;

			default:
				rc = EINVAL;
				break;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int find_top_objects(ADDRESS object,
			    TAG     object_tag,
			    ADDRESS object_handle,
			    ADDRESS parameters) {

	int rc;
	storage_object_t *obj = (storage_object_t *) object;
	dlist_t top_object_list = (dlist_t) parameters;
	ADDRESS trash;

	if (ListEmpty(obj->parent_objects)) {
		rc = InsertObject(top_object_list,obj,
				  get_tag_for_object(obj),
				  NULL,
				  AppendToList,
				  FALSE,
				  &trash);
	} else {
		rc = ForEachItem(obj->parent_objects,
				 find_top_objects,
				 parameters,
				 TRUE);
	}

	return rc;
}


int evms_can_replace(object_handle_t source_handle, object_handle_t target_handle) {

	int rc = 0;
	int src_rc = 0;
	int trg_rc = 0;
	object_type_t type;
	void  * thing = NULL;
	dlist_t top_objects = NULL;
	storage_object_t * source = NULL;
	storage_object_t * target = NULL;
	storage_object_t * parent = NULL;

	LOG_PROC_ENTRY();

	if (replace_plugin == NULL) {
		LOG_PROC_EXIT_INT(ENOSYS);
		return ENOSYS;
	}

	/* Check for valid source and target. */
	src_rc = translate_handle(source_handle, &thing, &type);
	if (src_rc == HANDLE_MANAGER_NO_ERROR) {
		switch (type) {
		case DISK:
		case SEGMENT:
		case REGION:
		case EVMS_OBJECT:
			source = (storage_object_t *) thing;
			break;

		case CONTAINER:
			LOG_ERROR("Containers cannot be replaced.\n");
			src_rc = EINVAL;
			break;

		case VOLUME:
			LOG_ERROR("Volumes cannot be replaced.\n");
			src_rc = EINVAL;
			break;

		case PLUGIN:
			LOG_ERROR("Plug-ins cannot be replaced.\n");
			src_rc = EINVAL;
			break;

		default:
			LOG_ERROR("Source handle %d maps to an unknown thing of type %#x.\n", source_handle, type);
			src_rc = EINVAL;
		}
	}

	if (src_rc == 0) {
		int tmp_rc;
		logical_volume_t * vol;

		/*
		 * Top objects cannot be replaced.  What is the use of replacing
		 * an object that is not in use with another object that is not
		 * in use?  This also prevents the use of any target objects
		 * created by plug-ins for their own internal moves as the target
		 * will be a top level object.
		 */
		if (is_top_object(source)) {
			LOG_DETAILS("Object %s is a top level object.  Top level objects cannot be replaced.\n",
				  source->name);
			src_rc = EINVAL;
		}

		/*
		 * Is the object a volume's object?  Certain rules apply to
		 * volume objects.
		 */
		if (ListEmpty(source->parent_objects) && (source->volume != NULL)) {
			if (source->volume->flags & VOLFLAG_COMPATIBILITY) {
				LOG_DETAILS("%s is the object for compatiblity volume %s.  Compatibility volume objects cannot be replaced.\n",
					  source->name, source->volume->name);
				src_rc = EINVAL;

			} else {
				/* It's an EVMS volume.
				 * A "raw" EVMS volume (one made on top of a
				 * nonEVMS object) has its own device-mapper
				 * mapping.  The volume can have its object
				 * replaced since we can change the mapping of
				 * the volume.
				 * A non-raw EVMS volume simply uses the
				 * device-mapper device of the EVMS object.
				 * Non-raw EVSM volumes can only be replaced if
				 * the volume is not mounted, since the
				 * major:minor of the volume will change due to
				 * it using a differnt object.
				 */
				if (source->object_type == EVMS_OBJECT) {
					if (is_volume_mounted(source->volume)) {
						LOG_DETAILS("%s is the object for volume %s.  The object can only be replaced if the volume is not mounted.\n",
							  source->name, source->volume->name);
						src_rc = EBUSY;
					}
				}
			}
		}

		/* Can't move non-data objects. */
		if (source->data_type != DATA_TYPE) {
			LOG_ERROR("Object %s is not a data object.  Only data objects can be replaced.\n",
				  source->name);
			src_rc = EINVAL;
		}

		if (!engine_can_online_copy()) {
			if (!engine_is_offline(source, &vol)) {
				LOG_DETAILS("Object %s is part of volume %s which is currently opened.  Replace must be done off-line.\n",
					  source->name, vol->name);
				src_rc = EINVAL;
			}
		}

		/* Can't replace if there are no available targets. */
		tmp_rc = engine_get_object_list(0, DATA_TYPE, NULL, NULL, VALID_INPUT_OBJECT, &top_objects);
		if (tmp_rc == 0) {
			if (ListEmpty(top_objects)) {
				LOG_DETAILS("There are no available target objects for the replace.\n");
				src_rc = EINVAL;
			}

			DestroyList(&top_objects, FALSE);

		} else {
			src_rc = tmp_rc;
		}
	}

	if (target_handle != 0) {
		trg_rc = translate_handle(target_handle, &thing, &type);
		if (trg_rc == HANDLE_MANAGER_NO_ERROR) {
			switch (type) {
			case DISK:
			case SEGMENT:
			case REGION:
			case EVMS_OBJECT:
				target = (storage_object_t *) thing;
				if (!is_top_object(target)) {
					LOG_DEBUG("Target %s is not a top level object.\n", target->name);
					trg_rc = EINVAL;
				}
				break;

			case CONTAINER:
				LOG_ERROR("Containers cannot be the target of a replace.\n");
				trg_rc = EINVAL;
				break;

			case VOLUME:
				LOG_ERROR("Volumes cannot be the target of a replace.\n");
				trg_rc = EINVAL;
				break;

			case PLUGIN:
				LOG_ERROR("Plug-ins cannot be the target of a replace.\n");
				trg_rc = EINVAL;
				break;

			default:
				LOG_ERROR("Target handle %d maps to an unknown thing of type %#x.\n", target_handle, type);
				trg_rc = EINVAL;
			}

			if (trg_rc == 0) {
				/* Can't move to non-data objects. */
				if (target->data_type != DATA_TYPE) {
					LOG_ERROR("Object %s is not a data object.  Only data objects can be the target of a replace.\n",
						  target->name);
					trg_rc = EINVAL;
				}

			}
		}
	}

	rc = (src_rc != 0) ? src_rc : trg_rc;

	if ((rc == 0) && (target != NULL)) {
		/* Safty checks */
		if (source == target) {
			LOG_ERROR("Object %s cannot replace itself.\n", source->name);
			rc = EINVAL;

		} else {
			/*
			 * The source object cannot be one of the objects that
			 * comprises the target object.
			 */
			dlist_t top_objects = CreateList();

			if (top_objects != NULL) {
				int tmp_rc;
				storage_object_t * obj;

				ForEachItem(source->parent_objects,
					    find_top_objects,
					    top_objects,
					    TRUE);

				DLIST_FOR_EACH(obj, top_objects, tmp_rc) {
					if (obj == target) {
						LOG_DETAILS("Target object %s comprises source object %s.\n", target->name, source->name);
						rc = EINVAL;
					}
				}

				DestroyList(&top_objects, FALSE);
			}

			if (ListEmpty(source->parent_objects)) {
				if ((source->object_type == EVMS_OBJECT) &&
				    (target->object_type != EVMS_OBJECT)) {
					if (target->size < source->size + EVMS_FEATURE_HEADER_SECTORS * 2) {
						LOG_DETAILS("Target %s is smaller than source %s.\n", target->name, source->name);
						rc = EINVAL;
					}
				} else {
					if (target->size < source->size) {
						LOG_DETAILS("Target %s is smaller than source %s.\n", target->name, source->name);
						rc = EINVAL;
					}
				}

				/*
				 * We can't do an online move of a raw EVMS volume
				 * object to an EVMS object because that would
				 * cause the volume's major:minor to change to use
				 * that of the EVMS object instead of the mapping
				 * used to make the raw EVMS volume.  Swapping of
				 * a major:minor can't be done while the volume
				 * is mounted.
				 */
				if ((source->volume != NULL) &&
				    is_volume_mounted(source->volume) &&
				    (source->object_type != EVMS_OBJECT) &&
				    (target->object_type == EVMS_OBJECT)) {
					LOG_DETAILS("non-EVMS object %s cannot be replaced with EVMS object %s while volume %s is mounted.\n",
						    source->name, target->name, source->volume->name);
					rc = EBUSY;
				}
			}
		}
	}

	if (rc == 0) {
		if (!ListEmpty(source->parent_objects)) {
			/*
			 * Ask the parent if it can replace its child.  Any
			 * parent will do.  If there are multiple, parents they
			 * should all be managed by the same plug-in.
			 */
			parent = dlist_get_current_item(source->parent_objects);
			rc = parent->plugin->functions.plugin->can_replace_child(parent, source, target);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_can_set_info will determine if a thing's information can be set.
 */
int evms_can_set_info(object_handle_t thing_handle) {

	int rc = 0;
	object_type_t type;
	void  * thing = NULL;
	int count = -1;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		task_context_t * task = engine_alloc(sizeof(task_context_t));

		if (task != NULL) {
			rc = translate_handle(thing_handle, &thing, &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {

				/*
				 * Check if the plug-in supports setting of object information
				 * by getting the option count for EVMS_Task_Set_Info.  If the
				 * option count is 0 or less, then the plug-in does not support
				 * the setting of object information.
				 */
				switch (type) {
				case DISK:
				case SEGMENT:
				case REGION:
				case EVMS_OBJECT:
					{
						storage_object_t * obj = (storage_object_t *) thing;
						task->object = obj;
						task->action = EVMS_Task_Set_Info;

						count = obj->plugin->functions.plugin->get_option_count(task);

						if (count <= 0) {
							LOG_DETAILS("Plug-in %s does not have any descriptors for setting object info.\n", obj->plugin->short_name);
							rc = ENOSYS;
						}
					}
					break;

				case CONTAINER:
					{
						storage_container_t * con = (storage_container_t *) thing;
						task->container = con;
						task->action = EVMS_Task_Set_Container_Info;

						count = con->plugin->functions.plugin->get_option_count(task);

						if (count <= 0) {
							LOG_DETAILS("Plug-in %s does not have any descriptors for setting container info.\n", con->plugin->short_name);
							rc = ENOSYS;
						}
					}
					break;

				case VOLUME:
					{
						logical_volume_t * vol = (logical_volume_t *) thing;
						task->volume = vol;
						task->action = EVMS_Task_Set_Info;

						if (vol->file_system_manager != NULL) {
							rc = vol->file_system_manager->functions.fsim->get_option_count(task);

							if (count <= 0) {
								LOG_DETAILS("Plug-in %s does not have any descriptors for setting volume info.\n", vol->file_system_manager->short_name);
								rc = ENOSYS;
							}

						} else {
							LOG_DETAILS("Volume %s does not have a file system interface module.\n", vol->name);
							rc = ENOSYS;
						}
					}
					break;

				default:
					LOG_ERROR("Cannot set info on a thing of type %#x.\n", type);
					rc = EINVAL;
					break;
				}
			}

			engine_free(task);

		} else {
			LOG_CRITICAL("Error allocating memory for a task context.\n");
			rc = ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static inline int get_plugin_extended_info(plugin_record_t         * plugin,
					   char                    * descriptor_name,
					   extended_info_array_t * * info) {
	int rc = 0;

	LOG_PROC_ENTRY();

	switch (GetPluginType(plugin->id)) {
	case EVMS_DEVICE_MANAGER:
	case EVMS_SEGMENT_MANAGER:
	case EVMS_REGION_MANAGER:
	case EVMS_FEATURE:
	case EVMS_ASSOCIATIVE_FEATURE:

		rc = plugin->functions.plugin->get_plugin_info(descriptor_name, info);
		break;

	case EVMS_FILESYSTEM_INTERFACE_MODULE:
		rc = plugin->functions.fsim->get_plugin_info(descriptor_name, info);
		break;

	case EVMS_CLUSTER_MANAGER_INTERFACE_MODULE:
		rc = plugin->functions.cluster->get_plugin_info(descriptor_name, info);
		break;

	default:
		LOG_ERROR("We don't get info for %d plug-in types.\n", GetPluginType(plugin->id) );
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void free_extended_info_object_contents(void * object) {

	int i, j;
	extended_info_array_t * eia = (extended_info_array_t *) object;
	extended_info_t       * ei;

	LOG_PROC_ENTRY();

	for (i = 0; i < eia->count; i++) {
		ei = &eia->info[i];

		if (ei->name != NULL) {
			engine_free(ei->name);
		}
		if (ei->title != NULL) {
			engine_free(ei->title);
		}
		if (ei->desc != NULL) {
			engine_free(ei->desc);
		}

		switch (ei->collection_type) {
		case EVMS_Collection_List:
			if (ei->collection.list != NULL) {
				if (ei->type == EVMS_Type_String) {
					for (j = 0; j < ei->collection.list->count; j++) {
						if (ei->collection.list->value[j].s != NULL) {
							engine_free(ei->collection.list->value[j].s);
						}
					}
				}
				engine_free(ei->collection.list);

			} else {
				LOG_WARNING("Collection says it has a list but the list pointer is NULL.\n");
			}
			break;

		case EVMS_Collection_Range:
			if (ei->collection.range != NULL) {
				engine_free((void *) ei->collection.range);

			} else {
				LOG_WARNING("Collection says it has a range but the range pointer is NULL.\n");
			}

		default:
			break;
		}

		if ((ei->group.group_number != 0) && (ei->group.group_name != NULL)) {
			engine_free(ei->group.group_name);
		}
	}

	LOG_PROC_EXIT_VOID();
}


int evms_get_extended_info(object_handle_t thing, char * descriptor_name, extended_info_array_t * * user_info) {
	int rc = 0;
	void * object;
	object_type_t type;
	extended_info_array_t * info = NULL;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {
		LOG_DEBUG("Get extended info for handle %d.\n", thing);
		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			switch (type) {
			case PLUGIN:{
					plugin_record_t * plugin = (plugin_record_t *) object;

					LOG_DEBUG("Handle %d maps to plug-in %s.\n", thing, plugin->short_name);
					rc = get_plugin_extended_info(plugin, descriptor_name, &info);
				}
				break;

			case DISK:
			case SEGMENT:
			case REGION:
			case EVMS_OBJECT:{
					storage_object_t * obj = (storage_object_t *) object;

					LOG_DEBUG("Handle %d maps to storage object %s.\n", thing, obj->name);
					rc = obj->plugin->functions.plugin->get_info(obj, descriptor_name, &info);
				}
				break;


			case CONTAINER:{
					storage_container_t * container = (storage_container_t *) object;

					LOG_DEBUG("Handle %d maps to container %s.\n", thing, container->name);
					rc = container->plugin->container_functions->get_container_info(container, descriptor_name, &info);
				}
				break;

			case VOLUME:{
					logical_volume_t * volume = (logical_volume_t *) object;

					LOG_DEBUG("Handle %d maps to volume %s.\n", thing, volume->name);
					if (volume->file_system_manager != NULL) {
						rc = volume->file_system_manager->functions.fsim->get_volume_info(volume, descriptor_name, &info);

					} else {
						/*
						 * The volume doesn't have an FSIM.
						 * Return an empty info array.
						 */
						info = engine_alloc(sizeof(extended_info_array_t));

						if (info == NULL) {
							rc = ENOMEM;
						}
					}
				}
				break;

			default:
				LOG_DEBUG("Handle %d maps to unknown object type %d.\n", thing, type);
				rc = EINVAL;
				break;
			}

			if ((rc == 0) && (info != NULL)) {

				/*
				 * Copy the extended info that was returned from the plug-in
				 * into an application memory object suitable for returning
				 * to the API caller.
				 */
				if (info->count == 0) {

					*user_info = alloc_app_struct(sizeof(info->count), NULL);
					if (*user_info != NULL) {
						/* Fill in the structure returned to the app. */
						(*user_info)->count = 0;

						/*
						 * Free the structure that was returned by the plug-in.
						 */
						engine_free(info);

					} else {
						rc = ENOMEM;
					}

				} else {
					/* The extended info array has at least one entry. */
					uint size = sizeof(extended_info_array_t) + (sizeof(extended_info_t) * (info->count - 1));

					*user_info = alloc_app_struct(size, free_extended_info_object_contents);

					if (*user_info != NULL) {
						/* Fill in the structure returned to the app. */
						memcpy(*user_info, info, size);

						/*
						 * Free the structure that was returned by the plug-in.
						 */
						engine_free(info);

					} else {
						rc = ENOMEM;
					}
				}
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int update_all_stop_data_states() {

	int rc;
	storage_object_t * obj;

	LOG_PROC_ENTRY();

	/*
	 * Objects that are no longer top level objects cannot be guaranteed
	 * to have stop data on them.  Once they are consumed by another
	 * object or container, the consumer can write anything to the object
	 * anywhere on the object, including the stop data area.
	 */
	DLIST_FOR_EACH(obj, DiskList, rc) {
		if (!is_top_object(obj)) {
			obj->flags &= ~SOFLAG_HAS_STOP_DATA;
		}
	}
	DLIST_FOR_EACH(obj, SegmentList, rc) {
		if (!is_top_object(obj)) {
			obj->flags &= ~SOFLAG_HAS_STOP_DATA;
		}
	}
	DLIST_FOR_EACH(obj, RegionList, rc) {
		if (!is_top_object(obj)) {
			obj->flags &= ~SOFLAG_HAS_STOP_DATA;
		}
	}
	CLEAR_OK_DLIST_ERROR(rc);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_set_info will set information, specified in options, for a thing.
 */
int evms_set_info(engine_handle_t  thing_handle,
		  option_array_t * options) {

	int rc = 0;
	object_type_t type;
	void *        thing;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(thing_handle, &thing, &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {

			switch (type) {
			case DISK:
			case SEGMENT:
			case REGION:
			case EVMS_OBJECT:
				{
					storage_object_t * obj = (storage_object_t *) thing;
					char * old_object_name[sizeof(obj->name)] = {'\0'};

					rc = engine_can_rename(obj);

					if (rc == 0) {
						/* Check if it's part of a compatibility volume. */
						if ((obj->volume != NULL) &&
						    (obj->object_type != EVMS_OBJECT) &&
						    (obj->feature_header == NULL)) {

							/*
							 * Check if it's the top object of the
							 * compatibility volume.
							 */
							if (ListEmpty(obj->parent_objects)) {
								/*
								 * This is the top object of a compatibility
								 * volume.  If the object's name changes,
								 * the volume name must change as well.
								 * save the current volume name so that
								 * we can check if it changed during the
								 * call to the plug-in's set_info().
								 */
								memcpy(old_object_name, obj->name, sizeof(obj->name));
							}
						}

						rc = obj->plugin->functions.plugin->set_info(obj, options);

						/*
						 * Check if we need to handle a possible name
						 * change.
						 */
						if (old_object_name[0] != '\0') {
							if (memcmp(old_object_name, obj->name, sizeof(obj->name)) != 0) {

								char * pVolName = obj->name;
								ADDRESS trash;

								/*
								 * The object is a top level object of a
								 * compatibility volume and its name has
								 * changed.  Delete the old volume and
								 * create a new one with the new name.
								 */

								/* Unregister the current volume name. */
								engine_unregister_name(obj->volume->name);

								/* Delete the volume from the VolumeList. */
								DeleteObject(VolumeList, obj->volume);

								if (!(obj->volume->flags & VOLFLAG_NEW)) {
									/* Put it on the VolumeDeleteList. */
									rc = InsertObject(VolumeDeleteList,
											  obj->volume,
											  VOLUME_TAG,
											  NULL,
											  AppendToList,
											  FALSE,
											  &trash);

									if (rc != DLIST_SUCCESS) {
										LOG_WARNING("Error code %d when putting volume %s on the VolumeDeleteList.\n", rc, obj->volume->name);
										rc = 0;
									}
								}

								/*
								 * Build a new volume name for the object.
								 */
								if (strncmp(pVolName, EVMS_DEV_NODE_PATH, strlen(EVMS_DEV_NODE_PATH)) != 0) {
									pVolName = engine_alloc(strlen(pVolName) + strlen(EVMS_DEV_NODE_PATH) + 1);
									if (pVolName != NULL) {
										strcpy(pVolName, EVMS_DEV_NODE_PATH);
										strcat(pVolName, obj->name);
									} else {
										LOG_CRITICAL("Could not get memory for building a volume name for object %s.\n", obj->name);
										rc = ENOMEM;
									}
								}

								if (rc == 0) {

									/*
									 * Make a new volume patterned after the
									 * old one.
									 */
									logical_volume_t * old_vol = obj->volume;
									rc = make_volume(obj, pVolName);

									if (rc == 0) {
										obj->volume->dev_major     = old_vol->dev_major;
										obj->volume->dev_minor     = old_vol->dev_minor;
										obj->volume->serial_number = old_vol->serial_number;
										obj->volume->flags        |= old_vol->flags;
										obj->volume->flags        |= (VOLFLAG_NEW | VOLFLAG_DIRTY | VOLFLAG_NEEDS_ACTIVATE);
									}

									engine_free(pVolName);
								}
							}
						}
					}
				}
				break;

			case CONTAINER:
				{
					storage_container_t * con = (storage_container_t *) thing;
					rc = con->plugin->container_functions->set_container_info(con, options);
				}
				break;

			case VOLUME:
				{
					logical_volume_t * vol = (logical_volume_t *) thing;
					if (vol->file_system_manager != NULL) {
						rc = vol->file_system_manager->functions.fsim->set_volume_info(vol, options);

					} else {
						LOG_DEBUG("Volume %s does not have a file system interface module.\n", vol->name);
						rc = ENOSYS;
					}
				}
				break;

			default:
				LOG_ERROR("Cannot set info on a thing of type %#x.\n", type);
				rc = EINVAL;
				break;
			}
		}
	}

	if (rc == 0) {
		/*
		 * A plug-in can (ab)use set_info() to add or remove child
		 * objects to/from the given object.  Since we don't know
		 * what went on inside the set_info() and how it affected
		 * other objects, update the stop data state on all the
		 * objects that can have stop data.
		 */
		update_all_stop_data_states();
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Mark dirty the feature header hanging off of this object.
 * This function is a service function for set_feature_header_dirty() below.
 * This function has its parameters structures so that it can be called by
 * the ForEachItem() dlist processor.
 */
static int dirty_feature_header(ADDRESS object,
				TAG     object_tag,
				ADDRESS object_handle,
				ADDRESS parameters) {
	int rc = 0;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	if (obj->feature_header != NULL) {
		obj->flags |= SOFLAG_FEATURE_HEADER_DIRTY;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Remove this object's feature headers and add the feature header sectors to
 * the kill list.
 */
int remove_feature_headers(ADDRESS object,
			   TAG     object_tag,
			   ADDRESS object_handle,
			   ADDRESS parameters) {
	int rc = 0;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	if (obj->feature_header != NULL) {
		LOG_DEBUG("Removing feature headers from object %s.\n", obj->name);

		engine_free(obj->feature_header);
		obj->feature_header = NULL;

		/* Wipe out both feature headers. */
		obj->plugin->functions.plugin->add_sectors_to_kill_list(obj, obj->size - (EVMS_FEATURE_HEADER_SECTORS * 2), EVMS_FEATURE_HEADER_SECTORS * 2);

		/*
		 * If the object had stop data on it, it will be wiped out by
		 * the kill sectors.
		 */
		obj->flags &= ~SOFLAG_HAS_STOP_DATA;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Mark this object's feature header(s) dirty.  The feature header(s) for this
 * object is/are in this object's child objects.
 * This function has its parameters structures so that it can be called by
 * the ForEachItem() dlist processor.
 */
int set_feature_header_dirty(ADDRESS object,
			     TAG     object_tag,
			     ADDRESS object_handle,
			     ADDRESS parameters) {
	int rc = 0;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	/* Only EVMS objects have feature headers. */
	if (obj->object_type == EVMS_OBJECT) {
		ForEachItem(obj->child_objects,
			    dirty_feature_header,
			    NULL,
			    TRUE);

		ForEachItem(obj->child_objects,
			    set_feature_header_dirty,
			    NULL,
			    TRUE);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int do_volume_delete_warnings(logical_volume_t * vol) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/*
	 * If the volume is new then it really doesn't have an FSIM associated with
	 * it and it's not already a compatibility volume so there are no warnings
	 * to post.
	 */
	if (vol->flags & VOLFLAG_NEW) {
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	/*
	 * The volume has an FSIM, notify the user that unmkfs will be run on the
	 * volume.
	 */
	if (vol->file_system_manager != NULL) {

		/* Check if the FSIM can unmkfs. */
		rc = vol->file_system_manager->functions.fsim->can_unmkfs(vol);
		if (rc == 0) {
			/*
			 * Only need to ask the user to run unmkfs if the current FSIM is
			 * the original FSIM, it isn't already marked for unmkfs, and the
			 * volume is not read-only.
			 */
			if ((vol->file_system_manager == vol->original_fsim) &&
			    !(vol->flags & (VOLFLAG_UNMKFS | VOLFLAG_READ_ONLY))) {

				char * choices[] = {"Continue", "Cancel", NULL};
				int answer = 0;	    /* Default is "Continue". */

				engine_user_message(&answer, choices,
						    "Volume \"%s\" has the %s file system on it.  "
						    "The file system will be removed along with the volume.\n",
						    vol->name, vol->file_system_manager->short_name);
				switch (answer) {
				case 0:
					/*
					 * Turn on the VOLFLAG_UNMKFS flag so that unmkfs will
					 * be run at commit time.  If the volume's VOLFLAG_MKFS
					 * is set we don't care.  The volume will be deleted and
					 * so won't have the chance to run mkfs.
					 */
					vol->flags |= VOLFLAG_UNMKFS;

					/* Mark that this volume has no FSIM. */
					vol->file_system_manager = NULL;
					break;

				default:
					/* Any other answer is a cancel. */
					rc = E_CANCELED;
					break;
				}
			}

		} else {
			LOG_ERROR("FSIM %s will not allow the file system to be removed from volume %s.  Error code is %d.\n", vol->file_system_manager->short_name, vol->name, rc);
		}

	} else {
		/*
		 * The volume does not have an FSIM assigned to it.  If it did not
		 * originally have and FSIM assigned to it, it may have a file system on
		 * it for which we have no FSIM.  If the volume is not marked read-only
		 * ask the user if we should run a crude "unmkfs" that wipes out the
		 * first 1MB of the disk.
		 */
		if ((vol->original_fsim == NULL) &&
		    !(vol->flags & VOLFLAG_READ_ONLY)) {
			char * choices[] = {"Write zeros", "Do not write zeros", "Cancel", NULL};
			int answer = 0;	    /* Default is "Write zeroes". */

			sector_count_t one_meg_in_sectors = (1024 * 1024) >> EVMS_VSECTOR_SIZE_SHIFT;

			engine_user_message(&answer, choices,
					    "Volume \"%s\" does not have a File System Interface Module (FSIM) assigned to it.  "
					    "The volume may have a file system on it, but none of the installed FSIM plug-ins recognizes it.  "
					    "Do you want to write zeros to the %sdisable any file system that may be on the volume?\n",
					    vol->name,
					    (vol->object->size > one_meg_in_sectors) ? "first 1MB of the volume to potentially " : "volume to ");
			switch (answer) {
			case 0:
				rc = vol->object->plugin->functions.plugin->add_sectors_to_kill_list(vol->object, 0, min(vol->vol_size, (1024 * 1024) >> EVMS_VSECTOR_SIZE_SHIFT));
				break;

			case 1:
				/* Nothing to do. */
				break;

			case 2:
				rc = E_CANCELED;
			default:
				break;

			}
		}
	}

	if (rc == 0) {
		/*
		 * If this is a compatibility volume and we did not schedule an unmkfs,
		 * warn the user that the last few sectors will have metadata written
		 * to them.
		 */
		if ((vol->flags & VOLFLAG_COMPATIBILITY) &&
		    !(vol->flags & VOLFLAG_UNMKFS)) {
			char * choices[] = {"Continue", "Cancel", NULL};
			int answer = 0;	    /* Default is "Continue". */

			engine_user_message(&answer, choices,
					    "Volume \"%s\" is not an EVMS volume.  "
					    "Removing a non-EVMS volume requires writing %d bytes of metadata at the end of the volume. "
					    "The metadata will overwrite any data that may be at the end of the volume.  "
					    "Do you want to continue with the delete?\n",
					    vol->name, (EVMS_FEATURE_HEADER_SECTORS * 2) << EVMS_VSECTOR_SIZE_SHIFT);

			if (answer != 0) {
				/* An answer other than "Continue" is considered a cancel. */
				rc = E_CANCELED;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int delete_volume(logical_volume_t * volume) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/* Check if the volume is opened. */
	if (!is_volume_opened(volume)) {

		/*
		 * If the volume has a file system, ask the FSIM if the file system can
		 * be removed.
		 */
		if (volume->file_system_manager != NULL) {
			rc = volume->file_system_manager->functions.fsim->can_unmkfs(volume);
		}

		if (rc == 0) {
			storage_object_t * obj = volume->object;

			rc = obj->plugin->functions.plugin->can_set_volume(obj, FALSE);

			if (rc == 0) {
				/*
				 * Display any warning message(s) for the removal of this
				 * volume.  do_volume_delete_warning() can return E_CANCELED if the
				 * user selects to cancel the removal.
				 */
				rc = do_volume_delete_warnings(volume);

				if (rc == 0) {
					/*
					 * If the object has a feature header attached,
					 * it is one for making the object into an EVMS
					 * volume.  Remove the feature header and kill
					 * the sectors so that it won't be found again.
					 * Mark that the volume needs to be deactivated.
					 */
					if (obj->feature_header != NULL) {
						LOG_DEBUG("Removing feature headers from object %s.\n", obj->name);
						rc = obj->plugin->functions.plugin->add_sectors_to_kill_list(obj, obj->size - (EVMS_FEATURE_HEADER_SECTORS * 2), EVMS_FEATURE_HEADER_SECTORS * 2);

						if (rc == 0) {
							engine_free(obj->feature_header);
							obj->feature_header = NULL;
							volume->flags |= VOLFLAG_NEEDS_DEACTIVATE;

						} else {
							LOG_SERIOUS("add_sectors_to_kill_list() returned error code %d when called to wipe out volume feature header on volume %s.\n", rc, volume->name);
						}
					}

					if (rc == 0) {

						/*
						 * If the volume is not new then put it on the list of
						 * volumes to be deleted by the kernel runtime.
						 */
						if (!(volume->flags & VOLFLAG_NEW)) {
							ADDRESS handle;

							LOG_DEBUG("Volume exists.  Put it on the delete list.\n");
							rc = InsertObject(VolumeDeleteList,
									  volume,
									  VOLUME_TAG,
									  NULL,
									  AppendToList,
									  FALSE,
									  &handle);

							/*
							 * If the volume's object is associated with another
							 * object, put the other object's volume into this
							 * volume's associated_volume field.
							 */
							if (obj->associated_object != NULL) {
								volume->associated_volume = obj->associated_object->volume;
							}
						}

						if (rc == 0) {
							/* Delete the volume from the master VolumeList. */
							rc = DeleteObject(VolumeList, volume);

							if (rc == DLIST_SUCCESS) {

								/* Destroy the volume handle. */
								destroy_handle(volume->app_handle);

								/* Unregister the volume's name. */
								engine_unregister_name(volume->name);

								/*
								 * Clear out the volume pointer on this object
								 * and on all the child objects in the tree.
								 * Notify all the objects that the volume has
								 * been deleted.
								 */
								set_volume_in_object(obj,
										     get_tag_for_object(obj),
										     NULL,
										     NULL);    /* volume pointer */

								/*
								 * If this is a volume we created but didn't
								 * commit yet (indicated by the VOLFLAG_NEW
								 * flag) then just throw away the volume
								 * structure.
								 */
								if (volume->flags & VOLFLAG_NEW) {
									LOG_DEBUG("Volume is new, so just toss it out.\n");
									engine_free(volume);
								}

							} else {
								LOG_WARNING("Error code %d when deleting volume %s from the VolumeList.\n", rc, volume->name);
							}

						} else {
							LOG_WARNING("Error code %d when inserting volume %s into the VolumeDeleteList.\n", rc, volume->name);
						}
					}
				}

			} else {
				LOG_ERROR("Object %s does not allow volume %s to be deleted.  Reason code is %d.\n", obj->name, volume->name, rc);
			}

		} else {
			LOG_ERROR("Volume %s cannot be deleted because the file system %s cannot be deleted.  Reason code is %d.\n", volume->name, volume->file_system_manager->short_name, rc);
		}

	} else {
		LOG_ERROR("Volume %s cannot be deleted because it is opened.\n", volume->name);
		if (volume->mount_point != NULL) {
			LOG_ERROR("Volume %s is mounted on %s.\n", volume->name, volume->mount_point);
		}
		rc = EBUSY;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int get_working_top_object(logical_volume_t   * vol,
			   storage_object_t * * obj) {

	int rc = 0;
	storage_object_t * working_top_obj = vol->object;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Find working top object for volume %s.\n", vol->name);

	*obj = NULL;

	while ((rc == DLIST_SUCCESS) && (working_top_obj != NULL) &&
	       (GetPluginType(working_top_obj->plugin->id) == EVMS_ASSOCIATIVE_FEATURE) &&
	       (working_top_obj->associated_object == NULL)) {
		TAG tag;

		rc = BlindGetObject(working_top_obj->child_objects,
				    &tag,
				    NULL,
				    FALSE,
				    (ADDRESS *) &working_top_obj);
	}

	if (rc == DLIST_SUCCESS) {
		*obj = working_top_obj;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * sync_volumes() will make sure that compatibility volumes have the
 * correct name, i.e., the EVMS_DEV_NODE_PATH prefix followed by the object
 * name.  (In the case where a segment is deleted or added to the disk, the
 * compatibility names of the objects may shift and hence the name of the
 * volume may change.)  It also makes sure that non-native EVMS volumes have
 * the same major:minor as the volume's object.
 */
int sync_volumes(void) {
	int rc = 0;
	logical_volume_t * vol;

	LOG_PROC_ENTRY();

	DLIST_FOR_EACH(vol, VolumeList, rc) {
		storage_object_t * working_top_obj;

		rc = get_working_top_object(vol, &working_top_obj);

		if (rc == DLIST_SUCCESS) {
			if (vol->flags & VOLFLAG_COMPATIBILITY) {
				/*
				 * Compatibility volumes must have their name
				 * derived from the volume's working object.
				 */
				if ((strcmp(vol->name + EVMS_DEV_NODE_PATH_LEN, working_top_obj->name) != 0)) {
					engine_user_message(NULL, NULL,
							    "Volume name %s has shifted to %s%s.\n",
							    vol->name, EVMS_DEV_NODE_PATH, working_top_obj->name);

					engine_unregister_name(vol->name);

					/* Give the volume its new name. */
					memset(vol->name, 0, sizeof(vol->name));
					strcpy(vol->name, EVMS_DEV_NODE_PATH);
					strcat(vol->name, working_top_obj->name);
					engine_register_name(vol->name);

					if (!(vol->flags & VOLFLAG_NEW)) {
						vol->flags |= VOLFLAG_NEEDS_ACTIVATE;
					}
				}
			}

			/*
			 * For volumes that are not featureless EVMS volumes, make sure
			 * the volume has the same major:minor as its object.  (Its
			 * immediate object, not the first object that is not an
			 * associative feature.)
			 */
			if (working_top_obj->feature_header == NULL) {
				if ((vol->dev_major != vol->object->dev_major) ||
				    (vol->dev_minor != vol->object->dev_minor)) {
					vol->dev_major = vol->object->dev_major;
					vol->dev_minor = vol->object->dev_minor;

					if (local_focus) {
						if (ensure_dev_node(vol->name, vol->dev_major, vol->dev_minor) == 0) {
							vol->flags |= VOLFLAG_ACTIVE;
							vol->flags &= ~VOLFLAG_NEEDS_ACTIVATE;
						}

					} else {
						if (remote_ensure_dev_node(&current_nodeid, vol->name, vol->dev_major, vol->dev_minor) == 0) {
							vol->flags |= VOLFLAG_ACTIVE;
							vol->flags &= ~VOLFLAG_NEEDS_ACTIVATE;
						}
					}
				}
			}

		} else {
			LOG_WARNING("Error code %d (%s) when trying to find the working top object of volume %s.\n", rc, evms_strerror(rc), vol->name);

			/* Let the syncing continue. */
			rc = DLIST_SUCCESS;
		}
	}

	CLEAR_OK_DLIST_ERROR(rc);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_delete(object_handle_t thing) {
	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			switch (type) {
			case VOLUME:
				{
					logical_volume_t * volume = (logical_volume_t *) object;

					LOG_DEBUG("Request to delete volume %s.\n", volume->name);

					rc = delete_volume(volume);
				}
				break;

			case CONTAINER:
				{
					storage_container_t * container = (storage_container_t *) object;
					dlist_t objects_consumed = CreateList();

					LOG_DEBUG("Request to destroy container %s.\n", container->name);

					if (objects_consumed != NULL) {

						rc = container->plugin->container_functions->delete_container(container, objects_consumed);
						LOG_DEBUG("Return code from delete_container() is %d: %s.\n",  rc, evms_strerror(rc));

						DestroyList(&objects_consumed, FALSE);

					} else {
						LOG_CRITICAL("Error allocating memory for a dlist of consumed objects.\n");
					}
				}
				break;

			case EVMS_OBJECT:
			case REGION:
			case SEGMENT:
			case DISK:
				{
					storage_object_t * obj = (storage_object_t *) object;

					LOG_DEBUG("Request to delete object %s.\n", obj->name);

					/*
					 * Make sure the object is a top level object.  We can't
					 * delete objects in the middle of a feature stack or
					 * under a volume has no parent.
					 */
					if (is_top_object(obj)) {
						dlist_t child_list = CreateList();

						if (child_list != NULL) {
							/*
							 * If the object has stop data on it,
							 * delete it.
							 */
							if (obj->flags & SOFLAG_HAS_STOP_DATA) {
								LOG_DEBUG("Removing stop data from object %s.\n", obj->name);
								rc = obj->plugin->functions.plugin->add_sectors_to_kill_list(obj, obj->size - (EVMS_FEATURE_HEADER_SECTORS * 2), EVMS_FEATURE_HEADER_SECTORS * 2);

								if (rc == 0) {
									obj->flags &= ~SOFLAG_HAS_STOP_DATA;

								} else {
									LOG_SERIOUS("add_sectors_to_kill_list() returned error code %d when called to wipe out stop data on object %s.\n", rc, obj->name);
								}
							}

							rc = obj->plugin->functions.plugin->delete(obj, child_list);

							if (rc == 0) {

								int tmp_rc;
								storage_object_t * child_obj;

								/*
								 * Now that the child objects are top
								 * objects, they will not contain feature
								 * headers.  Remove the feature headers from
								 * the child objects.
								 */
								ForEachItem(child_list,
									    remove_feature_headers,
									    NULL,
									    TRUE);

								/*
								 * Mark the children's child objects'
								 * feature headers dirty so that the volume
								 * complete flag will be set correctly on
								 * commit.
								 */
								ForEachItem(child_list,
									    set_feature_header_dirty,
									    NULL,
									    TRUE);

								/*
								 * The parent object may have written
								 * over the stop data when it owned the
								 * child.  Forget that the child had
								 * stop data.
								 */
								DLIST_FOR_EACH(child_obj, child_list, tmp_rc) {
									child_obj->flags &= ~SOFLAG_HAS_STOP_DATA;
								}

								/*
								 * Make sure the compatibility volume names
								 * are in sync.
								 */
								sync_volumes();
							}

							DestroyList(&child_list, FALSE);

						} else {
							LOG_CRITICAL("Error getting memory for a child objects list.\n");
							rc = ENOMEM;
						}

					} else {
						LOG_ERROR("Object %s is not a top level object.\n", obj->name);
						rc = EINVAL;
					}
				}
				break;

			default:
				LOG_ERROR("A thing with a type of %#x cannot be deleted.\n", type);
				rc = EINVAL;
				break;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Delete an object and all of its children.
 * This function is a service function for evms_destroy() below.
 * This function has its parameters structured so that it can be called by
 * the DLIST ForEachItem() processor.  In fact, this function uses
 * ForEachItem() to call itself recursively to delete its children.
 */
static int destroy_object(ADDRESS Object,
			  TAG     ObjectTag,
			  ADDRESS ObjectHandle,
			  ADDRESS Parameters) {
	int rc = 0;

	storage_object_t * obj = (storage_object_t *) Object;

	dlist_t child_list;

	LOG_PROC_ENTRY();

	/* Disks are not deleted on a destroy. */

	if (obj->object_type == DISK) {
		/*
		 * When the disk was consumed the stop data may have been
		 * trashed.  Forget that it has stop data.  It will be
		 * rewritten on commit.
		 */
		obj->flags &= ~SOFLAG_HAS_STOP_DATA;

	} else {
		child_list = CreateList();

		if (child_list != NULL) {
			/*
			 * Save the producing container so we can check it after the object
			 * is deleted.
			 */
			storage_container_t * producing_container = obj->producing_container;

			/* If the object has stop data on it, delete it. */
			if (obj->flags & SOFLAG_HAS_STOP_DATA) {
				LOG_DEBUG("Removing stop data from object %s.\n", obj->name);
				rc = obj->plugin->functions.plugin->add_sectors_to_kill_list(obj, obj->size - (EVMS_FEATURE_HEADER_SECTORS * 2), EVMS_FEATURE_HEADER_SECTORS * 2);

				if (rc == 0) {
					obj->flags &= ~SOFLAG_HAS_STOP_DATA;

				} else {
					LOG_SERIOUS("add_sectors_to_kill_list() returned error code %d when called to wipe out stop data on object %s.\n", rc, obj->name);
				}
			}

			rc = obj->plugin->functions.plugin->delete(obj, child_list);

			if (rc == 0) {
				/*
				 * Now that the child objects are top objects, they will not contain
				 * feature headers.  Remove the feature headers from the child
				 * objects.
				 */
				ForEachItem(child_list,
					    remove_feature_headers,
					    NULL,
					    TRUE);

				/*
				 * If the object was not produced by a container, do a recursive
				 * deletion of its children.
				 */
				if (producing_container == NULL) {
					rc = ForEachItem(child_list,
							 destroy_object,
							 NULL,
							 TRUE);
				}
			}

		} else {
			LOG_CRITICAL("Error allocating memory for a child object list.\n");
			rc = ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int destroy_volume(logical_volume_t * volume) {

	int rc = 0;
	storage_object_t * obj = volume->object;

	LOG_PROC_ENTRY();

	/*
	 * Ask the object below the volume if it can handle
	 * being deleted.
	 */
	rc = can_destroy_object(obj,
				get_tag_for_object(obj),
				NULL,
				NULL);

	if (rc == 0) {

		rc = delete_volume(volume);

		if (rc == 0) {
			rc = destroy_object(obj,
					    get_tag_for_object(obj),
					    NULL,
					    NULL);
		}

	} else {
		LOG_ERROR("Object %s does not allow volume %s to be deleted.  Reason code is %d.\n", obj->name, volume->name, rc);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_destroy() deletes an object and all the child objects in the tree
 * beneath the object.
 */
int evms_destroy(object_handle_t thing) {

	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			switch (type) {
			case VOLUME:
				{
					logical_volume_t * volume = (logical_volume_t *) object;

					LOG_DEBUG("Request to destroy volume %s.\n", volume->name);

					rc = destroy_volume(volume);

					if (rc == 0) {
						/*
						 * Make sure the compatibility volume names
						 * are in sync.
						 */
						sync_volumes();
					}
				}
				break;

			case CONTAINER:
				{
					storage_container_t * container = (storage_container_t *) object;
					dlist_t objects_consumed = CreateList();

					LOG_DEBUG("Request to destroy container %s.\n", container->name);

					if (objects_consumed != NULL) {

						rc = container->plugin->container_functions->delete_container(container, objects_consumed);
						LOG_DEBUG("Return code from call to %s plug-in's delete_container() is %d.\n", container->plugin->short_name, rc);

						if (rc == 0) {
							/* Destroy the objects that were in the container. */
							rc = ForEachItem(objects_consumed,
									 destroy_object,
									 NULL,
									 TRUE);
						}

						DestroyList(&objects_consumed, FALSE);

					} else {
						LOG_CRITICAL("Error allocating memory for a dlist of consumed objects.\n");
					}
				}
				break;

			case EVMS_OBJECT:
			case REGION:
			case SEGMENT:
				{
					storage_object_t * obj = (storage_object_t *) object;

					LOG_DEBUG("Request to destroy storage object %s.\n", obj->name);

					/*
					 * Make sure the object is a top level object.  We can't
					 * destroy objects in the middle of a feature stack.  A
					 * top level object has no parent.
					 */
					if (is_top_object(obj)) {

						rc = destroy_object(obj,
								    get_tag_for_object(obj),
								    NULL,
								    NULL);
						if (rc == 0) {
							/*
							 * Make sure the compatibility volume names
							 * are in sync.
							 */
							sync_volumes();
						}

					} else {
						/* The object is not a top level object. */

						LOG_ERROR("Object %s is not a top level object.\n", obj->name);
						rc = EINVAL;
					}
				}
				break;

			case DISK:
				{
					storage_object_t * obj = object;

					LOG_ERROR("Disk %s cannot be destroyed because disks cannot be destroyed.  Disks must be deleted.\n", obj->name);
					rc = EINVAL;
				}
				break;

			default:
				LOG_ERROR("A thing with a type of %#x cannot be destroyed.\n", type);
				rc = EINVAL;
				break;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int create_replace_object(storage_object_t   * source,
				 storage_object_t   * target,
				 storage_object_t * * replace_object) {
	int rc = 0;
	dlist_t input_list;
	dlist_t output_list;
	ADDRESS trash;

	LOG_PROC_ENTRY();

	input_list = CreateList();
	output_list = CreateList();

	if ((input_list != NULL) && (output_list != NULL)) {
		/*
		 * Order is important here.  The Replace feature takes
		 * the first object as the source and the second object
		 * as the target.
		 */
		rc = InsertObject(input_list,
				  source,
				  get_tag_for_object(source),
				  NULL,
				  AppendToList,
				  FALSE,
				  &trash);

		if (rc == DLIST_SUCCESS) {
			rc = InsertObject(input_list,
					  target,
					  get_tag_for_object(target),
					  NULL,
					  AppendToList,
					  FALSE,
					  &trash);

			if (rc == DLIST_SUCCESS) {
				rc = replace_plugin->functions.plugin->create(input_list, NULL, output_list);

                                if (rc == 0) {
					TAG tag;

					/*
					 * Any stop data on the target may get
					 * trashed by a parent object.  Forget
					 * that it is there.
					 */
					target->flags &= ~SOFLAG_HAS_STOP_DATA;

					/*
					 * Return the new replace object to the
					 * caller.
					 */
					rc = BlindGetObject(output_list, &tag, NULL, FALSE, (ADDRESS *) replace_object);
				}

			} else {
				LOG_WARNING("Error %d when putting target object %s into the input list.\n", rc, target->name);
			}

		} else {
			LOG_WARNING("Error %d when putting source object %s into the input list.\n", rc, source->name);
		}

		DestroyList(&input_list, FALSE);
		DestroyList(&output_list, FALSE);

	} else {
		if (input_list == NULL) {
			LOG_CRITICAL("Error allocating memory for an input dlist.\n");
		} else {
			DestroyList(&input_list, FALSE);
		}

		if (output_list == NULL) {
			LOG_CRITICAL("Error allocating memory for an output dlist.\n");
		} else {
			DestroyList(&output_list, FALSE);
		}

		rc = ENOMEM;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_replace(object_handle_t source_handle, object_handle_t target_handle) {

	int rc = 0;
	storage_object_t * source;
	storage_object_t * target;
	storage_object_t * replace_obj;
	object_type_t type;
	dlist_t source_parents = NULL;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();
	if (rc != 0) {
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	rc = evms_can_replace(source_handle, target_handle);
	if (rc != 0) {
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	/*
	 * No need to check the return codes from translate_handle()
	 * since evms_can_move() translated the handles and got good
	 * return codes.
	 */
	translate_handle(source_handle,
			 (void *) &source,
			 &type);

	translate_handle(target_handle,
			 (void *) &target,
			 &type);

	if (!ListEmpty(source->parent_objects)) {

		/* Get a copy of the source's parent list so
		 * we can insert the replace object after it
		 * has been created.  When the replace object
		 * is created it changes the source's parent
		 * list to point to the new replace object.
		 */
		source_parents = CreateList();
		if (source_parents != NULL) {

			rc = CopyList(source_parents, source->parent_objects, AppendToList);
			if (rc != DLIST_SUCCESS) {
				LOG_SERIOUS("Error making a copy of the parent list.  CopyList() returned error %d: %s\n", rc, evms_strerror(rc));
				DestroyList(&source_parents, FALSE);
				LOG_PROC_EXIT_INT(rc);
				return rc;
			}

		} else {
			LOG_CRITICAL("Error creating a dlist for source parents.\n");
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}
	}

	rc = create_replace_object(source, target, &replace_obj);

	if (rc == 0) {
		if (source_parents != NULL) {
			storage_object_t * parent;

			DLIST_FOR_EACH(parent, source_parents, rc) {
				rc = parent->plugin->functions.plugin->replace_child(parent, source, replace_obj);

				if (rc != 0) {
					int tmp_rc;
					LOG_WARNING("Parent object %s failed to replace child object %s with new child object %s.  Error code was %d: %s\n",
						    parent->name, source->name, replace_obj->name, rc, evms_strerror(rc));

					/* Undo any replaces that succeeded. */
					DLIST_FOR_EACH(parent, source_parents, tmp_rc) {
						parent->plugin->functions.plugin->replace_child(parent, replace_obj, source);
					}

					/*
					 * Bail out of the loop now.  Don't get
					 * the next item.  Doing so sets rc.
					 */
					break;
				}
			}

			CLEAR_OK_DLIST_ERROR(rc);

		} else {
			source->volume->object = replace_obj;
		}
	}

	if (source_parents != NULL) {
		DestroyList(&source_parents, FALSE);
	}

	if (rc == 0) {
		/*
		 * Some plug-ins NULL out the source's volume pointer
		 * when they abandon the source as a child.  However, in this
		 * case, the source has not really been abandoned; it is still
		 * part of the volume.
		 */
		source->volume = replace_obj->volume;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int make_dlist(handle_array_t * ha, dlist_t list) {

	int rc = 0;
	int i;
	void * object;
	object_type_t type;
	TAG tag = 0;

	LOG_PROC_ENTRY();

	/* If handle array pointer is NULL, just return the empty dlist_t. */
	if (ha != NULL) {
		for (i = 0; (rc == 0) && (i < ha->count); i++) {
			rc = translate_handle(ha->handle[i],
					      &object,
					      &type);
			if (rc == HANDLE_MANAGER_NO_ERROR) {

				switch (type) {
				case DISK:
					tag = DISK_TAG;
					break;
				case SEGMENT:
					tag = SEGMENT_TAG;
					break;
				case REGION:
					tag = REGION_TAG;
					break;
				case EVMS_OBJECT:
					tag = EVMS_OBJECT_TAG;
					break;
				case CONTAINER:
					tag = CONTAINER_TAG;
					break;
				case VOLUME:
					tag = VOLUME_TAG;
					break;
				default:
					rc = EINVAL;
				}

				if (rc == 0) {
					ADDRESS handle;

					rc = InsertObject(list,
							  object,
							  tag,
							  NULL,
							  AppendToList,
							  FALSE,
							  &handle);
				}
			}
		}
	}

	if (rc != 0) {
		DeleteAllItems(list, FALSE);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void free_info_object_contents(void * object) {

	handle_object_info_t * info = (handle_object_info_t *) object;

	LOG_PROC_ENTRY();

	switch (info->type) {
	case PLUGIN:
		break;

	case DISK:
	case SEGMENT:
	case REGION:
	case EVMS_OBJECT:
		if (info->info.object.parent_objects != NULL) {
			engine_free(info->info.object.parent_objects);
		}

		if (info->info.object.child_objects != NULL) {
			engine_free(info->info.object.child_objects);
		}

		break;

	case CONTAINER:
		if (info->info.container.objects_consumed != NULL) {
			engine_free(info->info.container.objects_consumed);
		}
		if (info->info.container.objects_produced != NULL) {
			engine_free(info->info.container.objects_produced);
		}

		break;

	case VOLUME:
		if (info->info.volume.mount_point != NULL) {
			engine_free(info->info.volume.mount_point);
		}

		break;

	default:
		break;
	}

	LOG_PROC_EXIT_VOID();
}


int evms_get_handle_object_type(object_handle_t handle,
				object_type_t * type) {
	int rc = 0;
	void * object;

	LOG_PROC_ENTRY();

	*type = 0;

	rc = check_engine_read_access();

	if (rc == 0) {
		rc = translate_handle(handle,
				      &object,
				      type);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Make an array of handles (handle_array_t) for the objects in a dlist_t.
 */
int make_handle_array(dlist_t list, handle_array_t * * ha) {
	int rc = 0;
	uint count;
	uint size;

	LOG_PROC_ENTRY();

	rc = GetListSize(list, &count);

	if (rc == 0) {
		LOG_DEBUG("Number of objects in the list:  %d\n", count);
		if (count > 1) {
			size = sizeof(handle_array_t) + ((count -1) * sizeof(object_handle_t));
		} else {
			size = sizeof(handle_array_t);
		}

		*ha = engine_alloc(size);
		if (*ha != NULL) {
			rc = ForEachItem(list,
					 make_handle_entry,
					 *ha,
					 TRUE);
		} else {
			rc = ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_get_info(object_handle_t thing, handle_object_info_t * * user_info) {
	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	*user_info = NULL;

	rc = check_engine_read_access();

	if (rc == 0) {
		handle_object_info_t * info = NULL;

		info = alloc_app_struct(sizeof(handle_object_info_t), free_info_object_contents);

		if (info != NULL) {
			LOG_DEBUG("Get info for handle %d.\n", thing);
			rc = translate_handle(thing,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				switch (type) {
				case PLUGIN:
					{
						plugin_record_t * plugin = (plugin_record_t *) object;

						LOG_DEBUG("Handle %d maps to plug-in %s.\n", thing, plugin->short_name);
						info->type = PLUGIN;

						rc = ensure_app_handle(plugin,
								       PLUGIN,
								       &plugin->app_handle);
						if (rc == 0) {
							info->info.plugin.handle                             = plugin->app_handle;
							info->info.plugin.id                                 = plugin->id;
							info->info.plugin.version.major                      = plugin->version.major;
							info->info.plugin.version.minor                      = plugin->version.minor;
							info->info.plugin.version.patchlevel                 = plugin->version.patchlevel;
							info->info.plugin.required_engine_api_version        = plugin->required_engine_api_version;
							info->info.plugin.required_plugin_api_version.plugin = plugin->required_plugin_api_version.plugin;
							info->info.plugin.required_container_api_version     = plugin->required_container_api_version;
							info->info.plugin.short_name                         = plugin->short_name;
							info->info.plugin.long_name                          = plugin->long_name;
							info->info.plugin.oem_name                           = plugin->oem_name;
							info->info.plugin.supports_containers                = (plugin->container_functions != NULL);
						}
					}
					break;

				case DISK:
				case SEGMENT:
				case REGION:
				case EVMS_OBJECT:
					{
						storage_object_t * obj = (storage_object_t *) object;

						LOG_DEBUG("Handle %d maps to storage object %s.\n", thing, obj->name);
						info->type = obj->object_type;

						rc = make_handle_array(obj->parent_objects, &info->info.object.parent_objects);
						if (rc == 0) {

							rc = make_handle_array(obj->child_objects, &info->info.object.child_objects);
							if (rc == 0) {

								rc = ensure_app_handle(obj,
										       obj->object_type,
										       &obj->app_handle);
								if (rc == 0) {
									info->info.object.handle = obj->app_handle;
									info->info.object.object_type = obj->object_type;
									info->info.object.data_type = obj->data_type;
								}

								if (rc == 0) {
									if (obj->plugin != NULL) {
										rc = ensure_app_handle(obj->plugin,
												       PLUGIN,
												       &obj->plugin->app_handle);
										if (rc == 0) {
											info->info.object.plugin = obj->plugin->app_handle;
										}

									} else {
										info->info.object.plugin = 0;
									}
								}

								if (rc == 0) {
									if (obj->producing_container != NULL) {

										rc = ensure_app_handle(obj->producing_container,
												       CONTAINER,
												       &obj->producing_container->app_handle);
										if (rc == 0) {
											info->info.object.producing_container = obj->producing_container->app_handle;
										}
									} else {
										info->info.object.producing_container = 0;
									}
								}

								if (rc == 0) {
									if (obj->consuming_container != NULL) {

										rc = ensure_app_handle(obj->consuming_container,
												       CONTAINER,
												       &obj->consuming_container->app_handle);
										if (rc == 0) {
											info->info.object.consuming_container = obj->consuming_container->app_handle;
										}
									} else {
										info->info.object.consuming_container = 0;
									}
								}

								if (rc == 0) {
									if (obj->volume != NULL) {

										rc = ensure_app_handle(obj->volume,
												       VOLUME,
												       &obj->volume->app_handle);
										if (rc == 0) {
											info->info.object.volume = obj->volume->app_handle;
										}
									} else {
										info->info.object.volume = 0;
									}
								}

								if (rc == 0) {
									if (obj->disk_group != NULL) {

										rc = ensure_app_handle(obj->disk_group,
												       CONTAINER,
												       &obj->disk_group->app_handle);
										if (rc == 0) {
											info->info.object.disk_group = obj->disk_group->app_handle;
										}
									} else {
										info->info.object.disk_group = 0;
									}
								}

								if (rc == 0) {
									info->info.object.dev_major = obj->dev_major;
									info->info.object.dev_minor = obj->dev_minor;
									info->info.object.flags = obj->flags;
									memcpy(info->info.object.name, obj->name, sizeof(obj->name));
									info->info.object.start = obj->start;
									info->info.object.size = obj->size;
									info->info.object.geometry = obj->geometry;
								}
							}
						}
					}
					break;

				case CONTAINER:
					{
						storage_container_t * con = (storage_container_t *) object;

						LOG_DEBUG("Handle %d maps to container %s.\n", thing, con->name);
						info->type = CONTAINER;

						rc = make_handle_array(con->objects_consumed, &info->info.container.objects_consumed);
						if (rc == 0) {
							rc = make_handle_array(con->objects_produced, &info->info.container.objects_produced);
							if (rc == 0) {

								rc = ensure_app_handle(con,
										       CONTAINER,
										       &con->app_handle);
								if (rc == 0) {
									info->info.container.handle = con->app_handle;
								}

								if (rc == 0) {
									if (con->plugin != NULL) {
										rc = ensure_app_handle(con->plugin,
												       PLUGIN,
												       &con->plugin->app_handle);
										if (rc == 0) {
											info->info.container.plugin = con->plugin->app_handle;
										}
									} else {
										info->info.container.plugin = 0;
									}
								}

								if (rc == 0) {
									if (con->disk_group != NULL) {

										rc = ensure_app_handle(con->disk_group,
												       CONTAINER,
												       &con->disk_group->app_handle);
										if (rc == 0) {
											info->info.container.disk_group = con->disk_group->app_handle;
										}
									} else {
										info->info.container.disk_group = 0;
									}
								}

								if (rc == 0) {
									info->info.container.flags = con->flags;
									memcpy(info->info.container.name, con->name, sizeof(con->name));
									info->info.container.size = con->size;
								}
							}
						}
					}
					break;

				case VOLUME:
					{
						logical_volume_t * vol = (logical_volume_t *) object;

						LOG_DEBUG("Handle %d maps to volume %s.\n", thing, vol->name);
						info->type = VOLUME;

						rc = ensure_app_handle(vol,
								       VOLUME,
								       &vol->app_handle);
						if (rc == 0) {
							info->info.volume.handle = vol->app_handle;
						}

						if (rc == 0) {
							if (vol->file_system_manager != NULL) {
								rc = ensure_app_handle(vol->file_system_manager,
										       PLUGIN,
										       &vol->file_system_manager->app_handle);
								if (rc == 0) {
									info->info.volume.file_system_manager = vol->file_system_manager->app_handle;
								}
							} else {
								info->info.volume.file_system_manager = 0;
							}
						}

						if (rc == 0) {
							if (vol->object != NULL) {
								rc = ensure_app_handle(vol->object,
										       vol->object->object_type,
										       &vol->object->app_handle);
								if (rc == 0) {
									info->info.volume.object = vol->object->app_handle;
								}
							} else {
								info->info.volume.object = 0;
							}
						}

						if (rc == 0) {
							if (vol->associated_volume != NULL) {
								rc = ensure_app_handle(vol->associated_volume,
										       VOLUME,
										       &vol->associated_volume->app_handle);
								if (rc == 0) {
									info->info.volume.associated_volume = vol->associated_volume->app_handle;
								}
							} else {
								info->info.volume.associated_volume = 0;
							}
						}

						if (rc == 0) {
							if (vol->disk_group != NULL) {

								rc = ensure_app_handle(vol->disk_group,
										       CONTAINER,
										       &vol->disk_group->app_handle);
								if (rc == 0) {
									info->info.volume.disk_group = vol->disk_group->app_handle;
								}
							} else {
								info->info.volume.disk_group = 0;
							}
						}

						if (rc == 0) {
							if (vol->mount_point != NULL) {
                                                                info->info.volume.mount_point = engine_alloc(strlen(vol->mount_point) + 1);
								strcpy(info->info.volume.mount_point, vol->mount_point);
							} else {
								info->info.volume.mount_point = NULL;
							}

							/*
							 * Reget volume limits since things like
							 * min_fs_size can change if the volume is live.
							 */
							get_volume_sizes_and_limits(vol);

							info->info.volume.fs_size       = vol->fs_size;
							info->info.volume.min_fs_size   = vol->min_fs_size;
							info->info.volume.max_fs_size   = vol->max_fs_size;
							info->info.volume.vol_size      = vol->vol_size;
							info->info.volume.max_vol_size  = vol->max_vol_size;
							info->info.volume.dev_major     = vol->dev_major;
							info->info.volume.dev_minor     = vol->dev_minor;
							info->info.volume.serial_number = vol->serial_number;
							info->info.volume.flags         = vol->flags;
							memcpy(info->info.volume.name, vol->name, sizeof(vol->name));
							memcpy(info->info.volume.dev_node, vol->dev_node, sizeof(vol->dev_node));
						}
					}
					break;

				default:
					LOG_DEBUG("Handle %d maps to unknown object type %d.\n", thing, type);
					rc = EINVAL;
					break;
				}
			}

			if ((rc != 0) && (info != NULL)) {
				evms_free(info);
				info = NULL;
			}


		} else {
			rc = ENOMEM;
		}

		*user_info = info;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int add_plugin_to_list(ADDRESS Object,
			      TAG     Tag,
			      ADDRESS ObjectHandle,
			      ADDRESS Parameters) {

	int rc = DLIST_SUCCESS;
	storage_object_t * object = (storage_object_t *) Object;
	dlist_t plugin_list = (dlist_t) Parameters;
	plugin_record_t * pPlugRec;
	ADDRESS handle;

	LOG_PROC_ENTRY();

	/* Only get the plug-in (feature) if this is an EVMS storage object. */
	if (object->object_type == EVMS_OBJECT) {

		/* Make sure the plug-in isn't already on the list. */
		GoToStartOfList(plugin_list);

		rc = GetObject(plugin_list,
			       PLUGIN_TAG,
			       NULL,
			       FALSE,
			       (ADDRESS *) &pPlugRec);

		while ((rc == DLIST_SUCCESS) && (pPlugRec != NULL) && (pPlugRec != object->plugin)) {
			rc = GetNextObject(plugin_list,
					   PLUGIN_TAG,
					   (ADDRESS *) &pPlugRec);
		}

		if ((rc == DLIST_EMPTY) ||
		    (rc == DLIST_END_OF_LIST)) {
			/*
			 * This plug-in record is not in the list.  Insert it into the
			 * list.
			 */
			rc = InsertObject(plugin_list,
					  object->plugin,
					  PLUGIN_TAG,
					  NULL,
					  AppendToList,
					  FALSE,
					  &handle);
		}

		if (rc == DLIST_SUCCESS) {
			/* Get the feature(s) for the child object(s). */
			rc = ForEachItem(object->child_objects,
					 add_plugin_to_list,
					 plugin_list,
					 TRUE);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_get_feature_list(object_handle_t    thing,
			  handle_array_t * * plugins) {
	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {
		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			if ((type == VOLUME) ||
			    (type == EVMS_OBJECT)) {
				dlist_t feature_list = CreateList();

				if (feature_list != NULL) {
					storage_object_t * storage_object;

					if (type == VOLUME) {
						logical_volume_t * volume = (logical_volume_t *) object;

						storage_object = volume->object;
					} else {
						storage_object = (storage_object_t *) object;
					}

					rc = add_plugin_to_list(storage_object,
								EVMS_OBJECT_TAG,
								NULL,
								feature_list);

					if (rc == DLIST_SUCCESS) {
						rc = make_user_handle_array(feature_list, plugins);
					}

					DestroyList(&feature_list, FALSE);


				} else {
					rc = ENOMEM;
				}


			} else {
				/* The thing is not a volume or object. */
				rc = EINVAL;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int find_object_handle_by_name(dlist_t list, char * name, object_handle_t * handle) {

	int rc = 0;
	storage_object_t * obj;
	TAG tag;
	boolean found = FALSE;

	LOG_PROC_ENTRY();

	GoToStartOfList(list);

	do {
		rc = BlindGetObject(list,
				    &tag,
                                    NULL,
				    FALSE,
				    (ADDRESS *) &obj);

		if (rc == DLIST_SUCCESS) {
			found = (strcmp(obj->name, name) == 0);

			if (!found) {
				rc = NextItem(list);
			}
		}
	} while ((rc == DLIST_SUCCESS) && !found);

	if (found) {
		rc = ensure_app_handle(obj,
				       obj->object_type,
				       &obj->app_handle);

		if (rc == 0) {
			*handle = obj->app_handle;
		}

	} else {
		rc = ENOENT;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int find_container_handle_by_name(char * name, object_handle_t * handle) {

	int rc = 0;
	storage_container_t * con;
	TAG tag;
	boolean found = FALSE;

	GoToStartOfList(ContainerList);

	do {
		rc = BlindGetObject(ContainerList,
				    &tag,
                                    NULL,
				    FALSE,
				    (ADDRESS *) &con);

		if (rc == DLIST_SUCCESS) {
			found = (strcmp(con->name, name) == 0);

			if (!found) {
				rc = NextItem(ContainerList);
			}
		}
	} while ((rc == DLIST_SUCCESS) && !found);

	if (found) {
		rc = ensure_app_handle(con,
				       CONTAINER,
				       &con->app_handle);
		if (rc == 0) {
			*handle = con->app_handle;
		}

	} else {
		rc = ENOENT;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int find_volume_handle_by_name(char * name, object_handle_t * handle) {

	int rc = 0;
	logical_volume_t * vol;
	TAG tag;
	boolean found = FALSE;

	GoToStartOfList(VolumeList);

	do {
		rc = BlindGetObject(VolumeList,
				    &tag,
                                    NULL,
				    FALSE,
				    (ADDRESS *) &vol);

		if (rc == DLIST_SUCCESS) {
			found = (strcmp(vol->name, name) == 0);

			if (!found) {
				rc = NextItem(VolumeList);
			}
		}
	} while ((rc == DLIST_SUCCESS) && !found);

	if (found) {
		rc = ensure_app_handle(vol,
				       VOLUME,
				       &vol->app_handle);
		if (rc == 0) {
			*handle = vol->app_handle;
		}

	} else {
		rc = ENOENT;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_get_object_handle_for_name(object_type_t type, char * name, object_handle_t * object_handle) {
	int rc = 0;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {
		rc = ENOENT;

		if (type & DISK) {
			rc = find_object_handle_by_name(DiskList, name, object_handle);
		}

		if ((rc == ENOENT) && (type & SEGMENT)) {
			rc = find_object_handle_by_name(SegmentList, name, object_handle);
		}

		if ((rc == ENOENT) && (type & REGION)) {
			rc = find_object_handle_by_name(RegionList, name, object_handle);
		}

		if ((rc == ENOENT) && (type & EVMS_OBJECT)) {
			rc = find_object_handle_by_name(EVMSObjectList, name, object_handle);
		}

		if ((rc == ENOENT) && (type & CONTAINER)) {
			rc = find_container_handle_by_name(name, object_handle);
		}

		if ((rc == ENOENT) && (type & VOLUME)) {
			rc = find_volume_handle_by_name(name, object_handle);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_direct_plugin_communication(plugin_handle_t plugin_handle,
				     object_handle_t thing,
				     boolean         target_kernel_plugin,
				     void          * arg) {

	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(plugin_handle,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			if (type == PLUGIN) {
				plugin_record_t * plugin = (plugin_record_t *) object;

				rc = translate_handle(thing,
						      &object,
						      &type);

				if (rc == HANDLE_MANAGER_NO_ERROR) {
					switch (GetPluginID(plugin->id)) {
					case EVMS_DEVICE_MANAGER:
					case EVMS_SEGMENT_MANAGER:
					case EVMS_REGION_MANAGER:
					case EVMS_FEATURE:
					case EVMS_ASSOCIATIVE_FEATURE:{

							LOG_DEBUG("Direct communication going to the %s feature.\n", plugin->short_name);

							rc = plugin->functions.plugin->direct_plugin_communication(object, target_kernel_plugin, arg);
						}
						break;

					default:
						rc = EINVAL;
						break;
					}
				}

			} else {
				rc = EINVAL;
			}
		}
	}

	if (rc == 0) {
		/*
		 * Direct plug-in communication can add or remove child
		 * objects to/from the given object.  Since we don't know
		 * what went on inside the set_info() and how it affected
		 * other objects, update the stop data state on all the
		 * objects that can have stop data.
		 */
		update_all_stop_data_states();
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

