/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: container.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "fullengine.h"
#include "container.h"
#include "engine.h"
#include "discover.h"
#include "handlemgr.h"
#include "common.h"
#include "internalAPI.h"
#include "object.h"
#include "message.h"


/*
 * Ask the plug-in if the object can be added to the container.
 */
int evms_can_add_to_container(object_handle_t object_handle,
			      object_handle_t container_handle) {

	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(object_handle,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			if ((type == DISK) ||
			    (type == SEGMENT) ||
			    (type == REGION) ||
			    (type == EVMS_OBJECT)) {

				storage_object_t * obj = (storage_object_t *) object;

				if (obj->consuming_container != NULL) {
					if (obj->consuming_container->plugin->container_functions != NULL) {

						rc = translate_handle(container_handle,
								      &object,
								      &type);

						if (rc == HANDLE_MANAGER_NO_ERROR) {
							if (type == CONTAINER) {
								storage_container_t * con = (storage_container_t *) object;

								if (con->plugin->container_functions != NULL) {
									rc = con->plugin->container_functions->can_add_object(obj, con);

								} else {
									LOG_WARNING("The plug-in %s for container %s does not support containers.\n", con->plugin->short_name, con->name);
									rc = EINVAL;
								}

							} else {
								LOG_DETAILS("Handle %d is not for a container.\n", container_handle);
							}
						}

					} else {
						LOG_DETAILS("The plug-in %s for object %s does not support containers.\n", obj->plugin->short_name, obj->name);
						rc = EINVAL;
					}

				} else {
					LOG_DETAILS("Object %s is already in container %s.\n", obj->name, obj->consuming_container->name);
				}

			} else {
				LOG_DETAILS("Handle %d is for a thing of type %d which cannot belong in a container.\n", object_handle, type);
				rc = EINVAL;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Ask the plug-in if the object can be removed from its container.
 */
int evms_can_remove_from_container(object_handle_t object_handle) {

	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(object_handle,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			if ((type == DISK) ||
			    (type == SEGMENT) ||
			    (type == REGION) ||
			    (type == EVMS_OBJECT)) {

				storage_object_t * obj = (storage_object_t *) object;

				if (obj->consuming_container != NULL) {
					if (obj->consuming_container->plugin->container_functions != NULL) {

						rc = obj->consuming_container->plugin->container_functions->can_remove_object(obj);

					} else {
						LOG_WARNING("The plug-in %s for container %s does not support containers.\n", obj->plugin->short_name, obj->consuming_container->name);
						rc = EINVAL;
					}

				} else {
					LOG_DETAILS("Object %s is not in a container.\n", obj->name);
					rc = EINVAL;
				}

			} else {
				LOG_DETAILS("Handle %d is for a thing of type %d which cannot belong in a container.\n", object_handle, type);
				rc = EINVAL;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_create_container(plugin_handle_t   plugin_handle,
			  handle_array_t  * input_objects,
			  option_array_t  * options,
			  object_handle_t * new_container_handle) {
	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(plugin_handle,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			if (type == PLUGIN) {
				plugin_record_t * pPlugRec = (plugin_record_t *) object;

				/* Make sure the plug-in does containers. */
				if (pPlugRec->container_functions != NULL) {

					dlist_t object_list = CreateList();

					if (object_list != NULL) {
						rc = make_dlist(input_objects, object_list);

						if (rc == 0) {
							storage_object_t * obj;
							storage_container_t * disk_group = NULL;

							obj = dlist_get_current_item(object_list);
							if (obj != NULL) {
								disk_group = obj->disk_group;
							}

							/*
							 * Make sure that the input object list contains
							 * top level objects and none of them insist on
							 * being a top object.
							 */
							rc = ForEachItem(object_list,
									 isa_valid_input_object,
									 disk_group,
									 TRUE);

							if (rc == 0) {
								storage_container_t * new_container;

								rc = pPlugRec->container_functions->create_container(object_list, options, &new_container);

								if (rc == 0) {
									storage_object_t * child_obj = NULL;

									/*
									 * Now that the objects are consumed
									 * any stop data on them can be
									 * overwritten.
									 */
									DLIST_FOR_EACH(child_obj, new_container->objects_consumed, rc) {
										child_obj->flags &= ~SOFLAG_HAS_STOP_DATA;
									}

									if (new_container->disk_group != NULL) {
										propigate_cluster_info(new_container->objects_produced);
									}

									SortList(ContainerList, compare_containers);

									switch (GetPluginType(pPlugRec->id)) {
									case EVMS_DEVICE_MANAGER:
										SortList(DiskList, compare_objects);
										break;
									case EVMS_SEGMENT_MANAGER:
										SortList(SegmentList, compare_objects);
										break;
									case EVMS_REGION_MANAGER:
										SortList(RegionList, compare_objects);
										break;
									case EVMS_FEATURE:
									case EVMS_ASSOCIATIVE_FEATURE:
										SortList(EVMSObjectList, compare_objects);
										break;
									default:
										break;
									}

									rc = create_handle(new_container,
											   CONTAINER,
											   &new_container->app_handle);
									if (rc == 0) {
										*new_container_handle = new_container->app_handle;

									} else {
										LOG_WARNING("Error %d creating a handle for container %s.\n", rc, new_container->name);
									}
								}
							}
						}

						DestroyList(&object_list, FALSE);

					} else {
						rc = ENOMEM;
					}
				}

			} else {
				rc = EINVAL;
			}

		} else {
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * evms_get_container_list returns a pointer to a handle_array_t with handles for
 * storage containers, optionally filtering on the region manager that manages
 * the container.  If the object handle for the region manager is 0, handles for
 * all of the storage containers will be returned.
 */
int evms_get_container_list(object_handle_t          plugin_handle,
			    object_handle_t          disk_group_handle,
			    container_search_flags_t flags,
			    handle_array_t       * * container_handle_list) {
	int rc = 0;
	void * object = NULL;
	object_type_t type;
	plugin_record_t * pPlugRec = NULL;
	storage_container_t * disk_group = NULL;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {
		if (plugin_handle != 0) {
			/* Translate the handle for the region manager to make sure it is valid */
			/* and to get the plugin_record_t for the region manager. */
			rc = translate_handle(plugin_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				if (type == PLUGIN) {
					pPlugRec = (plugin_record_t *) object;
				} else {
					rc = EINVAL;
				}
			}
		}

		if (disk_group_handle != 0){
			/*
			 * Translate the handle for the disk group to make sure
			 * it is valid and to get the disk group
			 * storage_container_t.
			 */
			rc = translate_handle(disk_group_handle,
					      &object,
					      &type);

			if (rc == HANDLE_MANAGER_NO_ERROR) {
				if (type == CONTAINER) {
					disk_group = (storage_container_t *) object;
				} else {
					rc = EINVAL;
				}
			}
		}

		if (rc == 0) {
			dlist_t container_list;

			rc = engine_get_container_list(pPlugRec, disk_group, flags, &container_list);

			if (rc == 0) {
				rc = make_user_handle_array(container_list, container_handle_list);

				/* We are finished with the list that was returned by */
				/* Engine_GetContainerList. */
				DestroyList(&container_list, FALSE);
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

