/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: discover.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "fullengine.h"
#include "discover.h"
#include "engine.h"
#include "internalAPI.h"
#include "crc.h"
#include "volume.h"
#include "commit.h"
#include "message.h"
#include "dm.h"
#include "memman.h"
#include "cluster.h"
#include "remote.h"
#include "volume.h"
#include <ldm_dpc.h>


boolean discover_in_progress = FALSE;

static int validate_feature_header(evms_feature_header_t * fh);

/******************************************************************************
 *
 * General purpose utility functions
 *
 ******************************************************************************/


/*
 * remove_corrupt_object() is used as the function of PruneList().  The caller
 * can issue a PruneList() on a list of storage_object_t structures and can
 * specify the pruning function as remove_corrupt_object.  This function will
 * check if a storage_object_t has its SOFLAG_CORRUPT set.  If so, it will
 * return TRUE so that the object is removed from the current list.
 */
boolean remove_corrupt_object(ADDRESS   object,
			      TAG       object_tag,
			      ADDRESS   object_handle,
			      ADDRESS   parameters,
			      boolean * free_memory,
			      uint    * error) {
	boolean result = FALSE;

	LOG_PROC_ENTRY();

	/* Never free item memory. */
	*free_memory = FALSE;

	/* Only handle storage objects. */
	if ((object_tag == DISK_TAG) ||
	    (object_tag == SEGMENT_TAG) |
	    (object_tag == REGION_TAG) ||
	    (object_tag == EVMS_OBJECT)) {

		storage_object_t * obj = (storage_object_t *) object;

		LOG_DEBUG("Examining object %s.\n", obj->name);

		if (obj->flags & SOFLAG_CORRUPT) {
			result = TRUE;
		}
	}

	/*
	 * Don't stop PruneList, even if we encounter an error.  The error was
	 * logged.
	 */
	*error = DLIST_SUCCESS;

	LOG_PROC_EXIT_BOOLEAN(result);
	return result;
}


static int set_not_claimed(ADDRESS object,
			   TAG     object_tag,
			   ADDRESS object_handle,
			   ADDRESS parameters) {

	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	/* Safety check */
	if ((object_tag == DISK_TAG) ||
	    (object_tag == SEGMENT_TAG) ||
	    (object_tag == REGION_TAG)) {

		LOG_DEBUG("Setting SOFLAG_NOT_CLAIMED flag on object %s.\n", obj->name);
		obj->flags |= SOFLAG_NOT_CLAIMED;
	}

	LOG_PROC_EXIT_INT(0);
	return 0;
}


static boolean remove_unclaimed_object(ADDRESS   object,
				       TAG       object_tag,
				       ADDRESS   object_handle,
				       ADDRESS   parameters,
				       boolean * free_memory,
				       uint    * error) {
	boolean result = FALSE;
	storage_object_t * obj = (storage_object_t *) object;
	dlist_t unclaimed_object_list = (dlist_t) parameters;
	int rc;

	LOG_PROC_ENTRY();

	/* Never free item memory */
	*free_memory = FALSE;

	/* Safety check */
	if ((object_tag == DISK_TAG) ||
	    (object_tag == SEGMENT_TAG) ||
	    (object_tag == REGION_TAG)) {

		LOG_DEBUG("Examining object %s.\n", obj->name);

		if (obj->flags & SOFLAG_NOT_CLAIMED) {
			ADDRESS trash;

			obj->flags &= ~SOFLAG_NOT_CLAIMED;

			/* Put the unclaimed object on the unclaimed_object_list. */
			LOG_DEBUG("Put object %s on the unclaimed list.\n", obj->name);
			rc = InsertObject(unclaimed_object_list,
					  object,
					  object_tag,
					  NULL,
					  AppendToList,
					  FALSE,
					  &trash);

			if (rc != DLIST_SUCCESS) {
				LOG_WARNING("Error code %d when putting object %s on the unclaimed object list.\n", rc, obj->name);
			}

			/*
			 * Return TRUE so that the object is removed from the list that is
			 * being processed.
			 */
			result = TRUE;

		} else {
			LOG_DEBUG("Object %s is not marked not claimed.\n", obj->name);
		}
	}

	LOG_PROC_EXIT_BOOLEAN(result);
	*error = 0;
	return result;
}


/*
 * Check if a dev node exists for a volume and that is has the correct minor
 * number.
 * Returns:
 * 0        dev node exists with the correct major:minor
 * ENOENT   dev node does not exist
 * EEXIST   dev node exists but has the wrong major:minor
 * other    some unexpected error code from stat()
 */
int hasa_dev_node(char * name, u_int32_t major, u_int32_t minor) {

	int rc = 0;
	struct stat statbuf;

	LOG_PROC_ENTRY();

	rc = stat(name, &statbuf);

	if (rc == 0) {

		/* dev node exists. Check if it has the right minor number. */
		dev_t devt = makedev(major, minor);

		if (statbuf.st_rdev != devt) {

			/* dev node exists but it has the wrong minor number. */
			rc = EEXIST;
		}

	} else {
		/* stat() failed.  Return the error code. */
		rc = errno;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int ensure_dev_node(char * name, u_int32_t major, u_int32_t minor) {

	int rc = 0;
	dev_t devt = makedev(major, minor);
	char dir_name[EVMS_OBJECT_NODE_PATH_LEN + EVMS_NAME_SIZE + 1];

	LOG_PROC_ENTRY();

	/* Make sure major:minor is valid. */
	if (major != 0) {
		switch (hasa_dev_node(name, major, minor)) {
		case 0:
			/* Dev node exists and has the correct major:minor. */
			break;

		case EEXIST:
			/* Dev node exists but has the wrong major:minor.  Fix it. */
			LOG_DEBUG("The device node for volume \"%s\" has the wrong major:minor.  I'm fixing it to %d:%d.\n", name, major, minor);

			rc = unlink(name);
			if (rc == 0) {
				rc = mknod(name, S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, devt);

				if (rc != 0) {
					LOG_WARNING("mknod(%s) failed with error code %d.\n", name, rc);
				}
			} else {
				rc = errno;
				LOG_WARNING("unlink(\"%s\") failed with error code %d: %s\n", name, rc, strerror(rc));
			}
			break;

		case ENOENT:
			strcpy(dir_name, name);
			*(strrchr(dir_name, '/')) = '\0';

			rc = make_directory(dir_name, NO_MESSAGES);

			if (rc == 0) {
				rc = mknod(name, S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, devt);

				if (rc != 0) {
					LOG_WARNING("mknod(%s) failed with error code %d.\n", name, rc);
				}


			} else {
				LOG_WARNING("Error code %d when making directory %s.\n", rc, dir_name);
			}
			break;

		default:
			LOG_WARNING("hasa_dev_node() for \"%s\", major %d, minor %d failed with unexpected error %d: %s.\n", name, major, minor, rc, strerror(rc));

		}

	} else {
		LOG_ERROR("Major:minor %d:%d is not valid.\n", major, minor);
		rc = EINVAL;
	}

	if (rc == 0) {
		LOG_DEBUG("Device node %s is for major %d, minor %d.\n", name, major, minor);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int make_object_dev_node(ADDRESS object,
			 TAG     object_tag,
			 ADDRESS object_handle,
			 ADDRESS parameters) {

	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	if (obj->flags & SOFLAG_ACTIVE) {
		char dev_name[EVMS_OBJECT_NODE_PATH_LEN + EVMS_NAME_SIZE + 1];

		strcpy(dev_name, EVMS_OBJECT_NODE_PATH);

		if (is_local_object(obj)) {
			if ((memcmp(&current_nodeid, &no_nodeid, sizeof(ece_nodeid_t)) != 0) &&
			    (memcmp(&current_nodeid, &my_nodeid, sizeof(ece_nodeid_t)) != 0)) {
				/*
				 * This is a remote object on shared storage which
				 * we will access locally.  Add the node's name to the
				 * device name.
				 */
				strcat(dev_name, nodeid_to_string(&current_nodeid));
				strcat(dev_name, "/");
			}
			strcat(dev_name, obj->name);
			ensure_dev_node(dev_name, obj->dev_major, obj->dev_minor);

		} else {
			strcat(dev_name, obj->name);
			remote_ensure_dev_node(&current_nodeid, dev_name, obj->dev_major, obj->dev_minor);
		}
	}

	LOG_PROC_EXIT_INT(DLIST_SUCCESS);
	return DLIST_SUCCESS;
}


static void propigate_cluster_info_to_container(storage_container_t * con) {
	
	storage_object_t * child_obj = dlist_get_current_item(con->objects_consumed);

	if ((child_obj->disk_group != NULL) &&
	    (con->disk_group == NULL)) {
		con->disk_group = child_obj->disk_group;

		if (child_obj->flags & SOFLAG_CLUSTER_PRIVATE) {
			con->flags |= SCFLAG_CLUSTER_PRIVATE;
		}
		if (child_obj->flags & SOFLAG_CLUSTER_SHARED) {
			con->flags |= SCFLAG_CLUSTER_SHARED;
		}
		if (child_obj->flags & SOFLAG_CLUSTER_DEPORTED) {
			con->flags |= SCFLAG_CLUSTER_DEPORTED;
		}
	}
}


void propigate_cluster_info(dlist_t objects) {

	int rc;
	storage_object_t * obj;

	DLIST_FOR_EACH(obj, objects, rc) {

		/* Disks don't have cluster info. */
		if (obj->object_type != DISK) {

			if (obj->producing_container != NULL) {
				storage_container_t * con = obj->producing_container;

				propigate_cluster_info_to_container(obj->producing_container);

				obj->disk_group = con->disk_group;

				if (con->flags & SCFLAG_CLUSTER_PRIVATE) {
					obj->flags |= SOFLAG_CLUSTER_PRIVATE;
				}
				if (con->flags & SCFLAG_CLUSTER_SHARED) {
					obj->flags |= SOFLAG_CLUSTER_SHARED;
				}
				if (con->flags & SCFLAG_CLUSTER_DEPORTED) {
					obj->flags |= SOFLAG_CLUSTER_DEPORTED;
				}

			} else {
				storage_object_t * child_obj = dlist_get_current_item(obj->child_objects);

				if (child_obj != NULL) {
					obj->disk_group = child_obj->disk_group;
					obj->flags |= (child_obj->flags & (SOFLAG_CLUSTER_PRIVATE | SOFLAG_CLUSTER_SHARED | SOFLAG_CLUSTER_DEPORTED));
				}
			}
		}
	}
}


/*
 * Return a string with a number of periods based on *status_count.
 * Increment *status_count.
 */
char * status_dots(int * status_count) {

	char * ptr;

	switch (*status_count % 4) {
	case 0:
		ptr = "...";
		break;
	case 1:
		ptr = "   ";
		break;
	case 2:
		ptr = ".  ";
		break;
	case 3:
		ptr = ".. ";
		break;
	default:
		ptr = "";
	}

	(*status_count)++;

	return ptr;
}


/******************************************************************************
 *
 * Functions for discovering logical disks
 *
 ******************************************************************************/

int discover_logical_disks(dlist_t object_list) {

	int rc = 0;
	int status_count = 0;
	plugin_record_t * pPlugRec = NULL;
	dlist_t result_object_list = CreateList();

	LOG_PROC_ENTRY();

	if (result_object_list != NULL) {

		/*
		 * Call all of the device manager plug-ins to have them discover their
		 * disks.
		 */
		GoToStartOfList(PluginList);
		rc = GetObject(PluginList,
			       PLUGIN_TAG,
			       NULL,
			       TRUE,
			       (ADDRESS *) &pPlugRec);

		while (pPlugRec != NULL) {
			if (GetPluginType(pPlugRec->id) == EVMS_DEVICE_MANAGER) {
				status_message("Discovering disks%s\n", status_dots(&status_count));
				LOG_DEBUG("Calling discover() in %s.\n", pPlugRec->short_name);
				rc = pPlugRec->functions.plugin->discover(object_list, result_object_list, TRUE);
				LOG_DEBUG("Return code from discover() is %d.\n", rc);

				/* Remove any corrupt objects from the list. */
				PruneList(result_object_list,
					  remove_corrupt_object,
					  NULL);

				/* Make sure we have a dev node for each object. */
				ForEachItem(result_object_list,
					    make_object_dev_node,
					    NULL,
					    TRUE);

				/* Remove any remaining objects from the source list. */
				DeleteAllItems(object_list, FALSE);

				/*
				 * Put the objects from the result list onto the source
				 * list.
				 */
				AppendList(object_list, result_object_list);
			}

			rc = GetNextObject(PluginList,
					   PLUGIN_TAG,
					   (ADDRESS *) &pPlugRec);
		}

		if ((rc == DLIST_END_OF_LIST) ||
		    (rc == DLIST_EMPTY)) {
			rc = DLIST_SUCCESS;
		}

		DestroyList(&result_object_list, FALSE);

	} else {
		/* There was an error in creating the result_object_list. */
		LOG_CRITICAL("Error allocating memory for the result object list.\n");
		rc = ENOMEM;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int discover_remote_logical_disks(ece_nodeid_t * node, dlist_t object_list) {

	int rc = 0;
	u_int32_t          num_disks;
	storage_object_t * disks;

	LOG_PROC_ENTRY();

	status_message("Discovering disks...\n");
	rc = remote_discover_disks(node, &num_disks, &disks);

        LOG_DEBUG("Number of discovered disks: %d.\n", num_disks);

	if (rc == 0) {
		int i;
		storage_object_t * new_disk;

		for (i = 0; (i < num_disks) && (rc == 0); i++) {
			rc = engine_allocate_logical_disk(disks[i].name, &new_disk);

			if (rc == 0) {
				void * trash;
				unsigned long plugin_id = (unsigned long)disks[i].plugin;

				new_disk->app_handle             = disks[i].app_handle;
				new_disk->object_type            = disks[i].object_type;
				new_disk->data_type              = disks[i].data_type;
				new_disk->dev_major 	         = disks[i].dev_major;
				new_disk->dev_minor 	         = disks[i].dev_minor;

				/*
				 * Remote daemon substitues plug-in ID for
				 * plug-in pointer.
				 */
			//	new_disk->plugin                 = disks[i].plugin;
				engine_get_plugin_by_ID((plugin_id_t) plugin_id, &(new_disk->plugin));

			//	new_disk->producing_container    = disks[i].producing_container;
			//	new_disk->consuming_container    = disks[i].consuming_container;
			//	new_disk->parent_objects         = disks[i].parent_objects;
			//	new_disk->child_objects          = disks[i].child_objects;
			//	new_disk->associated_object      = disks[i].associated_object;
				new_disk->flags                  = disks[i].flags;
				new_disk->start                  = disks[i].start;
				new_disk->size 	                 = disks[i].size;
			//	new_disk->volume 	         = disks[i].volume;
			//	new_disk->feature_header;        = disks[i].feature_header;
				new_disk->geometry               = disks[i].geometry;
				new_disk->private_data           = engine_alloc(sizeof(int));
				if (new_disk->private_data == NULL) {
					LOG_CRITICAL("Error getting memory for the disk's private data.\n");
					rc = ENOMEM;
					engine_free(new_disk);
					continue;
				}
			//	new_disk->consuming_private_data = disks[i].consuming_private_data;
				memcpy(&new_disk->name, disks[i].name, sizeof(disks[i].name));
				memcpy(&new_disk->uuid, disks[i].uuid, sizeof(disks[i].uuid));
				new_disk->disk_group             = disks[i].disk_group;

				rc = InsertObject(object_list,
						  new_disk,
						  DISK_TAG,
						  NULL,
						  AppendToList,
						  FALSE,
						  &trash);

				if (rc != DLIST_SUCCESS) {
					LOG_SERIOUS("Error code %d when inserting disk %s into the output list.\n", rc, new_disk->name);
					engine_free(new_disk);
					continue;
				}
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/******************************************************************************
 *
 * Functions for discovering disk segments
 *
 ******************************************************************************/

int discover_segments(dlist_t object_list) {

	int rc = 0;
	int status_count = 0;
	plugin_record_t * pPlugRec = NULL;
	dlist_t result_object_list = CreateList();
	dlist_t unclaimed_object_list = CreateList();
	boolean finished = FALSE;
	boolean last_pass = FALSE;

	LOG_PROC_ENTRY();

	if ((result_object_list != NULL) &&
	    (unclaimed_object_list != NULL)) {

		while ((rc == 0) && (!finished || last_pass)) {

			/*
			 * Mark all the objects in the list as unclaimed so that they
			 * can be removed at the end of the loop if no plug-in claims
			 * them.
			 */

			ForEachItem(object_list,
				    set_not_claimed,
				    NULL,
				    TRUE);

			/* Assume we will finish discovery on this loop. */
			finished = TRUE;

			GoToStartOfList(PluginList);
			rc = GetObject(PluginList,
				       PLUGIN_TAG,
				       NULL,
				       TRUE,
				       (ADDRESS *) &pPlugRec);

			while (pPlugRec != NULL) {

				if (GetPluginType(pPlugRec->id) == EVMS_SEGMENT_MANAGER) {
					status_message("Discovering segments%s\n", status_dots(&status_count));
					LOG_DEBUG("Calling discover() in %s.  last_pass is %s.\n", pPlugRec->short_name, (last_pass) ? "TRUE" : "FALSE");
					rc = pPlugRec->functions.plugin->discover(object_list, result_object_list, last_pass);
					LOG_DEBUG("Return code from discover() is %d.\n", rc);

					/* Remove any corrupt objects from the list. */
					PruneList(result_object_list,
						  remove_corrupt_object,
						  NULL);

					/* Make sure we have a dev node for each object. */
					ForEachItem(result_object_list,
						    make_object_dev_node,
						    NULL,
						    TRUE);

					propigate_cluster_info(result_object_list);

					/*
					 * Segment managers will return a positive number (the
					 * number of objects discovered) if they find any
					 * segments.  If rc is positive, turn off the finished
					 * flag so that we make another pass at discovery
					 * through the segment managers.  Segment managers
					 * return a 0 if they discovered nothing and had no
					 * errors.  They return a negative error code if
					 * something went wrong.  We ignore error codes to let
					 * the discovery continue.
					 */
					if (!last_pass) {
						if (rc > 0) {
							finished = FALSE;
						}
					}

					rc = 0;

					/*
					 * Remove any remaining objects from the source list.
					 */
					DeleteAllItems(object_list, FALSE);

					/*
					 * Put the objects from the result list onto the source
					 * list.  AppendList deletes them from the
					 * result_object_list.
					 */
					AppendList(object_list, result_object_list);
				}

				rc = GetNextObject(PluginList,
						   PLUGIN_TAG,
						   (ADDRESS *) &pPlugRec);
			}

			if ((rc == DLIST_END_OF_LIST) ||
			    (rc == DLIST_EMPTY)) {
				rc = DLIST_SUCCESS;
			}


			/*
			 * Remove any objects that have not been claimed and put them on the
			 * unclaimed_object_list.  These objects will never be discovered in
			 * subsequent passes of discovery.  Therefore we set them aside so
			 * that all the plug-ins don't bother reexamining them on every pass
			 * of discovery.
			 */
			PruneList(object_list,
				  remove_unclaimed_object,
				  unclaimed_object_list);

			if (finished && !last_pass) {
				last_pass = TRUE;
			} else {
				last_pass = FALSE;
			}
		}

		/*
		 * Put the unclaimed objects back on the object_list so that they
		 * can be examined by the next layer of discovery.
		 */
		CopyList(object_list, unclaimed_object_list, InsertAtStart);

		DestroyList(&result_object_list, FALSE);
		DestroyList(&unclaimed_object_list, FALSE);

	} else {
		/*
		 * There was an error in creating the result_object_list or the
		 * unclaimed_object_list.
		 */
		if (result_object_list == NULL) {
			LOG_CRITICAL("Error allocating memory for the result object list.\n");
		} else {
			DestroyList(&result_object_list, FALSE);
		}
		if (unclaimed_object_list == NULL) {
			LOG_CRITICAL("Error allocating memory for the unclaimed object list.\n");
		} else {
			DestroyList(&unclaimed_object_list, FALSE);
		}
		rc = ENOMEM;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/******************************************************************************
 *
 * Functions for discovering regions
 *
 ******************************************************************************/

int discover_regions(dlist_t object_list) {

	int rc = 0;
	int status_count = 0;
	plugin_record_t * pPlugRec = NULL;
	dlist_t result_object_list = CreateList();
	dlist_t unclaimed_object_list = CreateList();
	boolean finished = FALSE;
	boolean last_pass = FALSE;

	LOG_PROC_ENTRY();

	if ((result_object_list != NULL) &&
	    (unclaimed_object_list != NULL)) {

		while ((rc == 0) && (!finished || last_pass)) {

			/*
			 * Mark all the objects in the list as unclaimed so that they
			 * can be removed at the end of the loop if no plug-in claims
			 * them.
			 */

			ForEachItem(object_list,
				    set_not_claimed,
				    NULL,
				    TRUE);

			/* Assume we will finish discovery on this loop. */
			finished = TRUE;

			GoToStartOfList(PluginList);
			rc = GetObject(PluginList,
				       PLUGIN_TAG,
				       NULL,
				       TRUE,
				       (ADDRESS *) &pPlugRec);

			while (pPlugRec != NULL) {

				if (GetPluginType(pPlugRec->id) == EVMS_REGION_MANAGER) {
					status_message("Discovering regions%s\n", status_dots(&status_count));
					LOG_DEBUG("Calling discover() in %s.  last_pass is %s.\n", pPlugRec->short_name, (last_pass) ? "TRUE" : "FALSE");
					rc = pPlugRec->functions.plugin->discover(object_list, result_object_list, last_pass);
					LOG_DEBUG("Return code from discover() is %d.\n", rc);

					/* Remove any corrupt objects from the list. */
					PruneList(result_object_list,
						  remove_corrupt_object,
						  NULL);

					/* Make sure we have a dev node for each object. */
					ForEachItem(result_object_list,
						    make_object_dev_node,
						    NULL,
						    TRUE);

					propigate_cluster_info(result_object_list);

					/*
					 * Region managers will return a positive number (the
					 * number of objects discovered) if they find any
					 * regions.  If rc is positive, turn off the finished
					 * flag so that we make another pass at discovery
					 * through the region managers.  Region managers return
					 * a 0 if they discovered nothing and had no errors.
					 * They return a negative error code if something went
					 * wrong.  We ignore error codes to let the discovery
					 * continue.
					 */
					if (!last_pass) {
						if (rc > 0) {
							finished = FALSE;
						}
					}

					rc = 0;

					/*
					 * Remove any remaining objects from the source list.
					 */
					DeleteAllItems(object_list, FALSE);

					/*
					 * Put the objects from the result list onto the source
					 * list.  AppendList deletes them from the
					 * result_object_list.
					 */
					AppendList(object_list, result_object_list);
				}

				rc = GetNextObject(PluginList,
						   PLUGIN_TAG,
						   (ADDRESS *) &pPlugRec);
			}

			if ((rc == DLIST_END_OF_LIST) ||
			    (rc == DLIST_EMPTY)) {
				rc = DLIST_SUCCESS;
			}

			/*
			 * Remove any objects that have not been claimed and put them on the
			 * unclaimed_object_list.  These objects will never be discovered in
			 * subsequent passes of discovery.  Therefore we set them aside so
			 * that all the plug-ins don't bother reexamining them on every pass
			 * of discovery.
			 */
			PruneList(object_list,
				  remove_unclaimed_object,
				  unclaimed_object_list);

			if (finished && !last_pass) {
				last_pass = TRUE;
			} else {
				last_pass = FALSE;
			}
		}

		/*
		 * Put the unclaimed objects back on the object_list so that they
		 * can be examined by the next layer of discovery.
		 */
		CopyList(object_list, unclaimed_object_list, InsertAtStart);

		DestroyList(&result_object_list, FALSE);
		DestroyList(&unclaimed_object_list, FALSE);

	} else {
		/*
		 * There was an error in creating the result_object_list or the
		 * unclaimed_object_list.
		 */
		if (result_object_list == NULL) {
			LOG_CRITICAL("Error allocating memory for the result object list.\n");
		} else {
			DestroyList(&result_object_list, FALSE);
		}
		if (unclaimed_object_list == NULL) {
			LOG_CRITICAL("Error allocating memory for the unclaimed object list.\n");
		} else {
			DestroyList(&unclaimed_object_list, FALSE);
		}
		rc = ENOMEM;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/******************************************************************************
 *
 * Support functions for making volumes
 *
 ******************************************************************************/

/*
 * Set the volume pointer in an object.  Set the volume pointer in its child
 * objects.
 * This function has its parameters structured so that it can be called by
 * ForEachItem() to process a dlist_t of objects.
 */
int set_volume_in_object(ADDRESS object,
			 TAG     object_tag,
			 ADDRESS object_handle,
			 ADDRESS parameters) {

	logical_volume_t * volume = (logical_volume_t *) parameters;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	/*
	 * This function can be called for a child of any object.
	 */
	switch (object_tag) {
	case DISK_TAG:
	case SEGMENT_TAG:
	case REGION_TAG:
	case EVMS_OBJECT_TAG:
		{
			LOG_DEBUG("Set volume %s in object %s.\n", (volume != NULL) ? volume->name : "(none)", obj->name);
			
			if (!discover_in_progress) {
				/* If the volume for this object is changing and it has a
				 * feature header, mark the feature header dirty so that
				 * it will be rewritten with the new volume data.
				 */
				if (obj->volume != volume) {
					if (obj->feature_header != NULL) {
						obj->flags |= SOFLAG_FEATURE_HEADER_DIRTY;
					}
				}
			}

			obj->volume = volume;

			/*
			 * Clear the volume pointer in a disk object if it has multiple
			 * parents or its single parent is a segment manager (which
			 * could potentially make more parents).
			 */
			if (object_tag == DISK_TAG) {
				uint parent_count = 0;

				GetListSize(obj->parent_objects, &parent_count);
				if (parent_count >1) {
					obj->volume = NULL;

				} else {
					if (parent_count == 1) {
						storage_object_t * parent_object;
						TAG tag;
						int rc;

						rc = BlindGetObject(obj->parent_objects,
								    &tag,
								    NULL,
								    FALSE,
								    (ADDRESS *) &parent_object);

						if (rc == DLIST_SUCCESS) {

							/* If the parent is a segment manager,
							 * clear the volume pointer in the object.
							 */
							if (GetPluginType(parent_object->plugin->id) == EVMS_SEGMENT_MANAGER) {
								obj->volume = NULL;
							}

						} else {
							/*
							 * Couldn't get the parent object.
							 * Assume the worst; clear the volume pointer.
							 */
							obj->volume = NULL;
						}
					}
				}
			}

			/*
			 * If this is not discovery time (this function can be called on
			 * a create or revert of a volume) notify the feature of the
			 * object's change in volume status.
			 */
			if (!discover_in_progress) {
				obj->plugin->functions.plugin->set_volume(obj, (volume != NULL));
			}

			/*
			 * If the object was not produced by a container, then call
			 * this function recursively to set the volume pointers in
			 * the child objects.  If the object was produced by a
			 * container then its child objects may not belong exclusively
			 * to this volume, so we can't set the volume pointer in the
			 * child objects; the recursion stops.
			 */
			if (obj->producing_container == NULL) {
				ForEachItem(obj->child_objects,
					    set_volume_in_object,
					    volume,
					    TRUE);
			}
		}
		break;

	default:
		/*Ignore any other kind of objects. */
		break;
	}

	LOG_PROC_EXIT_INT(DLIST_SUCCESS);
	return DLIST_SUCCESS;
}

/*
 * If a volume has an FSIM, get the file system size and limits.
 * Else set default values.
 */
int get_volume_sizes_and_limits(logical_volume_t * vol) {

	int rc = 0;
	plugin_record_t * fsim = vol->file_system_manager;

	LOG_PROC_ENTRY();

	if (fsim == NULL) {
		/* Volume has no FSIM.  Set default sizes and limits. */
		vol->fs_size      = round_down_to_hard_sector(vol->vol_size, vol->object);
		vol->min_fs_size  = 0;
		vol->max_fs_size  = round_down_to_hard_sector(-1, vol->object);
		vol->max_vol_size = round_down_to_hard_sector(-1, vol->object);

	} else {
		/*
		 * If the FSIM has changed, don't call the new FSIM for the sizes and
		 * limits.
		 */
		if (fsim == vol->original_fsim) {
			int rc2;

			rc = fsim->functions.fsim->get_fs_size(vol, &vol->fs_size);

			if (rc != 0) {
				LOG_WARNING("FSIM %s returned error code %d from call to get_fs_size() for volume %s.\n", fsim->short_name, rc, vol->name);
			}

			rc2 = fsim->functions.fsim->get_fs_limits(vol, &vol->min_fs_size,
								  &vol->max_fs_size,
								  &vol->max_vol_size);

			if (rc2 != 0) {
				LOG_WARNING("FSIM %s returned error code %d from call to get_fs_limits() for volume %s.\n", fsim->short_name, rc, vol->name);
			}

			if (rc == 0) {
				rc = rc2;
			}

			/* Make sure sizes are multiples of the hard sector size. */
			vol->fs_size      = round_down_to_hard_sector(vol->fs_size, vol->object);
			vol->min_fs_size  = round_up_to_hard_sector(vol->min_fs_size, vol->object);
			vol->max_fs_size  = round_down_to_hard_sector(vol->max_fs_size, vol->object);
			vol->max_vol_size = round_down_to_hard_sector(vol->max_vol_size, vol->object);

		} else {
			if (vol->original_fsim == NULL) {
				LOG_DEBUG("Cannot get sizes and limits for volume %s because it does not have a file system on it right now.\n", vol->name);
			} else {
				LOG_DEBUG("Cannot get sizes and limits for volume %s because it is being converted from %s to %s.\n", vol->name, vol->original_fsim->short_name, vol->file_system_manager->short_name);
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Given a volume, see if there is an FSIM that will claim that it manages the
 * volume.
 */
void find_fsim_for_volume(logical_volume_t * volume) {

	int rc = 0;
	boolean found = FALSE;
	plugin_record_t * pPlugRec = NULL;

	LOG_PROC_ENTRY();

	volume->file_system_manager = NULL;

	GoToStartOfList(PluginList);
	rc = GetObject(PluginList,
		       PLUGIN_TAG,
		       NULL,
		       TRUE,
		       (ADDRESS *) &pPlugRec);

	while ((rc == DLIST_SUCCESS) && (pPlugRec != NULL) && (!found)) {

		if (GetPluginType(pPlugRec->id) == EVMS_FILESYSTEM_INTERFACE_MODULE) {
			found = (pPlugRec->functions.fsim->probe(volume) == 0);

			if (found) {
				LOG_DEBUG("Volume %s belongs to %s.\n", volume->name, pPlugRec->short_name);
				volume->file_system_manager = pPlugRec;
				volume->original_fsim = volume->file_system_manager;
				volume->original_fsim_private_data = volume->private_data;
			}
		}

		rc = GetNextObject(PluginList,
				   PLUGIN_TAG,
				   (ADDRESS *) &pPlugRec);
	}

	get_volume_sizes_and_limits(volume);

	LOG_PROC_EXIT_VOID();
}


int make_volume_dev_node(logical_volume_t * vol) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/*
	 * Make sure the volume has a valid major:minor before we
	 * make a dev node for it.
	 */
	if (vol->dev_major != 0) {
		if (is_local_volume(vol)) {
			rc = ensure_dev_node(vol->name, vol->dev_major, vol->dev_minor);

		} else {
			rc = remote_ensure_dev_node(&current_nodeid, vol->name, vol->dev_major, vol->dev_minor);
		}

		if (rc == 0) {
			memcpy(vol->dev_node, vol->name, sizeof(vol->name));
                        vol->flags &= ~VOLFLAG_NEEDS_ACTIVATE;

			if (vol->object->flags & SOFLAG_ACTIVE) {
				vol->flags |= VOLFLAG_ACTIVE;
			}

		} else {
			memset(vol->dev_node, '\0', sizeof(vol->dev_node));
		}

	} else {
                rc = EINVAL;
        }

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int get_volume_dev_node_and_fsim(logical_volume_t * vol) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/* Safety check */
	if (vol == NULL) {
		LOG_WARNING("Volume pointer is NULL.\n");
		LOG_PROC_EXIT_INT(EINVAL);
		return EINVAL;
	}

	if (vol->dev_major != 0) {
		if (make_volume_dev_node(vol) == 0) {

			if (vol->object->flags & SOFLAG_ACTIVE) {
				vol->flags |= VOLFLAG_ACTIVE;

				find_fsim_for_volume(vol);

				if (is_volume_mounted(vol)) {
					LOG_DEBUG("Volume \"%s\" is mounted on %s.\n", vol->name, vol->mount_point);
				}
			}
		}
	}

	if (!(vol->flags & VOLFLAG_ACTIVE)) {
		vol->flags |= VOLFLAG_NEEDS_ACTIVATE;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/******************************************************************************
 *
 * Functions for discovering EVMS objects
 *
 ******************************************************************************/

/*
 * A structure for grouping EVMS objects by volume.
 */
typedef struct volume_objects {
	u_int64_t               volume_serial;
	dlist_t                 objects;
	struct volume_objects * next;
	char                    volume_name[EVMS_VOLUME_NAME_SIZE];
	u_int32_t               flags;
} volume_objects_t;


/*
 * The structure of the parameters passed to extract_plugin_object() in its
 * Parameters pointer.
 */
typedef struct extract_object_parms_s {
	plugin_id_t id;
	dlist_t     plugin_object_list;
} extract_object_parms_t;


/*
 * If this object is managed by the plug-in specified in the parameters, then
 * put it into the plugin_object_list and return TRUE to remove it from the
 * current list.
 * This function has its parameters structured so that it can be called by
 * PruneList.
 */
static boolean extract_plugin_object(ADDRESS   Object,
				     TAG       ObjectTag,
				     ADDRESS   ObjectHandle,
				     ADDRESS   Parameters,
				     boolean * FreeMemory,
				     uint    * Error) {
	boolean result = FALSE;
	storage_object_t * object = (storage_object_t *) Object;
	extract_object_parms_t * parms = (extract_object_parms_t *) Parameters;

	LOG_PROC_ENTRY();

	/* Never free the object memory. */

	*FreeMemory = FALSE;

	/*
	 * If this object has the same plug-in ID as the one we're looking for
	 * (specified in the parms) then add it to the list specified in the parms
	 * and return TRUE so that it is removed from the main list that is being
	 * traversed.  Else, don't put the object in the list and return FALSE so
	 * that the object remains in the main list.
	 */

	if (object->feature_header->feature_id == parms->id) {
		ADDRESS handle;

		*Error = InsertObject(parms->plugin_object_list,
				      Object,
				      EVMS_OBJECT_TAG,
				      NULL,
				      AppendToList,
				      FALSE,
				      &handle);

		result = TRUE;

	} else {
		*Error = DLIST_SUCCESS;
	}

	LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
	return result;
}


/*
 * Convert a feature header in disk endian (little endian) format to the CPU
 * endian format.
 */
void inline feature_header_disk_to_cpu(evms_feature_header_t * fh) {

	fh->signature                 = DISK_TO_CPU32(fh->signature);
	fh->version.major             = DISK_TO_CPU32(fh->version.major);
	fh->version.minor             = DISK_TO_CPU32(fh->version.minor);
	fh->version.patchlevel        = DISK_TO_CPU32(fh->version.patchlevel);
	fh->engine_version.major      = DISK_TO_CPU32(fh->engine_version.major);
	fh->engine_version.minor      = DISK_TO_CPU32(fh->engine_version.minor);
	fh->engine_version.patchlevel = DISK_TO_CPU32(fh->engine_version.patchlevel);
	fh->flags                     = DISK_TO_CPU32(fh->flags);
	fh->feature_id                = DISK_TO_CPU32(fh->feature_id);
	fh->sequence_number           = DISK_TO_CPU64(fh->sequence_number);
	fh->alignment_padding         = DISK_TO_CPU64(fh->alignment_padding);
	fh->feature_data1_start_lsn   = DISK_TO_CPU64(fh->feature_data1_start_lsn);
	fh->feature_data1_size        = DISK_TO_CPU64(fh->feature_data1_size);
	fh->feature_data2_start_lsn   = DISK_TO_CPU64(fh->feature_data2_start_lsn);
	fh->feature_data2_size        = DISK_TO_CPU64(fh->feature_data2_size);
	fh->volume_serial_number      = DISK_TO_CPU64(fh->volume_serial_number);
	fh->volume_system_id          = DISK_TO_CPU32(fh->volume_system_id);
	fh->object_depth              = DISK_TO_CPU32(fh->object_depth);
}


/*
 * Check for a valid feature header -- has a signature and the CRC is correct.
 *
 * Returns:
 * ENOENT    if no signature
 * EINVAL    if has signature but CRC is bad
 *           or version is bad
 */
static int validate_feature_header(evms_feature_header_t * fh) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (DISK_TO_CPU32(fh->signature) == EVMS_FEATURE_HEADER_SIGNATURE) {
		u_int32_t old_crc;
		u_int32_t new_crc;

		old_crc = DISK_TO_CPU32(fh->crc);
		fh->crc = 0;
		new_crc = calculate_CRC(0xFFFFFFFFL, fh, sizeof(evms_feature_header_t));

		if ((new_crc == old_crc) || (old_crc == EVMS_MAGIC_CRC)) {

			/* Convert the feature header to CPU endian format. */
			feature_header_disk_to_cpu(fh);

			/*
			 * Make sure the version of the feature header is one that
			 * we can understand.
			 */
			if (fh->version.major == EVMS_FEATURE_HEADER_MAJOR) {
				if (fh->version.minor <= EVMS_FEATURE_HEADER_MINOR) {
					if (fh->version.minor == EVMS_FEATURE_HEADER_MINOR) {
						if (fh->version.patchlevel <= EVMS_FEATURE_HEADER_PATCHLEVEL) {
							/* Feature header looks OK. */

						} else {
							LOG_WARNING("Feature header is version %d.%d.%d.  The Engine handles version %d.%d.%d or compatible.\n", fh->version.major, fh->version.minor, fh->version.patchlevel, EVMS_FEATURE_HEADER_MAJOR, EVMS_FEATURE_HEADER_MINOR, EVMS_FEATURE_HEADER_PATCHLEVEL);
							rc = EINVAL;
						}
					}

				} else {
					LOG_WARNING("Feature header is version %d.%d.%d.  The Engine handles version %d.%d.%d or compatible.\n", fh->version.major, fh->version.minor, fh->version.patchlevel, EVMS_FEATURE_HEADER_MAJOR, EVMS_FEATURE_HEADER_MINOR, EVMS_FEATURE_HEADER_PATCHLEVEL);
					rc = EINVAL;
				}

			} else {
				LOG_WARNING("Feature header is version %d.%d.%d.  The Engine handles version %d.%d.%d or compatible.\n", fh->version.major, fh->version.minor, fh->version.patchlevel, EVMS_FEATURE_HEADER_MAJOR, EVMS_FEATURE_HEADER_MINOR, EVMS_FEATURE_HEADER_PATCHLEVEL);
				rc = EINVAL;
			}

		} else {
			LOG_DEBUG("Bad CRC. old(%#x) new(%#x)\n", old_crc, new_crc);
			rc = EINVAL;
		}

	} else {
		LOG_DEBUG("Sector does not have a feature header signature.\n");
		rc = ENOENT;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Allocate a buffers and read in the primary and secondary feature headers.
 * Check for a valid feature header -- has a signature and the CRC is
 * correct.
 *
 * Returns:
 * non-0    something really bad happened, e.g, memory allocation failure.
 *
 * For the primary and secondary return codes and feature header pointers:
 *
 * *rc      *feature_header
 * ---      -----------------
 * 0        pointer to buffer   if good feature header
 * ENOENT   NULL                if no signature
 * EINVAL   NULL                if has signature but CRC is bad
 *                              or version is bad
 * error    NULL                if any other error occurs
 */
static int read_feature_header(storage_object_t        * object,
			       evms_feature_header_t * * primary_feature_header,
			       int                     * primary_rc,
			       evms_feature_header_t * * secondary_feature_header,
			       int                     * secondary_rc) {

	int rc = 0;
	evms_feature_header_t * primary_fh = NULL;
	evms_feature_header_t * secondary_fh = NULL;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Look for feature headers on object %s.\n", object->name);

	secondary_fh = engine_alloc(sizeof(evms_feature_header_t) * 2);
	primary_fh = engine_alloc(sizeof(evms_feature_header_t));

	if ((primary_fh != NULL) &&
	    (secondary_fh != NULL)) {
		rc = object->plugin->functions.plugin->read(object,
							    object->size - (EVMS_FEATURE_HEADER_SECTORS * 2),
							    EVMS_FEATURE_HEADER_SECTORS * 2,
							    secondary_fh);
		if (rc == 0) {
			/*
			 * The sectors for both potential feature headers were read in
			 * successfully.  Copy the memory from the primary feature
			 * header location to the primary_fh buffer.
			 */

			memcpy(primary_fh, secondary_fh + 1, sizeof(evms_feature_header_t));

			/* Validate the feature headers. */
			*primary_rc = validate_feature_header(primary_fh);
			*secondary_rc = validate_feature_header(secondary_fh);

		} else {
			/*
			 * An error occurred trying to read both feature headers.
			 * Try to read them separately.
			 */

			*primary_rc = object->plugin->functions.plugin->read(object,
									     object->size - EVMS_FEATURE_HEADER_SECTORS,
									     EVMS_FEATURE_HEADER_SECTORS,
									     primary_fh);
			if (*primary_rc == 0) {
				/* Validate the primary feature header. */
				*primary_rc = validate_feature_header(primary_fh);
			}

			*secondary_rc = object->plugin->functions.plugin->read(object,
									       object->size - (EVMS_FEATURE_HEADER_SECTORS * 2),
									       EVMS_FEATURE_HEADER_SECTORS,
									       secondary_fh);
			if (*secondary_rc == 0) {
				/* Validate the secondary feature header. */
				*secondary_rc = validate_feature_header(secondary_fh);
			}
		}

		/*
		 * If either of the feature headers had and error on reading or on
		 * validation, free the feature header memory; it's useless.
		 */
		if (*primary_rc != 0) {
			engine_free(primary_fh);
			primary_fh = NULL;
		}
		if (*secondary_rc != 0) {
			engine_free(secondary_fh);
			secondary_fh = NULL;
		}

	} else {
		LOG_CRITICAL("Error allocating memory to read in a feature header.\n");
		if (primary_fh != NULL) {
			engine_free(primary_fh);
			primary_fh = NULL;
		}
		if (secondary_fh != NULL) {
			engine_free(secondary_fh);
			secondary_fh = NULL;
		}
		rc = ENOMEM;
	}

	*primary_feature_header   = primary_fh;
	*secondary_feature_header = secondary_fh;

	/* Any over all error code is returned as individual error codes as well. */
	if (rc != 0) {
		*primary_rc   = rc;
		*secondary_rc = rc;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Read in the feature header at the top of the object, if there is one, and
 * put a pointer to it into the storage_object_t.
 * This function has its parameters structured so that it can be called by
 * ForEachItem to process a list of objects.
 */
static int get_feature_header(ADDRESS Object,
			      TAG     ObjectTag,
			      ADDRESS ObjectHandle,
			      ADDRESS Parameters) {

	int secondary_rc = 0;
	int primary_rc = 0;
	storage_object_t * object = (storage_object_t *) Object;
	evms_feature_header_t * primary_fh = NULL;
	evms_feature_header_t * secondary_fh = NULL;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Examining object %s for feature headers.\n", object->name);

	/* If the object already has a feature header, then we're finished. */
	if (object->feature_header == NULL) {

		/* Read in the feature headers at the back of the object. */
		read_feature_header(object, &primary_fh, &primary_rc, &secondary_fh, &secondary_rc);

		switch (primary_rc) {
		case 0:
			LOG_DEBUG("Primary feature header read successfully.\n");

			switch (secondary_rc) {
			case 0:
				LOG_DEBUG("Secondary feature header read successfully.\n");

				/*
				 * Both feature headers were read in successfully.
				 * Check the sequence numbers.
				 */
				if (secondary_fh->sequence_number == primary_fh->sequence_number) {

					/*
					 * The sequence numbers match.  Life is good.  Put
					 * the primary copy (arbitrary choice) of the
					 * feature header into the storage object and free
					 * the secondary copy.
					 */
					LOG_DEBUG("Primary and secondary feature headers match.\n");
					object->feature_header = primary_fh;
					engine_free(secondary_fh);

				} else {
					/*
					 * The sequence numbers don't match.  Use the
					 * feature header with the highest sequence number,
					 * throw out the other one, and mark the object's
					 * feature header dirty so that the feature headers
					 * will be rewritten to disk.
					 */
					if (primary_fh->sequence_number > secondary_fh->sequence_number) {
						LOG_WARNING("Primary feature header for object %s has a higher sequence number (%"PRIu64") than the secondary feature header (%"PRIu64").  Using the primary copy.\n", primary_fh->object_name, primary_fh->sequence_number, secondary_fh->sequence_number);
						object->feature_header = primary_fh;
						engine_free(secondary_fh);

					} else {
						LOG_WARNING("Secondary feature header for object %s has a higher sequence number (%"PRIu64") than the primary feature header (%"PRIu64").  Using the secondary copy.\n", secondary_fh->object_name, secondary_fh->sequence_number, primary_fh->sequence_number);
						object->feature_header = secondary_fh;
						engine_free(primary_fh);
					}

					object->flags |= SOFLAG_FEATURE_HEADER_DIRTY;
				}
				break;

			case ENOENT:
			case EINVAL:
				/*
				 * The primary feature header was found, but the
				 * secondary feature header either had a bad signature
				 * or a bad CRC.  Keep the primary feature header.
				 * Assume that the secondary feature header is corrupt
				 * and needs to be rewritten.
				 */
				LOG_WARNING("Secondary feature header not found on object %s.  Using the primary copy.\n", object->name);
				object->feature_header = primary_fh;
				object->flags |= SOFLAG_FEATURE_HEADER_DIRTY;
				break;

			default:
				/*
				 * We got the primary feature header OK but some
				 * unexpected error occurred when reading the secondary
				 * feature header.  If we get an unexpected error we
				 * don't use any of the feature headers.
				 */
				engine_user_message(NULL, NULL, "Error code %d (%s) when reading the secondary copy of feature header on object %s.  Ignoring good primary feature header.\n", secondary_rc, strerror(secondary_rc), object->name);
				engine_free(primary_fh);
				break;
			}
			break;

		case ENOENT:
		case EINVAL:
			switch (secondary_rc) {
			case 0:
				/*
				 * The primary feature header either had a bad
				 * signature or a bad CRC.  The secondary feature header
				 * was read successfully.  Keep the secondary feature
				 * header.  Assume the primary feature header is
				 * corrupt and needs to be rewritten.
				 */
				LOG_WARNING("Primary feature header not found on object %s.\n", object->name);
				LOG_WARNING("Secondary feature header read successfully.  Using the secondary copy.\n");
				object->feature_header = secondary_fh;
				object->flags |= SOFLAG_FEATURE_HEADER_DIRTY;
				break;

			case ENOENT:
			case EINVAL:
				/*
				 * Both the primary and secondary feature headers either
				 * had a bad signature or a bad CRC.  Assume that there
				 * are no feature headers on this object.
				 */
				LOG_DEBUG("Primary feature header not found on object %s.\n", object->name);
				LOG_DEBUG("Secondary feature header not found on object %s.\n", object->name);
				break;

			default:
				/*
				 * The primary feature header either has a bad
				 * signature or the CRC is bad.  Regardless, we got an
				 * unexpected error code when reading the secondary
				 * feature header.  If we get an unexpected error we
				 * don't use any of the feature headers.
				 */
				LOG_WARNING("Primary feature header not found on object %s.\n", object->name);
				engine_user_message(NULL, NULL, "Error code %d (%s) when reading the secondary copy of feature header on object %s.\n", secondary_rc, strerror(secondary_rc), object->name);
				break;
			}
			break;

		default:
			/*
			 * We got an unexpected error when reading the primary feature
			 * header.  If we get an unexpected error we don't use any of
			 * the feature headers.
			 */
			engine_user_message(NULL, NULL, "Error code %d (%s) when reading the primary copy of feature header on object %s.\n", primary_rc, strerror(primary_rc), object->name);

			switch (secondary_rc) {
			case 0:
				engine_user_message(NULL, NULL, "Ignoring good secondary feature header.\n");
				engine_free(secondary_fh);
				break;

			case ENOENT:
			case EINVAL:
				LOG_WARNING("Secondary feature header not found.\n");
				break;

			default:
				/*
				 * We got an unexpected error when reading the
				 * secondary feature header.
				 */
				engine_user_message(NULL, NULL, "Error code %d (%s) when reading the secondary copy of feature header on object %s.\n", secondary_rc, strerror(secondary_rc), object->name);
				break;
			}
			break;
		}
	}

	/*
	 * Always return a good error code.  This function is called by a
	 * ForEachItem().  An error code would cause the ForEachItem() to abort.
	 * We want to process all of the objects in the list.
	 */
	LOG_PROC_EXIT_INT(DLIST_SUCCESS);
	return DLIST_SUCCESS;
}


static int get_greatest_object_depth(ADDRESS object,
				     TAG     object_tag,
				     ADDRESS object_handle,
				     ADDRESS parameters) {

	storage_object_t * obj = (storage_object_t *) object;
	u_int32_t * object_depth = (u_int32_t *) parameters;

	LOG_PROC_ENTRY();

	if (obj->feature_header != NULL) {
		if (obj->feature_header->object_depth > *object_depth) {
			*object_depth = obj->feature_header->object_depth;
		}
	}

	LOG_PROC_EXIT_INT(DLIST_SUCCESS);
	return DLIST_SUCCESS;
}


static int discover_objects_by_plugin(dlist_t object_list, dlist_t resulting_object_list) {

	int rc = 0;
	int status_count = 0;
	dlist_t plugin_object_list = CreateList();
	dlist_t new_object_list = CreateList();

	LOG_PROC_ENTRY();

	if ((plugin_object_list != NULL) &&
	    (new_object_list != NULL)) {
		/*
		 * Loop through the object list finding groups of objects that have the
		 * same top level feature.  Extract the objects with the same top level
		 * feature, put them into a separate list, and pass the list to the
		 * feature for discovery.
		 */
		while (!ListEmpty(object_list) && (rc == 0)) {
			TAG                tag;
			storage_object_t * obj;

			rc = BlindExtractObject(object_list,
						&tag,
						NULL,
						(ADDRESS *) &obj);

			if ((rc == DLIST_SUCCESS) && (obj != NULL)) {
				plugin_record_t * feature;

				/*
				 * If the object has a feature header, start the discovery
				 * process on the object.
				 */
				if (obj->feature_header != NULL) {
					rc = engine_get_plugin_by_ID(obj->feature_header->feature_id, &feature);

					if (rc == 0) {
						/*
						 * Skip associative class features.  They are handled
						 * later after the volumes have been discovered.
						 */

						if (GetPluginType(feature->id) != EVMS_ASSOCIATIVE_FEATURE) {
							ADDRESS trash;

							/* Clear out the plugin_object_list */
							DeleteAllItems(plugin_object_list, FALSE);

							/*
							 * Put the extracted object on the list of objects
							 * for the plug-in.
							 */
							rc = InsertObject(plugin_object_list,
									  obj,
									  tag,
									  NULL,
									  AppendToList,
									  FALSE,
									  &trash);

							if (rc == DLIST_SUCCESS) {
								extract_object_parms_t parms;

								/*
								 * Clear the new_object_list to receive the
								 * new objects from the next call to a
								 * plug-in's discover();
								 */
								DeleteAllItems(new_object_list, FALSE);

								/*
								 * Pull the other objects that have the same
								 * feature ID off the list and add them to the
								 * plugin_object_list.
								 */
								parms.id = feature->id;
								parms.plugin_object_list = plugin_object_list;

								rc = PruneList(object_list,
									       extract_plugin_object,
									       (ADDRESS) &parms);

								if (rc == DLIST_SUCCESS) {

									status_message("Discovering EVMS objects%s\n", status_dots(&status_count));
									rc = feature->functions.plugin->discover(plugin_object_list, new_object_list, TRUE);

									/*
									 * Remove any corrupt objects from the
									 * list.
									 */
									PruneList(new_object_list,
										  remove_corrupt_object,
										  NULL);

									/* Make sure we have a dev node for each object. */
									ForEachItem(new_object_list,
										    make_object_dev_node,
										    NULL,
										    TRUE);

									propigate_cluster_info(new_object_list);

									/*
									 * If discover() succeeded, put the new
									 * objects on the resulting_object_list.
									 */
									if (rc == 0) {
										CopyList(resulting_object_list,
											 new_object_list,
											 AppendToList);
									}
								}

								/*
								 * We have handled any errors.  Clear the
								 * error code so that the loop will not
								 * terminate.
								 */
								rc = 0;
							}

						} else {
							/*
							 * We hit an associative class feature.  Build a
							 * temporary volume structure to hold the volume
							 * information from the feature header structure and
							 * hang it off of the object so that we will know
							 * how to create the real volume when the
							 * associative feature discovers its object.
							 */
							logical_volume_t * volume = (logical_volume_t *) engine_alloc(sizeof(logical_volume_t));

							if (volume != NULL) {

								volume->serial_number = obj->feature_header->volume_serial_number;
								strcpy(volume->name, obj->feature_header->volume_name);

								/* (Ab)Use the consuming_private_data field in
								 * the object to hold the temporary volume.
								 * If we put it in the volume field, later code
								 * won't be able to find this object as a top
								 * object since it will think it is part of a
								 * volume.
								 */
								obj->consuming_private_data = volume;

							} else {
								LOG_CRITICAL("Error allocating memory for a logical volume structure.\n");
							}
						}

					} else {
						LOG_WARNING("Error code %d when trying to get plug-in for feature ID %d (%#x) from feature header in object %s.\n", rc, obj->feature_header->feature_id, obj->feature_header->feature_id, obj->name);
					}

				} else {
					LOG_WARNING("Object %s does not contain a feature header.\n", obj->name);
				}
			}
		}

	} else {
		/* Failed to create one or more lists. */
		if (plugin_object_list == NULL) {
			LOG_CRITICAL("Error allocating memory for the plugin_object_list.\n");
		}
		if (new_object_list == NULL) {
			LOG_CRITICAL("Error allocating memory for the new_object_list.\n");
		}
		rc = ENOMEM;
	}

	if (plugin_object_list != NULL)	DestroyList(&plugin_object_list, FALSE);
	if (new_object_list != NULL) DestroyList(&new_object_list, FALSE);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


typedef struct extract_by_depth_parms_s {
	dlist_t     result_list;
	u_int32_t   depth;
} extract_by_depth_parms_t;


/*
 * Check if the feature header hanging off an object has the specified depth.
 * If so, put the object on the result_list and return TRUE so that the
 * object will be removed from the current list.  Else, return FALSE so that
 * the object stays on the list.
 * This function has its parameters structured so that it can be called by the
 * PruneList() dlist processor on a list of objects.
 */
static boolean extract_object_by_depth(ADDRESS   object,
				       TAG       object_tag,
				       ADDRESS   object_handle,
				       ADDRESS   parameters,
				       boolean * free_memory,
				       uint    * error) {
	boolean result = FALSE;
	storage_object_t * obj = (storage_object_t *) object;
	extract_by_depth_parms_t * parms = (extract_by_depth_parms_t *) parameters;

	LOG_PROC_ENTRY();

	/* Never free item memory. */
	*free_memory = FALSE;

	if (obj->feature_header != NULL) {
		if (obj->feature_header->object_depth == parms->depth) {
			ADDRESS trash;

			*error = InsertObject(parms->result_list,
					      object,
					      get_tag_for_object(obj),
					      NULL,
					      AppendToList,
					      FALSE,
					      &trash);
			result = TRUE;
		}

	} else {
		LOG_WARNING("Object %s does not have a feature header.  It is being removed from the list.\n", obj->name);
		result = TRUE;
	}

	LOG_PROC_EXIT_BOOLEAN_INT(result, *error);
	return result;

}


/*
 * Given a VolObj list, do the discovery of the objects that make up the volume.
 */
static void discover_objects_for_volume(volume_objects_t * pVolObj) {

	int rc = 0;
	extract_by_depth_parms_t extract_parms;
	dlist_t                  object_list = CreateList();

	LOG_PROC_ENTRY();

	extract_parms.depth = 0;
	extract_parms.result_list = object_list;

	if (object_list != NULL) {

		/* Find the greatest object_depth of the objects in the list. */
		ForEachItem(pVolObj->objects,
			    get_greatest_object_depth,
			    &extract_parms.depth,
			    TRUE);

		while ((rc == DLIST_SUCCESS) && (extract_parms.depth > 0)) {

			/*
			 * Get the feature header for the top level feature contained
			 * within each object on the list.  The feature header is hung off a
			 * pointer in the storage object.  We ignore errors that may occur
			 * during the reading of feature headers.  If the error was bad
			 * enough, then there won't be a feature header and thus won't be
			 * anything to discover on that object.
			 */
			ForEachItem(pVolObj->objects,
				    get_feature_header,
				    NULL,
				    TRUE);

			/*
			 * Clear out the working object_list for this iteration through
			 * the loop.
			 */
			DeleteAllItems(object_list, FALSE);

			/*
			 * Build a list of objects that have the same object_depth
			 * in the feature header.  The objects are extracted from the
			 * pVolObj->objects list.
			 */
			rc = PruneList(pVolObj->objects,
				       extract_object_by_depth,
				       &extract_parms);

			if (rc == DLIST_SUCCESS) {

				/*
				 * Given the list of objects with the same object_depth,
				 * group the objects by plug-in and pass each group to its
				 * plug-in for discovery.  Objects resulting from the
				 * discovery are put back on the pVolObj->objects list for
				 * the next pass through this loop.
				 */
				rc = discover_objects_by_plugin(object_list, pVolObj->objects);

				if (rc == 0) {
					/* Move to the next depth. */
					extract_parms.depth--;
				}
			}
		}

		if (rc == DLIST_SUCCESS) {

			uint list_size = 0;

			/*
			 * If there is only one item in the list, then it is a candidate for
			 * becoming a volume.
			 */
			GetListSize(pVolObj->objects, &list_size);

			if (list_size == 1) {
				storage_object_t * obj;

				rc = GetObject(pVolObj->objects,
					       EVMS_OBJECT_TAG,
					       NULL,
					       FALSE,
					       (ADDRESS *) &obj);

				if (rc == DLIST_SUCCESS) {

					/*
					 * A flag in the feature header (flags were copied to
					 * the volume_objects_t structure) says whether this
					 * object should be made into a volume or left as an
					 * object.
					 */
					if (!(pVolObj->flags & EVMS_VOLUME_DATA_OBJECT)) {
						rc = make_evms_volume_for_object(obj,
										 pVolObj->volume_name,
										 pVolObj->volume_serial);
						get_volume_dev_node_and_fsim(obj->volume);
					}
				}
			}
		}

		DestroyList(&object_list, FALSE);

	} else {
		/* Error allocating memory for the lists. */
		if (object_list == NULL) {
			LOG_CRITICAL("Error getting memory for an object list.\n");
		} else {
			DestroyList(&object_list, FALSE);
		}
		rc = ENOMEM;
	}

	LOG_PROC_EXIT_VOID();
}


/*
 * Put an object onto a VolObj list (a list of objects which volume comprises).
 * If a VolObj list has not yet been made for the volume, make one.
 */
static int add_object_to_VolObj_list(storage_object_t * obj, volume_objects_t * * pVolObjList) {
	int rc = 0;
	volume_objects_t * pVolObj;

	LOG_PROC_ENTRY();

	pVolObj = *pVolObjList;
	while ((pVolObj != NULL) && (pVolObj->volume_serial != obj->feature_header->volume_serial_number)) {
		pVolObj = pVolObj->next;
	}

	if (pVolObj == NULL) {
		LOG_DEBUG("Create new VolObj list for volume %s.\n", obj->feature_header->volume_name);
		pVolObj = (volume_objects_t *) engine_alloc(sizeof(volume_objects_t));

		if (pVolObj != NULL) {
			pVolObj->flags = obj->feature_header->flags;
			pVolObj->volume_serial = obj->feature_header->volume_serial_number;
			memcpy(pVolObj->volume_name, obj->feature_header->volume_name, sizeof(pVolObj->volume_name));
			pVolObj->objects = CreateList();

			pVolObj->next = *pVolObjList;
			*pVolObjList = pVolObj;

		} else {
			rc = ENOMEM;
			LOG_CRITICAL("Error when allocating memory for a volume_object structure.\n");
		}
	}

	if (rc == 0) {
		ADDRESS handle;

		LOG_DEBUG("Add object %s to volume object list for volume %s.\n", obj->name, pVolObj->volume_name);
		rc = InsertObject(pVolObj->objects,
				  obj,
				  get_tag_for_object(obj),
				  NULL,
				  AppendToList,
				  TRUE,
				  &handle);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Initial processing of objects.
 * Objects with stop data have been removed from the list.  Non-data type
 * objects have been removed from the list.
 * If the object does not have a feature header, make it into a compatibility
 * volume.
 * If the object has a feature header on it, check if it is the special
 * feature header that is put on objects that don't have features but are made
 * into EVMS volumes.  If so, make the object into an EVMS volume.
 * Otherwise, the object has EVMS features on it.  Put the object into a list
 * of object that are grouped by volume.
 */
static int process_object(ADDRESS object,
			  TAG     object_tag,
			  ADDRESS object_handle,
			  ADDRESS parameters) {
	int rc = 0;
	storage_object_t * obj = (storage_object_t *) object;
	volume_objects_t * * pVolObjList = (volume_objects_t * *) parameters;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Processing object %s.\n", obj->name);

	/*
	 * If the object does not have a feature header, then make it into a
	 * compatibility volume.
	 */
	if (obj->feature_header == NULL) {
		make_compatibility_volume_for_object(obj);
		get_volume_dev_node_and_fsim(obj->volume);

	} else {
		/*
		 * The object has a feature header.  If the feature header has an
		 * object_depth of 1 and a feature_id of 0, then it is a feature
		 * header for making an EVMS volume.
		 */

		if ((obj->feature_header->object_depth == 1) &&
		    (obj->feature_header->feature_id == EVMS_VOLUME_FEATURE_ID)) {

			if (!(obj->feature_header->flags & EVMS_VOLUME_DATA_OBJECT)) {
				/* Don't let errors stop the discovery process. */
				make_evms_volume_for_object(obj,
							    obj->feature_header->volume_name,
							    obj->feature_header->volume_serial_number);
				get_volume_dev_node_and_fsim(obj->volume);
			}

		} else {
			/*
			 * This object has feature(s) on it.  Put it into a
			 * VolObjList.
			 */
			rc = add_object_to_VolObj_list(obj, pVolObjList);

			if (rc != 0) {
				LOG_CRITICAL("Error when putting object %s into the VolObj list.  Return code was %d.\n", obj->name, rc);
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Check if an object has the data_type of DATA_TYPE.  If so, return FALSE to
 * leave it on the list.  If not, return TRUE so that it will be removed from
 * the list.
 * This function has its parameters structured so that it can be called by the
 * PruneList() dlist processor on a list of objects.
 */
boolean remove_non_data_object(ADDRESS   object,
			       TAG       object_tag,
			       ADDRESS   object_handle,
			       ADDRESS   parameters,
			       boolean * free_memory,
			       uint    * error) {
	boolean result = FALSE;
	int rc = 0;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Object %s has a data_type of %d.\n", obj->name, obj->data_type);

	/* Never free the item memory. */
	*free_memory = FALSE;

	/*
	 * If the object is a free space object, return TRUE to remove it from this
	 * list.
	 */
	if (obj->data_type == FREE_SPACE_TYPE) {
		result = TRUE;

	} else {
		/*
		 * If the object is not a data object (e.g., it is a metadata object)
		 * return TRUE to remove it from this list.
		 */
		if (obj->data_type != DATA_TYPE) {
			result = TRUE;
		}
	}

	LOG_PROC_EXIT_BOOLEAN_INT(result, rc);
	*error = rc;
	return result;
}


/*
 * Check if an object has stop data on it.  If so, return TRUE so that it will
 * be removed from its current list.  Else, return FALSE so that the object
 * stays in the list.
 * This function has its parameters structured so that it can be called by the
 * PruneList() dlist processor on a list of objects.
 */
static evms_feature_header_t stop_data;

static boolean remove_stop_data_object(ADDRESS   object,
				       TAG       object_tag,
				       ADDRESS   object_handle,
				       ADDRESS   parameters,
				       boolean * free_memory,
				       uint    * error) {
	boolean result = FALSE;
	int rc = 0;
	storage_object_t * obj = (storage_object_t *) object;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Request to remove object %s if it has stop data.\n", obj->name);

	/* Never free the item memory. */
	*free_memory = FALSE;

	/* Default, if anything fails, is that there are no stop data. */
	obj->flags &= ~SOFLAG_HAS_STOP_DATA;

	/* Only data objects can have stop data. */
	if (obj->data_type == DATA_TYPE) {

		/* Check for stop data at the end of the object. */
		rc = obj->plugin->functions.plugin->read(obj, obj->size - 1, 1, &stop_data);

		if (rc == 0) {
			if (validate_feature_header(&stop_data) == 0) {
				/*
				 * validate_feature_header() converts the in-memory
				 * feature header to CPU endian format.
				 */
				if (stop_data.flags & EVMS_VOLUME_DATA_STOP) {
					LOG_DEBUG("Object %s has stop data on it.\n", obj->name);
					obj->flags |= SOFLAG_HAS_STOP_DATA;
					result = TRUE;
				}
			}
		}
	}

	/* Don't allow an error to kill the processing of the list. */
	*error = DLIST_SUCCESS;

	LOG_PROC_EXIT_BOOLEAN_INT(result, rc);
	return result;
}


int discover_evms_objects(dlist_t object_list) {
	int rc = 0;

	LOG_PROC_ENTRY();

	/*
	 * Pull any metadata objects off the list.
	 * Pull any free space objects off the object_list.
	 */
	rc = PruneList(object_list,
		       remove_non_data_object,
		       NULL);

	if (rc == DLIST_SUCCESS) {

		/* Remove from the list any objects with stop data on them. */
		rc = PruneList(object_list,
			       remove_stop_data_object,
			       NULL);

		if (rc == DLIST_SUCCESS) {
			volume_objects_t * VolObjList = NULL;


			/*
			 * Get the feature headers for the top level feature contained
			 * within each object on the list.  The feature header if found,
			 * is hung off a pointer in the storage object.
			 */
			ForEachItem(object_list,
				    get_feature_header,
				    NULL,
				    TRUE);

			/* Process the objects on the object_list. */
			rc = ForEachItem(object_list,
					 process_object,
					 &VolObjList,
					 TRUE);

			if (rc == 0) {

				/*
				 * Do the feature discovery for each group of objects
				 * that belong to a volume.
				 */
				volume_objects_t * pVolObj;

				for (pVolObj = VolObjList; ((rc == 0) && (pVolObj != NULL)); pVolObj = pVolObj->next) {
					discover_objects_for_volume(pVolObj);
				}

				/* Clean up our lists. */
				while (VolObjList != NULL) {
					pVolObj = VolObjList;
					VolObjList = pVolObj->next;

					DestroyList(&(pVolObj->objects), FALSE);
					engine_free(pVolObj);
				}
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static boolean extract_associative_object(ADDRESS   Object,
					  TAG       ObjectTag,
					  ADDRESS   ObjectHandle,
					  ADDRESS   Parameters,
					  boolean * FreeMemory,
					  uint    * Error) {

	storage_object_t * obj = (storage_object_t *) Object;
	boolean result = FALSE;
	int rc = 0;
	int status_count = 0;
	logical_volume_t * volume;
	plugin_record_t * feature;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Examining object %s.\n", obj->name);

	*Error = 0;

	/* Never free the item memory. */
	*FreeMemory = FALSE;

	/*
	 * Check if a feature header even exists since this might be a topmost
	 * object that is not a volume.
	 */
	if (obj->feature_header == NULL) {
		LOG_DEBUG("Object does not have a feature header.\n");
		LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
		return result;
	}

	/*
	 * Filter out objects that have an associative class feature for a
	 * parent.
	 */
	if (GetPluginType(obj->feature_header->feature_id) != EVMS_ASSOCIATIVE_FEATURE) {
		LOG_DEBUG("Object's feature header is not for an associative class feature.\n");
		LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
		return result;
	}

	/*
	 * Extract the temporary volume structure that was attached to the
	 * object when it was discovered.
	 */

	volume = (logical_volume_t *) obj->consuming_private_data;

	if (volume == NULL) {
		LOG_DEBUG("Object has an associative class feature header, but does not have a volume structure attached.\n");
		LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
		return result;
	}

	obj->consuming_private_data = NULL;

	rc = engine_get_plugin_by_ID(obj->feature_header->feature_id, &feature);

	if (rc == 0) {

		/*
		 * discover() takes a dlist_t for input objects, so we have
		 * to build them, even though we know there is only one
		 * object for the discover().
		 */

		dlist_t object_list = CreateList();

		if (object_list != NULL) {
			ADDRESS handle;

			rc = InsertObject(object_list,
					  obj,
					  EVMS_OBJECT_TAG,
					  NULL,
					  AppendToList,
					  FALSE,
					  &handle);

			if (rc == DLIST_SUCCESS) {
				dlist_t new_object_list = CreateList();

				if (new_object_list != NULL) {
					status_message("Discovering EVMS associative objects%s\n", status_dots(&status_count));
					rc = feature->functions.plugin->discover(object_list, new_object_list, TRUE);

					LOG_DEBUG("Return code from %s discovery() was %d.\n", feature->short_name, rc);

					/* Remove any corrupt objects from the list. */
					PruneList(new_object_list,
						  remove_corrupt_object,
						  NULL);

					ForEachItem(new_object_list,
						    make_object_dev_node,
						    NULL,
						    TRUE);

					propigate_cluster_info(new_object_list);

					if (rc == 0) {
						storage_object_t * new_object;

						rc = GetObject(new_object_list,
							       EVMS_OBJECT_TAG,
							       NULL,
							       FALSE,
							       (ADDRESS *) &new_object);

						/*
						 * Make a volume for the new object if the
						 * feature header doesn't say to make it
						 * into an object.
						 */
						if ((rc == DLIST_SUCCESS) &&
						    (!(obj->feature_header->flags & EVMS_VOLUME_DATA_OBJECT))) {

							/* Make a volume for the object. */
							rc = make_evms_volume_for_object(new_object,
											 volume->name,
											 volume->serial_number);
							get_volume_dev_node_and_fsim(obj->volume);
						}

						/*
						 * Return TRUE so that this object will be
						 * removed from the list.
						 */
						result = TRUE;

						if ((rc == DLIST_END_OF_LIST) ||
						    (rc == DLIST_EMPTY)) {
							rc = DLIST_SUCCESS;
						}

					} else {
						/*
						 * If the associative feature failed to do
						 * its discovery, that's not a catastrophic
						 * error worth killing the Engine.  Clear
						 * the error code.
						 */
						if (rc == EVMS_FEATURE_FATAL_ERROR) {
							rc = 0;
						}
					}

					DestroyList(&new_object_list, FALSE);

				} else {
					LOG_CRITICAL("Error allocating memory for a resulting objects list.\n");
					rc = ENOMEM;
				}
			}

			DestroyList(&object_list, FALSE);

		} else {
			LOG_CRITICAL("Error allocating memory for an input objects list.\n");
			rc = ENOMEM;
		}

	} else {
		LOG_DEBUG("Unable to find the plug-in for feature ID %d.\n", obj->feature_header->feature_id);
	}

	/* Release the temporary volume structure. */
	engine_free(volume);

	LOG_PROC_EXIT_BOOLEAN_INT(result, rc);
	*Error = rc;
	return result;
}


int discover_associative_features(dlist_t object_list) {
	int rc = 0;

	LOG_PROC_ENTRY();

	/*
	 * Use PruneList to run the ObjectList checking for associative class
	 * features, calling their discover() functions, and making them into
	 * volumes if necessary.  If the object is made into a volume it will be
	 * removed from the ObjectList.
	 */

	rc = PruneList(object_list,
		       extract_associative_object,
		       NULL);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int compare_objects(ADDRESS   object1,
		    TAG       object1_tag,
		    ADDRESS   object2,
		    TAG       object2_tag,
		    uint    * error) {

	int result = 0;

	storage_object_t * obj1 = (storage_object_t *) object1;
	storage_object_t * obj2 = (storage_object_t *) object2;

	LOG_PROC_ENTRY();

	/* Validate parameters. */
	if ((object1_tag != DISK_TAG) &&
	    (object1_tag != SEGMENT_TAG) &&
	    (object1_tag != REGION_TAG) &&
	    (object1_tag != EVMS_OBJECT_TAG)) {
		*error = EINVAL;
		LOG_PROC_EXIT_INT(*error);
		return(result);
	}

	if ((object2_tag != DISK_TAG) &&
	    (object2_tag != SEGMENT_TAG) &&
	    (object2_tag != REGION_TAG) &&
	    (object2_tag != EVMS_OBJECT_TAG)) {
		*error = EINVAL;
		LOG_PROC_EXIT_INT(*error);
		return(result);
	}

	/* First sort by plug-in */
	result = strcmp(obj1->plugin->short_name, obj2->plugin->short_name);

	if (result == 0) {
		if (object1_tag == object2_tag) {
			switch (object1_tag) {
			case DISK_TAG:
			case REGION_TAG:
			case EVMS_OBJECT_TAG:
				/*
				 * If the objects are produced from containers,
				 * sort by the container name.
				 */
				if ((obj1->producing_container != NULL) &&
				    (obj2->producing_container != NULL)) {
					result = strcmp(obj1->producing_container->name,
							obj2->producing_container->name);
				}

				if (result == 0) {
					/* Sort by data type. */
					if (obj1->data_type != obj2->data_type) {
						switch (obj1->data_type) {
						case DATA_TYPE:
							/* Data objects come first. */
							result = -1;
							break;
						case FREE_SPACE_TYPE:
							/* Free space objects are second. */
							if (obj2->data_type == DATA_TYPE) {
								result = 1;
							} else {
								result = -1;
							}
							break;
						case META_DATA_TYPE:
							/* Metadata objects are last. */
							result = 1;
							break;
						default:
							break;
						}

					} else {
						/* Objects have the same data type. */
						/* Sort by name. */
						result = strcmp(obj1->name, obj2->name);
					}
				}
				break;

			case SEGMENT_TAG:
				{
					/* First, sort by disk. */
					int rc;
					TAG tag;
					storage_object_t * child1;
					storage_object_t * child2;

					rc = BlindGetObject(obj1->child_objects,
							    &tag,
							    NULL,
							    FALSE,
							    (ADDRESS *) &child1);

					if (rc == DLIST_SUCCESS) {
						rc = BlindGetObject(obj2->child_objects,
								    &tag,
								    NULL,
								    FALSE,
								    (ADDRESS *) &child2);

						if (rc == DLIST_SUCCESS) {
							result = strcmp(child1->name, child2->name);
						}

					}

					if (result == 0) {
						/* Segments are sorted by starting LBA. */
						if (obj1->start < obj2->start) {
							result = -1;
						} else if (obj1->start > obj2->start) {
							result = 1;
						} else {
							result = 0;
						}
					}
				}
				break;

			default:
				break;
			}

		} else {
			/*
			 * Objects are not of the same type.  This should not happen.
			 * Just in case, compare the object names.
			 */
			result = strcmp(obj1->name, obj2->name);
		}
	}

	*error = 0;

	LOG_PROC_EXIT_INT(result);
	return(result);
}


int compare_containers(ADDRESS   object1,
		       TAG       object1_tag,
		       ADDRESS   object2,
		       TAG       object2_tag,
		       uint    * error) {

	int result = 0;

	storage_container_t * con1 = (storage_container_t *) object1;
	storage_container_t * con2 = (storage_container_t *) object2;

	LOG_PROC_ENTRY();

	/* Validate parameters. */
	if (object1_tag != CONTAINER_TAG) {
		*error = EINVAL;
		LOG_PROC_EXIT_INT(*error);
		return(result);
	}

	if (object2_tag != CONTAINER_TAG) {
		*error = EINVAL;
		LOG_PROC_EXIT_INT(*error);
		return(result);
	}

	/* First sort by plug-in */
	result = strcmp(con1->plugin->short_name, con2->plugin->short_name);

	if (result == 0) {
		/* Next, sort by container name. */
		result = strcmp(con1->name, con2->name);
	}

	*error = 0;

	LOG_PROC_EXIT_INT(result);
	return(result);
}


int compare_volumes(ADDRESS   object1,
		    TAG       object1_tag,
		    ADDRESS   object2,
		    TAG       object2_tag,
		    uint    * error) {

	int result = 0;

	logical_volume_t * vol1 = (logical_volume_t *) object1;
	logical_volume_t * vol2 = (logical_volume_t *) object2;

	LOG_PROC_ENTRY();

	/* Validate parameters. */
	if (object1_tag != VOLUME_TAG) {
		*error = EINVAL;
		LOG_PROC_EXIT_INT(*error);
		return(result);
	}

	if (object2_tag != VOLUME_TAG) {
		*error = EINVAL;
		LOG_PROC_EXIT_INT(*error);
		return(result);
	}

	/* Sort by volume name. */
	result = strcmp(vol1->name, vol2->name);

	*error = 0;

	LOG_PROC_EXIT_INT(result);
	return(result);
}


int do_discovery(void) {

	int rc = 0;
	dlist_t object_list;
	struct timezone tz;
	struct timeval discover_start_time;
	struct timeval discover_stop_time;
	struct timeval discover_time;
	int discover_hours;
	int discover_minutes;

	gettimeofday(&discover_start_time, &tz);

	LOG_PROC_ENTRY();

	rc = make_directory(EVMS_OBJECT_NODE_DIR, NO_MESSAGES);

	if (rc == 0) {
		object_list = CreateList();
		if (object_list != NULL) {

			discover_in_progress = TRUE;

			if (local_focus) {
				rc = discover_logical_disks(object_list);

			} else {
				rc = discover_remote_logical_disks(&current_nodeid, object_list);
			}

			if (rc == 0) {

				SortList(DiskList, compare_objects);
				rc = discover_segments(object_list);
				if (rc == 0) {

					SortList(SegmentList, compare_objects);
					rc = discover_regions(object_list);
					if (rc == 0) {

						SortList(RegionList, compare_objects);
						rc = discover_evms_objects(object_list);
						if (rc == 0) {

							dlist_t top_object_list;

							rc = engine_get_object_list(0,
										    DATA_TYPE,
										    NULL,
										    NULL,
										    TOPMOST,
										    &top_object_list);

							if (rc == 0) {
								rc = discover_associative_features(top_object_list);
								if (rc == 0) {

									SortList(EVMSObjectList, compare_objects);
									SortList(ContainerList, compare_containers);
									SortList(VolumeList, compare_volumes);
								}

								DestroyList(&top_object_list, FALSE);
							}
						}
					}
				}
			}

			discover_in_progress = FALSE;

			DestroyList(&object_list, FALSE);

		} else {
			LOG_CRITICAL("Error creating the object_list.\n");
			rc = ENOMEM;
		}

	} else {
		engine_user_message(NULL,NULL,"Error creating the %s directory: %s\n", EVMS_OBJECT_NODE_DIR, strerror(rc));
	}

	/*
	 * Tell the Local Disk Manager to turn of its cache now that discovery
	 * is complete.
	 */
	if (local_disk_manager != NULL) {
		ldm_dpc_t dpc = {function:STOP_CACHE};
		local_disk_manager->functions.plugin->direct_plugin_communication(NULL, FALSE, &dpc);
	}

	gettimeofday(&discover_stop_time, &tz);

	discover_time.tv_sec = discover_stop_time.tv_sec - discover_start_time.tv_sec;
	discover_time.tv_usec = discover_stop_time.tv_usec - discover_start_time.tv_usec;
	if (discover_time.tv_usec < 0) {
		discover_time.tv_sec --;
		discover_time.tv_usec += 1000000;
	}

	discover_hours = discover_time.tv_sec / 3600;
	discover_time.tv_sec %= 3600;
	discover_minutes = discover_time.tv_sec / 60;
	discover_time.tv_sec %= 60;

	LOG_DEFAULT("Discovery time: %02d:%02d:%02ld.%06ld\n", discover_hours, discover_minutes, discover_time.tv_sec, discover_time.tv_usec);

	status_message("Discovery finished.\n");

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

