/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: expand.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "fullengine.h"
#include "expand.h"
#include "engine.h"
#include "handlemgr.h"
#include "common.h"
#include "commit.h"
#include "discover.h"
#include "object.h"
#include "memman.h"
#include "internalAPI.h"
#include "message.h"


int get_object_expand_points(storage_object_t * obj, u_int64_t * max_delta_size, dlist_t expand_points) {

	int rc = 0;

	LOG_PROC_ENTRY();

	rc = obj->plugin->functions.plugin->can_expand(obj, max_delta_size, expand_points);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int get_volume_expand_points(logical_volume_t * volume, dlist_t expand_points) {

	int rc = 0;
	sector_count_t max_delta_size = -1;

	LOG_PROC_ENTRY();

	if (volume->flags & VOLFLAG_MKFS) {
		LOG_ERROR("Volume %s cannot be expanded because it is scheduled to have a file system installed on it.\n", volume->name);
		LOG_PROC_EXIT_INT(EINVAL);
		return EINVAL;
	}

	/*
	 * If the volume currently has an FSIM and the FSIM is not scheduled to be
	 * removed, use the FSIM's limits.
	 */
	if ((volume->original_fsim != NULL) &&
	    !(volume->flags & VOLFLAG_UNMKFS)) {

		/*
		 * Reget the current size and limits because the file system could be
		 * mounted and be active.
		 */
		rc = get_volume_sizes_and_limits(volume);

		if (rc == 0) {
			max_delta_size = volume->max_vol_size - volume->vol_size;
		}
	}

	if (rc == 0) {
		rc = get_object_expand_points(volume->object, &max_delta_size, expand_points);
	}

	if (rc == 0) {
		/*
		 * If the volume has an FSIM and the FSIM is not scheduled to be removed
		 * and there are expand points, check if the FSIM can handle the expand.
		 *
		 */
		if ((volume->original_fsim != NULL) &&
		    !(volume->flags & VOLFLAG_UNMKFS) &&
		    !ListEmpty(expand_points)) {

			max_delta_size = round_down_to_hard_sector(max_delta_size, volume->object);

			rc = volume->original_fsim->functions.fsim->can_expand_by(volume, &max_delta_size);

			/*
			 * The FSIM cannot handle the expand.  Delete all the expand
			 * points from the list.
			 */
			if (rc != 0) {
				/*
				 * Can't use DeleteAllItems(expand_points, TRUE)
				 * because it uses free().  Shrink points are
				 * allocted with engine_alloc() which may add a
				 * debug header.  Extract each item and use
				 * engine_free().
				 */
				TAG tag;
				void * thing;

				while (BlindExtractObject(expand_points, &tag, NULL, &thing) == DLIST_SUCCESS) {
					engine_free(thing);
				}
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Find out if an object can be expanded.  The "thing" must be either a volume
 * or a top level object.  What we'll do is call the internal
 * get_expand_points() and see if any expand candidates show up.  If we get
 * expand points then return 0 to say that the object can be expanded.  Else,
 * return an error.
 */
int evms_can_expand(object_handle_t thing) {

	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {
		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {
			dlist_t expand_point_list = CreateList();

			if (expand_point_list != NULL) {
				switch (type) {
				case DISK:
				case SEGMENT:
				case REGION:
				case EVMS_OBJECT:
					{
						storage_object_t * obj = object;

						/*
						 * Only top level objects can be expanded.
						 */
						if (is_top_object(obj)) {
							dlist_t expand_points = CreateList();

							if (expand_points != NULL) {
								u_int64_t max_delta_size = -1;

								rc = get_object_expand_points(obj, &max_delta_size, expand_points);

								if (rc == 0) {
									if (ListEmpty(expand_points)) {
										rc = ENOENT;
									}
								}

								DestroyList(&expand_points, FALSE);

							} else {
								LOG_CRITICAL("Error allocating memory for an expand point list.\n");
								rc = ENOMEM;
							}

						} else {
							LOG_DETAILS("Object %s is not a top level object.  Only volumes and top level objects can be expanded.\n", obj->name);
							rc = EINVAL;
						}
					}
					break;

				case VOLUME:
					{
						logical_volume_t * volume = (logical_volume_t *) object;

						if (volume->flags & (VOLFLAG_ACTIVE | VOLFLAG_NEW)) {
							dlist_t expand_points = CreateList();

							if (expand_points != NULL) {
								rc = get_volume_expand_points(volume, expand_points);

								if (rc == 0) {
									if (ListEmpty(expand_points)) {
										rc = ENOENT;
									}
								}

								DestroyList(&expand_points, FALSE);

							} else {
								LOG_CRITICAL("Error allocating memory for an expand point list.\n");
								rc = ENOMEM;
							}

						} else {
							LOG_DETAILS("Volume %s cannot be expanded because it is not active.\n", volume->name);
							rc = EINVAL;
						}
					}
					break;

				default:
					LOG_ERROR("Cannot expand object of type %#x.\n", type);
					rc = EINVAL;
					break;
				}

				DestroyList(&expand_point_list, FALSE );

			} else {
				rc = ENOMEM;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Given an expand_info_t, add the app_handle for the object and the size to an
 * expand_handle_array_t.
 * The parameters of this function are structured so that it can be called by
 * ForEachItem().
 */
static int make_expand_handle_entry(ADDRESS object,
				    TAG     object_tag,
				    ADDRESS object_handle,
				    ADDRESS parameters) {
	int rc = 0;

	expand_object_info_t * object_info = (expand_object_info_t *) object;
	expand_handle_array_t * eha = (expand_handle_array_t *) parameters;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Current number of entries in handle array:  %d.\n", eha->count);

	if (object_tag == EXPAND_OBJECT_TAG) {
		LOG_DEBUG("Add entry for storage object %s.\n", object_info->object->name);

		rc = ensure_app_handle(object_info->object,
				       object_info->object->object_type,
				       &(object_info->object->app_handle));
		if (rc == 0) {
			eha->expand_point[eha->count].object = object_info->object->app_handle;
			eha->expand_point[eha->count].max_expand_size = object_info->max_expand_size;
			eha->count++;
		}

	} else {
		LOG_WARNING("Cannot make an expand handle entry from an object with tag %ld.\n", object_tag);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Make an array of expand handles (expand_handle_array_t) for the objects in
 * a dlist_t.
 */
static int make_expand_handle_array(dlist_t list, expand_handle_array_t * * eha) {
	int rc = 0;
	uint count;
	uint size;

	LOG_PROC_ENTRY();

	rc = GetListSize(list, &count);

	if (rc == 0) {
		LOG_DEBUG("Number of objects in the list:  %d\n", count);
		if (count > 1) {
			size = sizeof(expand_handle_array_t) + ((count -1) * sizeof(expand_handle_t));
		} else {
			size = sizeof(expand_handle_array_t);
		}

		*eha = alloc_app_struct(size, NULL);
		if (*eha != NULL) {
			rc = ForEachItem(list,
					 make_expand_handle_entry,
					 *eha,
					 TRUE);
		} else {
			rc = ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Find out which objects in the stack can expand.  evms_get_expand_points()
 * must be targeted at either a volume or a top level object.  Any plug-in
 * that can expand will add an expand_object_info_t to the expand_points
 * dlist.  The dlist is then converted into an expand_handle_array_t for return
 * to the user.
 */
int evms_get_expand_points(object_handle_t thing, expand_handle_array_t * * expand_points) {

	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {

		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {

			if (expand_points != NULL) {
				dlist_t expand_point_list = CreateList();

				if (expand_point_list != NULL) {

					switch (type) {
					case VOLUME:
						{
							logical_volume_t * volume = (logical_volume_t *) object;

							if (volume->flags & (VOLFLAG_ACTIVE | VOLFLAG_NEW)) {
								rc = get_volume_expand_points(volume, expand_point_list);

							} else {
								LOG_DETAILS("Volume %s cannot be expanded because it is not active.\n", volume->name);
								rc = EINVAL;
							}
						}
						break;

					case EVMS_OBJECT:
					case REGION:
					case SEGMENT:
					case DISK:
						{
							storage_object_t * obj = (storage_object_t *) object;
							/*
							 * Only top level objects can be expanded.
							 */
							if (is_top_object(obj)) {
								u_int64_t max_delta_size = -1;

								rc = get_object_expand_points(obj, &max_delta_size, expand_point_list);

							} else {
								LOG_ERROR("Object %s is not a top level object.  Only volumes and top level objects can be expanded.\n", obj->name);
								rc = EINVAL;
							}
						}
						break;

					default:
						LOG_ERROR("An object of type %d cannot be expanded.\n", type);
						rc = EINVAL;
						break;
					}

					if (rc == 0) {
						rc = make_expand_handle_array(expand_point_list, expand_points);
					}

					DestroyList(&expand_point_list, FALSE);
				}

			} else {
				LOG_ERROR("The pointer to the expand points list cannot be NULL.\n");
				rc = EINVAL;
			}

		} else {
			LOG_CRITICAL("Error when allocating memory for an expand point list.\n");
			rc = ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * If the object is part of a volume and the volume has no FSIM, warn the
 * user that there is no FSIM to handle the expand.  The user has the
 * choice of continuing or aborting.
 * Returns 0 if the user wants to continue, E_CANCELED if the user wants to abort.
 */
static int warn_if_no_fsim(storage_object_t * obj) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (obj->volume != NULL) {
		if (obj->volume->original_fsim == NULL) {
			char * choices[] = {"Continue", "Cancel", NULL};
			int answer = 0;	    /* Initialize to "Continue" */

			engine_user_message(&answer, choices,
					    "WARNING: Volume \"%s\" does not have an associated File System Interface Module which could coordinate the expansion of the file system on the volume.  "
					    "The file system will not be expanded.  "
					    "You may need to run a separate utility to expand the file system after this operation completes.  "
					    "Expanding the volume may make the file system unusable.\n",
					    obj->volume->name);

			if (answer != 0) {
				rc = E_CANCELED;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Expand the size of an object.  The expand gets targeted at the top object of
 * the feature stack.  The plug-ins pass it on down until it hits the object
 * that is to do the real expansion.
 */
int evms_expand(object_handle_t thing, handle_array_t * input_objects, option_array_t * options) {

	int rc = 0;
	void * object;
	object_type_t type;

	LOG_PROC_ENTRY();

	rc = check_engine_write_access();

	if (rc == 0) {

		rc = translate_handle(thing,
				      &object,
				      &type);

		if (rc == HANDLE_MANAGER_NO_ERROR) {

			if ((type == EVMS_OBJECT) ||
			    (type == REGION) ||
			    (type == SEGMENT) ||
			    (type == DISK)) {

				storage_object_t * obj = (storage_object_t *) object;

				if (obj->volume != NULL) {
					if (!(obj->volume->flags & (VOLFLAG_ACTIVE | VOLFLAG_NEW))) {
						LOG_DETAILS("Object %s cannot be expanded because volume %s is not active.\n", obj->name, obj->volume->name);
						rc = EINVAL;

					} else {
						if (obj->volume->flags & VOLFLAG_MKFS) {
							LOG_ERROR("Volume %s cannot be expanded because it is scheduled to have a file system installed on it.\n", obj->volume->name);
							rc = EINVAL;
						}
					}
				}

				if (rc == 0) {
					dlist_t input_object_list = CreateList();

					if (input_object_list != NULL) {
						rc = make_dlist(input_objects, input_object_list);

						if (rc == 0) {

							/*
							 * Make sure all of the input objects are storage
							 * objects.  (The caller could have supplied handles
							 * for bogus items such as plug-ins or volumes.)
							 */
							rc = ForEachItem(input_object_list,
									 isa_valid_input_object,
									 obj->disk_group,
									 TRUE);

							if (rc == DLIST_SUCCESS) {

								/*
								 * Find the top object in the feature stack.
								 */
								storage_object_t * top_object = obj;
								uint parent_count = 0;

								rc = GetListSize(top_object->parent_objects, &parent_count);

								while ((rc == DLIST_SUCCESS) && (parent_count == 1)) {
									TAG tag;

									rc = BlindGetObject(top_object->parent_objects,
											    &tag,
											    NULL,
											    FALSE,
											    (ADDRESS *) &top_object);

									if (rc == DLIST_SUCCESS) {
										rc = GetListSize(top_object->parent_objects, &parent_count);
									}
								}

								/* Clean up error codes that are OK. */
								if ((rc == DLIST_EMPTY) || (rc == DLIST_END_OF_LIST)) {
									rc = DLIST_SUCCESS;
								}

								if ((rc == DLIST_SUCCESS) && (parent_count <= 1)) {

									/*
									 * Warn the user if there is no FSIM to
									 * handle the expand.
									 */
									rc = warn_if_no_fsim(obj);

									if (rc == 0) {
										/*
										 * If the object is not part of a
										 * volume and it has stop data,
										 * remove the stop data from its
										 * current location.  It will be
										 * reapplied during commit.
										 */
										if ((top_object->volume == NULL) &&
										    (top_object->flags & SOFLAG_HAS_STOP_DATA)) {
											top_object->plugin->functions.plugin->add_sectors_to_kill_list(top_object, top_object->size - (EVMS_FEATURE_HEADER_SECTORS * 2), EVMS_FEATURE_HEADER_SECTORS * 2);
											top_object->flags &= ~SOFLAG_HAS_STOP_DATA;
										}

										/*
										 * Start the expand at the top level
										 * object.
										 */
										rc = top_object->plugin->functions.plugin->expand(top_object, obj, input_object_list, options);

										if (rc == 0) {

											int tmp_rc;
											storage_object_t * child_obj;

											DLIST_FOR_EACH(child_obj, obj->child_objects, tmp_rc) {
												/*
												 * Make sure the child objects
												 * are marked as being in the
												 * same volume.
												 */
												set_volume_in_object(child_obj,
														     get_tag_for_object(child_obj),
														     NULL,
														     obj->volume);

												/*
												 * Any stop data on the
												 * childrean can be wiped
												 * out by the parent.
												 */
												child_obj->flags &= ~SOFLAG_HAS_STOP_DATA;
											}

											if (obj->object_type == EVMS_OBJECT) {
												/*
												 * Mark all the child feature
												 * headers dirty to ensure that
												 * they get the correct depth
												 * and get written in the right
												 * place.
												 */
												ForEachItem(obj->child_objects,
													    mark_feature_headers_dirty,
													    NULL,
													    TRUE);
											}

											/*
											 * If the object is part of a
											 * volume...
											 */
											if (obj->volume != NULL) {

												/*
												 * Mark it dirty so that the
												 * feature headers get written
												 * to their new locations.
												 */
												obj->volume->flags |= VOLFLAG_DIRTY;

												if (!(obj->volume->flags & VOLFLAG_COMPATIBILITY)) {
													obj->volume->flags |= VOLFLAG_NEEDS_ACTIVATE;
												}

												/*
												 * Update the volume size.
												 */
												obj->volume->vol_size = top_object->size;

												if (top_object->feature_header != NULL) {
													/*
													 * Object has EVMS volume
													 * feature headers.
													 * Subtract them from the
													 * size.
													 */
													obj->volume->vol_size -= EVMS_FEATURE_HEADER_SECTORS * 2;
												}

												obj->volume->vol_size = round_down_to_hard_sector(obj->volume->vol_size, top_object);
											}
										}
									}

								} else {
									if (parent_count > 1) {
										LOG_ERROR("Object %s cannot be expanded because it has multiple parents.\n", obj->name);
										rc = EINVAL;
									} else {
										LOG_WARNING("Error code %d encountered when trying to find the top level object.\n", rc);
									}
								}

							} else {
								LOG_ERROR("One or more items in the input object list is not a storage object.\n");
								rc = EINVAL;
							}

						} else {
							LOG_ERROR("Error code %d when making a dlist from the input_objects handle array.\n", rc);
							rc = EINVAL;
						}

						DestroyList(&input_object_list, FALSE);

					} else {
						LOG_CRITICAL("Error allocating memory to create the input object list.\n");
						rc = ENOMEM;
					}
				}

			} else {
				LOG_ERROR("Object of type %d cannot be expanded.\n", type);
				rc = EINVAL;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * engine_can_expand_by() is a service that the Engine provides for the
 * plug-ins.  A plug-in calls engine_can_expand_by() to find out if all of its
 * parent objects, including the volume, would approve of an expansion by the
 * specified amount.  The Engine walks up the object tree calling each
 * plug-in's can_expand_by() function to see if it would approve of the
 * expansion.  Each plug-in can update the "size" if the resulting size of its
 * object would have a different delta size than the delta size of its child
 * object.  If the top of the object stack is reached and the object is part of
 * a volume, the Engine then checks the resulting delta size against any limits
 * that an FSIM may have imposed on the volume.
 */
int engine_can_expand_by(storage_object_t * object, sector_count_t * delta_size) {

	int rc = 0;
	storage_object_t * curr_object = object;
	sector_count_t original_delta_size = *delta_size;

	LOG_PROC_ENTRY();

	while ((rc == 0) && !ListEmpty(curr_object->parent_objects)) {
		TAG tag;

		rc = BlindGetObject(curr_object->parent_objects,
				    &tag,
				    NULL,
				    FALSE,
				    (ADDRESS *) &curr_object);

		if (rc == DLIST_SUCCESS) {
			rc = curr_object->plugin->functions.plugin->can_expand_by(curr_object, delta_size);
		}
	}

	/* Clean up error codes that are OK. */
	if ((rc == DLIST_EMPTY) || (rc == DLIST_END_OF_LIST)) {
		rc = DLIST_SUCCESS;
	}

	if (rc == 0) {
		/*
		 * If this object is part of a volume and the volume has an FSIM that is
		 * not scheduled to be removed, check to make sure the final size does
		 * not go over the FSIM's maximum size it can handle for its backing
		 * storage object.
		 */
		if (object->volume != NULL) {
			if ((object->volume->original_fsim != NULL) &&
			    !(object->volume->flags & VOLFLAG_UNMKFS)) {
				rc = object->volume->original_fsim->functions.fsim->can_expand_by(object->volume, delta_size);
			}
		}
	}

	if (rc == 0) {
		if (*delta_size != original_delta_size) {
			rc = EAGAIN;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

