/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libbsd.so
 *
 *   File: helpers.c
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "bsd_plugin.h"



storage_object_t *  allocate_bsd_segment( storage_object_t *object )
{
        int   rc;
        void *handle;
        DISKSEG  *seg=NULL;
        seg_private_data_t  *pdata=NULL;


        LOG_ENTRY();

        rc = bsd_eng_funcs->allocate_segment( NULL, &seg );
        if (rc) {
                LOG_EXIT_PTR(NULL);
                return NULL;
        }

        rc = InsertObject ( (dlist_t)          seg->child_objects,
                            (void *)           object,
                            (TAG)              object->object_type,
                            (void *)           NULL,
                            (Insertion_Modes)  AppendToList,
                            (boolean)          TRUE,
                            (void **)         &handle );

        if (rc == 0) {

                seg->plugin      = bsd_plugin;
                seg->object_type = SEGMENT;

                memcpy(&seg->geometry, &object->geometry, sizeof(geometry_t) );

                seg->private_data = calloc(1, sizeof(seg_private_data_t) );
                if (seg->private_data) {

                        pdata = (seg_private_data_t *)seg->private_data;

                        pdata->signature    = BSD_SEGMENT_PDATA_SIGNATURE ;
                        pdata->logical_disk = object;
                }
                else {
                        LOG_ERROR("call to malloc segment private storage area failed\n");
                        bsd_eng_funcs->free_segment( seg );
                        seg = NULL;
                }
        }
        else {
                LOG_ERROR("call to insert DISK storage object in segment child_objects dlist failed, RC= %d\n", rc );
                bsd_eng_funcs->free_segment( seg );
                seg = NULL;
        }

        LOG_EXIT_PTR(seg);
        return seg;
}


void free_bsd_segment( DISKSEG *seg )
{
        LOG_ENTRY();
        LOG_DEBUG("segment name= %s\n", seg->name );

        if (seg->private_data) free(seg->private_data);

        bsd_eng_funcs->free_segment( seg );

        LOG_EXIT_VOID();
}


DISKSEG * create_bsd_metadata_segment( storage_object_t   *object,
                                       lba_t               start,
                                       sector_count_t      size,                                      
                                       u_int32_t           object_flags )
{
        DISKSEG *metadata;

        LOG_ENTRY();

        metadata = allocate_bsd_segment( object );
        if (metadata) {
                metadata->size        = size;
                metadata->start       = start;
                metadata->data_type   = META_DATA_TYPE;
                metadata->flags       = object_flags;

		if (object->object_type == DISK) {
                        sprintf( metadata->name, "%s_bsd_disklabel", object->name );
                }
                else {
                        sprintf( metadata->name, "%s.bsd_disklabel", object->name );
                }
	       
        }

        LOG_EXIT_PTR(metadata);
        return metadata;
}


DISKSEG * create_bsd_data_segment( LOGICALDISK        *ld,
                                   lba_t               start,
                                   sector_count_t      size, 
				   bsd_partition_t    *p,
                                   u_int32_t           minor,
                                   u_int32_t           object_flags )
{
        DISKSEG *seg;
        seg_private_data_t *pdata=NULL;

        LOG_ENTRY();

        seg = allocate_bsd_segment( ld );

        if (seg) {
                seg->size              = size;
                seg->start             = start;
                seg->data_type         = DATA_TYPE;
                seg->flags             = object_flags;

                pdata = (seg_private_data_t *) seg->private_data;

		memcpy((void *)&pdata->p_record, (void *)p, sizeof(bsd_partition_t));

                pdata->minor           = minor;

                if (ld->object_type == DISK) {
                        sprintf( seg->name, "%s%d", ld->name, minor );
                }
                else {
                        sprintf( seg->name, "%s.%d", ld->name, minor );
                }

        }

        LOG_EXIT_PTR(seg);
        return seg;
}


int  remove_bsd_segment_from_list( dlist_t seglist, DISKSEG *seg )
{
        int rc;

        LOG_ENTRY();
        LOG_DEBUG("segment name= %s\n", seg->name );

        rc = DeleteObject( seglist, seg );

        if (rc == DLIST_SUCCESS) {
                bsd_eng_funcs->unregister_name( seg->name );
        }

        LOG_EXIT_INT(rc);
        return rc;
}


int insert_bsd_segment_into_list( dlist_t seglist, DISKSEG *seg)
{
        int rc;

        LOG_ENTRY();

        rc = bsd_eng_funcs->register_name( seg->name );

        if (rc==0) {
                rc = insert_bsd_segment_into_ordered_list( seglist, seg );
                if (rc) {
                        bsd_eng_funcs->unregister_name( seg->name );
                }
        }

        LOG_EXIT_INT(rc);
        return rc;
}


int insert_bsd_segment_into_ordered_list( dlist_t seglist, DISKSEG *seg )
{
        int         rc;
        void       *handle=NULL;
        DISKSEG    *seg2;
        lba_t       seg2_end_lba;
        boolean     overlapping=FALSE;


        LOG_ENTRY();
        LOG_DEBUG("seg name= %s   seg start= %"PRIu64"  ends= %"PRIu64"  size= %"PRIu64"\n", seg->name, seg->start, seg->start+seg->size-1, seg->size );


        rc = GoToStartOfList( seglist );
        if (rc == DLIST_SUCCESS) {

                while (rc == DLIST_SUCCESS) {

                rc = GetObject( seglist, SEGMENT_TAG,NULL,TRUE, (ADDRESS *) &seg2 );

                if (rc == DLIST_SUCCESS) {

                        seg2_end_lba = seg2->start + seg2->size - 1;

                        // test and set ... overlapping segments flag
                        if ( (  seg->start >= seg2->start )&&
                             (  seg->start <= seg2_end_lba)) {
                                overlapping = TRUE;
                        }
                        else if ( ( seg->start  <  seg2->start ) &&
                                  ( seg2->start <= (seg->start + seg->size - 1)) ) {
                                overlapping = TRUE;
                        }
                        else {
                                overlapping = FALSE;
                        }

                if ( overlapping == TRUE ) {

                        LOG_DEBUG("Error ... Overlapping Segments ...\n");
                        LOG_DEBUG("seg2:   name: %s\n", seg2->name );
                        LOG_DEBUG("       start: %"PRIu64"\n", seg2->start );
                        LOG_DEBUG("        size: %"PRIu64"\n", seg2->size );
                        LOG_DEBUG("         end: %"PRIu64"\n", seg2_end_lba );
                        LOG_DEBUG(" overlap lba: %"PRIu64"\n", seg->start );

                        rc = EINVAL;    // must be genuine partition overlap
                        break;          // break out of loop ... looking for insertion pt

                }


                // test for ... valid insertion point
                if (seg2->start > seg->start ) break;

                rc = NextItem( seglist );
                }

                }
        }


        switch (rc) {

        case DLIST_SUCCESS:      /* Ok, found a segment we should insert in front of */

                rc = InsertObject ( (dlist_t)          seglist,
                                    (void *)           seg,
                                    (TAG)              SEGMENT_TAG,
                                    (void *)           NULL,
                                    (Insertion_Modes)  InsertBefore,
                                    (boolean)          TRUE,
                                    (void **)         &handle );
                break;

        case DLIST_EMPTY:
        case DLIST_END_OF_LIST: /* This new segment lays out on the disk at a higher */
                                /* LBA than any other existing segment.  So, just    */
                                /* insert at end of the segment list.                */

                rc = InsertObject ( (dlist_t)          seglist,
                                    (void *)           seg,
                                    (TAG)              SEGMENT_TAG,
                                    (void *)           NULL,
                                    (Insertion_Modes)  AppendToList,
                                    (boolean)          TRUE,
                                    (void **)         &handle );
                break;

        default:     /* REAL ERROR ...  */
                LOG_ERROR("error, insertion failed ... RC= %d\n",  rc);
                break;
        }


        LOG_EXIT_INT(rc);
        return rc;
}


boolean prune_bsd_seg_objects_from_list( ADDRESS   Object,
                                         TAG       ObjectTag,
                                         ADDRESS   ObjectHandle,
                                         ADDRESS   Parameters,
                                         boolean  *FreeMemory,
                                         uint     *Error )
{
        *FreeMemory = FALSE;           // tells dlist not to free any memory
        *Error      = DLIST_SUCCESS;   // tells dlist we were successful

        // prune any storage object belonging to us
        if ( ((DISKSEG *)Object)->plugin == bsd_plugin ) {

                free_bsd_segment((storage_object_t *) Object);

                return TRUE;               // tells dlist to prune the object
        }

        return FALSE;                  // tells dlist not to prune
}

