/*
 *
 *   Copyright (C) International Business Machines  Corp., 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <plugin.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "osi_endian.h"
#include "ogfs_ondisk.h"
#include "ogfsfsim.h"
#include "cluster_config.h"

static plugin_record_t *my_plugin_record = &ogfs_plugin_record;
static engine_functions_t *EngFncs = NULL;

/**
 *	ogfs_get_volume_type - identify the type of volume from metadata in vol->private_data
 *	@volume: address of logical volume structure corresponding to OGFS volume
 *
 *	This routine examines the metadata structure pointed to by the given
 *	volume structure's private_data field. We return a enum type corresponding
 *	to the type of ogfs volume with think it represents.
 */
static ogfs_volume_type_t ogfs_get_volume_type(logical_volume_t *volume)
{
	ogfs_volume_type_t type = OGFS_UNKNOWN_VOL;
	
	if (volume->private_data != NULL) {
		uint32 magic = *((uint32 *)volume->private_data);
		
		if (magic == OGFS_MAGIC) {
			ogfs_meta_header_t *ogfs_mh;
	
			ogfs_mh = (ogfs_meta_header_t *)volume->private_data;

			if (ogfs_mh->mh_type == OGFS_METATYPE_SB) {
				type = OGFS_FS_VOL;
			} else if (ogfs_mh->mh_type == OGFS_METATYPE_LH) {
				type = OGFS_JOURNAL_VOL;
			}
		} else if (magic == OGFS_CLUSTER_GLOBAL) {
			type = OGFS_CI_VOL;
		}
	}
	
	return type;
}

/**
 *	ogfs_setup - plug-in initialization routine
 *	@engine_function_table: address of function table of engine services
 *
 *	This routine is called by the engine when the FSIM is loaded.
 *	We initialize some necessary globals and do any other necessary
 *	initialization.
 */
static int ogfs_setup(engine_functions_t *engine_function_table)
{
	int rc = 0;
		
	EngFncs = engine_function_table;	
	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_cleanup - allows for cleanup when plugin is being unloaded
 *	@void
 *
 *	This routine retrieves all volumes we have claimed and frees any
 *	memory we may have allocated and hung off each volume->private_data
 *	pointer.
 */
static void ogfs_cleanup(void)
{
	int rc;
	dlist_t global_volumes;
	logical_volume_t *volume;

	LOG_ENTRY();

	rc = EngFncs->get_volume_list(my_plugin_record, NULL, 0, &global_volumes);
	if (rc == 0) {
		while (ExtractObject(global_volumes, VOLUME_TAG, NULL, (void**)&volume) == 0) {
			if (volume->private_data != NULL) {
				EngFncs->engine_free(volume->private_data);
				volume->private_data = NULL;
			}
		}
	}

	LOG_EXIT_VOID();
	return;
}

/**
 *	ogfs_discard - cleanup data structures associated with an ogfs volume
 *	@volume: pointer to logical volume structure
 *
 *	This routine is invoked by the engine when it wants us to forget
 *	about a volume. We free our in-memory metadata structure pointed
 *	to by the volume->private_data field.
 */
static int ogfs_discard(logical_volume_t *volume)
{
	int rc = 0;
	
	LOG_ENTRY();
	
	if (volume->private_data != NULL) {
		EngFncs->engine_free(volume->private_data);
		volume->private_data = NULL;
	}
		
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	fsim_read_bytes - read specific number of bytes from a volume
 *
 *	volume: pointer to logical volume structure
 *	fd: file handle of an opened device to read from
 *	offset: byte offset from beginning of volume for start of read
 *	bytes_to_read: number of bytes to read
 *	buffer: buffer to transfer data in to
 */
static int fsim_read_bytes(logical_volume_t *volume,
			   int     fd,
			   int64_t offset,
			   int32_t bytes_to_read,
			   void   *buffer)
{
	int rc = 0;
	size_t bytes_read;
	
	LOG_ENTRY();
	
	bytes_read = EngFncs->read_volume(volume, fd, buffer,
					  bytes_to_read, offset);
	if (bytes_read != bytes_to_read)
		rc = EIO;
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	fsim_write_bytes - write specific number of bytes to a volume
 *
 *	volume: pointer to logical volume structure
 *	fd: file descriptor of an opened device to write to
 *	offset: byte offset from beginning of volume for start of write
 *	bytes_to_read: number of bytes to write
 *	buffer: buffer to transfer data out of
 */
static int fsim_write_bytes(logical_volume_t *volume,
			   int     fd,
			   int64_t offset,
			   int32_t bytes_to_write,
			   void   *buffer)
{
	int rc = 0;
	size_t bytes_written;
	
	LOG_ENTRY();
	
	bytes_written = EngFncs->read_volume(volume, fd, buffer,
					     bytes_to_write, offset);
	if (bytes_written != bytes_to_write)
		rc = EIO;
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_fs_limits - obtain minimum and maximum filesystem limits
 *	@volume: the address of the logical volume structure
 *	@min_fs_size: address to return the minimum filesystem size in sectors
 *	@max_fs_size: address to return the maximum filesystem size in sectors
 *	@max_object_size: address to return the maximum size for the volume
 *
 *	This routine returns limit information for only a type OGFS_FS_TYPE volume.
 */
static int ogfs_get_fs_limits(logical_volume_t *volume,
			      sector_count_t *min_fs_size,
			      sector_count_t *max_fs_size,
			      sector_count_t *max_object_size)
{
	int rc;
	
	LOG_ENTRY();
	
	if (ogfs_get_volume_type(volume) == OGFS_FS_VOL) {
		rc = ENOSYS;
	} else {
		rc = EPERM;
	}
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_restrict_metatypes - return ogfs superblock or journal header
 *	@buf: buffer containing metadata in disk-order as input
 *	@ogfs_mh: buffer containing metadata in cpu-order as output
 *
 *	This routine copies metadata in disk-order in the given buffer
 *	over to the ogfs_mh buffer in cpu-order. We only care about
 *	external logs and an ogfs filesystem super-block metadata types
 *	and other are summarily ignored.
 */
static int ogfs_restrict_metatypes(char *buf, ogfs_meta_header_t *ogfs_mh)
{
	int rc = 0;
	
	LOG_ENTRY();
		
	switch (ogfs_mh->mh_type) {
	case OGFS_METATYPE_SB:
		ogfs_sb_in((struct ogfs_sb *)ogfs_mh, buf);
		break;
	case OGFS_METATYPE_LH:
		ogfs_log_header_in((struct ogfs_log_header *)ogfs_mh, buf);
		break;
	default:
		rc = EINVAL;
	}
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_read_meta_header - read a block containing ogfs_meta_header
 *	@volume: pointer to evms volume structure
 *	@ogfs_mh: pointer to a buffer to copy ogfs meta data
 *	@fd: the file descriptor of an opened volume device
 *	@offset: offset from start of volume to read an ogfs disk structure
 *
 *	This routine reads an ogfs basic block from the given volume
 *	device and checks for the magic number. It converts the ogfs
 *	meta_header to cpu-order and makes additional checks to restrict what
 *	sort of meta-data we currently care about.
 */
static int ogfs_read_meta_header(logical_volume_t *volume,
				 ogfs_meta_header_t *ogfs_mh,
				 int fd,
				 int64_t offset)
{
	int rc;
	char *buf;

	LOG_ENTRY();
	
	/*
	 * Allocate a buffer for the read which will contain the metadata
	 * in disk-order. On success, the ogfs_mh will point to the copy
	 * of this metadata in cpu-order.
	 */
	buf = EngFncs->engine_alloc(OGFS_BASIC_BLOCK);
	if (buf != NULL) {	
		/*
		 * Read a block to contain at least an ogfs_meta_header and a
		 * bit more (like superblock info).
		 */
		rc = fsim_read_bytes(volume, fd, offset, OGFS_BASIC_BLOCK, buf);
		if (rc == 0) {
			ogfs_meta_header_in(ogfs_mh, buf);
			if (ogfs_mh->mh_magic != OGFS_MAGIC) {
				rc = EINVAL;
			} else {
				rc = ogfs_restrict_metatypes(buf, ogfs_mh);
			}
		}
		EngFncs->engine_free(buf);
	} else {
		rc = ENOMEM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_super_block - get an ogfs filesystem superblock
 *	@volume: pointer to evms volume structure
 *	@ogfs_mh: pointer to a buffer to copy ogfs meta data
 *	@fd: the file descriptor of an opened volume device
 */
inline static int ogfs_get_super_block(logical_volume_t *volume,
				       ogfs_meta_header_t *ogfs_mh,
				       int fd)
{
	int rc;
	
	rc = ogfs_read_meta_header(volume, ogfs_mh, fd, OGFS_SB_OFFSET);
	if (rc == 0 && ogfs_mh->mh_type != OGFS_METATYPE_SB)
		rc = EINVAL;
		
	return rc;
}

/**
 *	ogfs_get_log_header - get an ogfs external journal log header
 *	@volume: pointer to evms volume structure
 *	@ogfs_mh: pointer to a buffer to copy ogfs meta data
 *	@fd: the file descriptor of an opened volume device
 */
inline static int ogfs_get_log_header(logical_volume_t *volume,
				      ogfs_meta_header_t *ogfs_mh,
				      int fd)
{
	int rc;
	
	rc = ogfs_read_meta_header(volume, ogfs_mh, fd, OGFS_LH_OFFSET);
	if (rc == 0 && ogfs_mh->mh_type != OGFS_METATYPE_LH)
		rc = EINVAL;
		
	return rc;
}

/**
 *	ogfs_get_meta_header - returns ogfs metadata from given volume
 *	@volume: the volume to read from
 *	@ogfs_mh: the buffer to place the ogfs metadata within
 *
 *	This routine opens the given volume device and attempts to read
 *	certain types of ogfs metadata from the the volume. On success
 *	the return buffer will contain either an ogfs superblock or ogfs
 *	journal header.
 */
static int ogfs_get_meta_header(logical_volume_t *volume,
				ogfs_meta_header_t *ogfs_mh)
{
	int rc, fd;
	
	LOG_ENTRY();

	fd = EngFncs->open_volume(volume, O_RDONLY, 0);
	if (fd < 0) {
		rc = -fd;
	} else {
		/* Try getting an ogfs superblock */
		rc = ogfs_get_super_block(volume, ogfs_mh, fd);
		if (rc != 0) {
			/* Try getting an ogfs journal header */
			rc = ogfs_get_log_header(volume, ogfs_mh, fd);
			if (rc == 0) {
				/* Mark the journal volume as unmountable */
				volume->flags |= VOLFLAG_NOT_MOUNTABLE;
			}
		}
		EngFncs->close_volume(volume, fd);
	}
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_filesys_metadata - get ogfs superblock or journal metadata
 *	@volume: pointer to evms volume structure
 *	@metadata: address of pointer to have point at filesys related metadata
 *
 *	This routine attempts to read a ogfs block containing data that is not
 *	cluster related such as an ogfs filesystem superblock or external
 *	journal log header.
 */
static int ogfs_get_filesys_metadata(logical_volume_t *volume, char **metadata)
{
	int rc;
	ogfs_meta_header_t *ogfs_mh;

	LOG_ENTRY();

	/*
	 * Allocate space of OGFS_BASIC_BLOCK size which should be sufficient
	 * to read various meta types (ogfs_sb, ogfs_log_header, etc.)
	 */
	ogfs_mh = EngFncs->engine_alloc(OGFS_BASIC_BLOCK);

	if (ogfs_mh) {
		/*
		 * See if we get an ogfs_meta_header plus additional type
		 * specific data, e.g. superblock or log header.
		 */
		rc = ogfs_get_meta_header(volume, ogfs_mh);

		if (rc == 0) {
			/* return pointer to ogfs metadata */
			*metadata = (char *)ogfs_mh;
		} else {
			/* could not get ogfs metadata */
			EngFncs->engine_free(ogfs_mh);
		}
	} else {
		rc = ENOMEM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_read_cluster_info - read a block containing cluster configuration
 *	@volume: pointer to evms volume structure
 *	@ogfs_mh: pointer to a buffer to copy ogfs global cluster info
 *	@fd: the file descriptor of an opened volume device
 *
 *	This routine reads an cluster information block from the given volume
 *	device and checks for the magic number. It converts the ogfs cluster
 *	information read from disk to cpu-order.
 */
static int ogfs_read_cluster_info(logical_volume_t *volume,
				  cluster_global_t *ogfs_cg,
				  int fd)
{
	int rc;
	char *buf;

	LOG_ENTRY();
	
	buf = EngFncs->engine_alloc(CIDEV_BLOCKSIZE);
	if (buf != NULL) {	
		rc = fsim_read_bytes(volume, fd, OGFS_CG_OFFSET, CIDEV_BLOCKSIZE, buf);
		if (rc == 0) {
			cluster_global_in(ogfs_cg, buf);
			if (ogfs_cg->cg_magic != OGFS_CLUSTER_GLOBAL) {
				/* bad magic */
				rc = EINVAL;
			} else if ((((ogfs_cg->cg_version >> 16) & 0xff) !=
					((OGFS_CLUSTER_VERSION >> 16) & 0xff))
					|| (((ogfs_cg->cg_version >> 8) & 0xff) !=
					((OGFS_CLUSTER_VERSION >> 8) & 0xff))) {
				/* version mismatch */
				rc = EINVAL;
			}
		}
		EngFncs->engine_free(buf);
	} else {
		rc = ENOMEM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_cluster_info - returns ogfs cluster info from given volume
 *	@volume: the volume to read from
 *	@ogfs_cg: the buffer to place the ogfs cluster information in
 *
 *	This routine opens the given volume device and attempts to read
 *	certain types of ogfs cluster config data from the the volume.
 */
static int ogfs_get_cluster_info(logical_volume_t *volume,
				 cluster_global_t *ogfs_cg)
{
	int rc, fd;
	
	LOG_ENTRY();

	fd = EngFncs->open_volume(volume, O_RDONLY, 0);
	if (fd < 0) {
		rc = -fd;
	} else {
		rc = ogfs_read_cluster_info(volume, ogfs_cg, fd);
		EngFncs->close_volume(volume, fd);
	}
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_cluster_metadata - get ogfs cluster configuration metadata
 *	@volume: pointer to evms volume structure
 *	@metadata: address of pointer to have point at cluster related metadata
 *
 *	This routine attempts to read an ogfs cluster block containing data
 *	that contains cluster configuration information. 
 */
static int ogfs_get_cluster_metadata(logical_volume_t *volume, char **metadata)
{
	int rc;
	cluster_global_t *ogfs_cg;

	LOG_ENTRY();

	ogfs_cg = EngFncs->engine_alloc(CIDEV_BLOCKSIZE);

	if (ogfs_cg) {
		rc = ogfs_get_cluster_info(volume, ogfs_cg);

		if (rc == 0) {
			/* return pointer to ogfs metadata */
			*metadata = (char *)ogfs_cg;
			/* mark this volume as unmountable */
			volume->flags |= VOLFLAG_NOT_MOUNTABLE;
		} else {
			/* could not get ogfs metadata */
			EngFncs->engine_free(ogfs_cg);
		}
	} else {
		rc = ENOMEM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_probe - determine if this volume contains ogfs metadata
 *	@volume: the volume we are probing
 *
 *	This routine attempts to read a block from the given volume to
 *	determine if it should be claimed by this fsim if it contains the
 *	expected ogfs filesystem or cluster configuration metadata.
 */
static int ogfs_probe(logical_volume_t *volume)
{
	int rc;
	char *metadata;

	LOG_ENTRY();

	rc = ogfs_get_filesys_metadata(volume, &metadata);
	if (rc != 0)
		rc = ogfs_get_cluster_metadata(volume, &metadata);
		
	if (rc == 0) {
		volume->private_data = metadata;
	} else {
		volume->private_data = NULL;
	}
	
	LOG_EXIT_INT(rc);
	return rc;
}

/*
 *	ogfs_can_mkfs - can we install a filesystem on a given volume
 *	@volume: the address of the logical volume structure
 *
 *	This routine is invoked to determine if we can make a filesystem
 *	on the given volume.
 */
static int ogfs_can_mkfs(logical_volume_t *volume)
{
	int rc = 0;

	LOG_ENTRY();

	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If mounted, can't format. */
		rc = EBUSY;
	} else if (!(volume->flags & VOLFLAG_CLUSTER_SHARED) ||
			(volume->vol_size * EVMS_VSECTOR_SIZE) < MINOGFS) {
		/* Not a cluster shared volume of the minimum size */
		rc = EPERM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/*
 *	ogfs_can_unmkfs - can we remove out filesystem or metadata from this volume
 *	@volume: the address of the logical volume structure
 *
 *	This routine is invoked to determine if we can nuke enough metadata
 *	so that we no longer recognize this volume as belonging to ogfs.
 */
static int ogfs_can_unmkfs(logical_volume_t *volume)
{
	int rc;
	
	LOG_ENTRY();

	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If mounted, can't unmkfs. */
		rc = EBUSY;
	} else if (ogfs_get_volume_type(volume) == OGFS_FS_VOL) {
		/*
		 * We only allow the filesystem to be removed. The cidev and
		 * external journals are indentified by the filesystem and
		 * they are also removed along with it at commit time.
		 */
		rc = 0;
	} else {
		rc = EPERM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_can_fsck - can we fsck the given volume
 *	@volume: the address of the logical volume structure
 *
 *	This routine is invoked to determine if the volume can be checked
 *	for filesystem inconsistencies.
 */
static int ogfs_can_fsck(logical_volume_t *volume)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_get_fs_size(logical_volume_t *volume, sector_count_t *size)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_expand(logical_volume_t *volume, sector_count_t *new_size)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_volume_log_names - return list of external log volume names for a volume
 *	@volume: the address of the filesystem's logical volume structure
 *
 *	This routine digs through the journal index info to find the device names
 *	for the external log volumes (if any) and returns a linked list containing the
 *	names.
 */
static struct list_element *ogfs_get_volume_log_names(logical_volume_t *volume)
{
	struct list_element *head = NULL;
	
	return head;
}

/**
 *	ogfs_unclaim_volumes - unclaim volumes in volume list for the given names
 *	@volumes: list of logical volume structures to examine
 *	@volnames: list of volume names that show be unclaimed from volumes list
 *
 *	This routine loops through the list of logical volumes trying to find the
 *	logical volume structures matching either the cidev and/or external journal
 *	volume names given (sorry, ogfs doesn't track them by uuid) associated with
 *	a filesystem volume.
 *
 */
static struct list_element *ogfs_unclaim_volumes(dlist_t volumes, struct list_element *volnames)
{
	int tag;
	logical_volume_t *volume;
	struct list_element *head = NULL, *vnames = NULL;

	/*
	 * Make a copy of original names list to remove search entries as
	 * matches are found.
	 */
	vnames = duplicate_list(volnames);
		
	GoToStartOfList(volumes);

	while (BlindExtractObject(volumes, (TAG *)&tag, NULL, (void**)&volume) == 0
		&& vnames != NULL) {
		struct list_element *element;
		
		element = vnames;
		do {
			if (strcasecmp((const char *)element->data,
					(const char *)volume->dev_node) == 0) {
				struct list_element *next = element->next;
				/*
				 * Unclaim the volume; add it to the return unclaimed
				 * volumes list and then remove the search volume name
				 * from the search list.
				 */
				EngFncs->unassign_fsim_from_volume(volume);
				append_list_element(&head, volume);
				delete_list_element(&vnames, element);
				element = (vnames == NULL ? NULL : next);
				break;
			} else {
				element = element->next;
			}
		} while (element != vnames);
	}
	delete_all_list_elements(vnames, FALSE);
	
	return head;
}

/**
 *	ogfs_unmkfs_setup - handle unassignment from volume
 *	@volume: the address of the logical volume structure
 *
 *	This routine handles the ogfs fsim being removed from a volume.
 *	We find the external journals and the cidev and call the proper
 *	engine service to unclaim them as well.
 *
 *	We then create a list of the volumes associated with the filesystem
 *	volume and place the list head pointer in the reserved field of the
 *	superblock for access later from the ogfs_remove_filesystem() routine.
 */
static int ogfs_unmkfs_setup(logical_volume_t *volume)
{
	int rc = 0;
	
	LOG_ENTRY();
	
	if (ogfs_get_volume_type(volume) == OGFS_FS_VOL) {
		dlist_t ogfs_vols;
		struct ogfs_sb *ogfs_sb;
		struct list_element *volnames = NULL,*unclaimed_volumes = NULL;
		
		ogfs_sb = (struct ogfs_sb *)volume->private_data;
		
		/* get list of external journal volume names */
		volnames = ogfs_get_volume_log_names(volume);
		
		/* append cidev volume name */
		append_list_element(&volnames, strdup(ogfs_sb->sb_locktable));
		
		/* get list of ogfs claimed volumes */
		rc = EngFncs->get_volume_list(my_plugin_record, NULL, 0, &ogfs_vols);		
		if (rc == 0) {
			/*
			 * Unclaim the named volumes and return a list of the logical
			 * volume structures that were actually unclaimed
			 */
			unclaimed_volumes = ogfs_unclaim_volumes(ogfs_vols, volnames);
		}
		delete_all_list_elements(volnames, TRUE);
		
		/*
		 * Anchor the list of volumes structures associated with our
		 * filesystem volume that we just unclaimed. Place the anchor
		 * pointer within the reserved area of the superblock for use
		 * later at actual ogfs_unmkfs() time.
		 */
		*((struct list_element **)(ogfs_sb->sb_reserved)) = unclaimed_volumes;
	} else {
		rc = EPERM;
	}
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_remove_cidev - remove the cluster information global structure
 *	@volume: the address of the logical volume structure
 *
 *	This routine wipes out the cluster information global on-disk structure
 *	so that we don't claim this on a re-discover.
 */
static int ogfs_remove_cidev(logical_volume_t *volume)
{
	int fd, rc;
	
	LOG_ENTRY();
	
	fd = EngFncs->open_volume(volume, O_RDWR | O_EXCL, 0);
	
	if (fd >= 0) {
		char *zeroblock;
		
		zeroblock = EngFncs->engine_alloc(CIDEV_BLOCKSIZE);
		if (zeroblock) {
			memset((void *)zeroblock, 0, CIDEV_BLOCKSIZE);
			/* zero cluster global config */
			rc = fsim_write_bytes(volume, fd, OGFS_CG_OFFSET,
					      CIDEV_BLOCKSIZE, zeroblock);
			if (rc == 0) {
				/* free cluster info in private data */
				EngFncs->engine_free(volume->private_data);
				volume->private_data = NULL;
			}
			EngFncs->engine_free(zeroblock);
		} else {
			rc = ENOMEM;
		}
		EngFncs->close_volume(volume, fd);
	} else {
		rc = -fd;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_remove_external_journal - remove first log header at start of volume
 *	@volume: the address of the logical volume structure
 *
 *	This routine nukes the log header at the start of an external journal
 *	volume.
 */
static int ogfs_remove_external_journal(logical_volume_t *volume)
{
	int fd, rc;
	
	LOG_ENTRY();
	
	fd = EngFncs->open_volume(volume, O_RDWR | O_EXCL, 0);
	
	if (fd >= 0) {	
		char *zeroblock;
		
		zeroblock = EngFncs->engine_alloc(OGFS_BASIC_BLOCK);
		if (zeroblock) {
			memset((void *)zeroblock, 0, OGFS_BASIC_BLOCK);	
			/* zero journal log header */
			rc = fsim_write_bytes(volume, fd, OGFS_LH_OFFSET,
					      OGFS_BASIC_BLOCK, zeroblock);
			if (rc == 0) {
				/* free log header in private data */
				EngFncs->engine_free(volume->private_data);
				volume->private_data = NULL;
			}
			EngFncs->engine_free(zeroblock);
		} else {
			rc = ENOMEM;
		}
		EngFncs->close_volume(volume, fd);
	} else {
		rc = -fd;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_remove_associated_volumes - remove journals and cluster information
 *	@volumes: list of volumes associated with filesystem volume to remove
 *
 *	This routine walks the list and invokes either ogfs_remove_cidev() or
 *	ogfs_remove_journal() to remove the cluster information volume and
 *	any external journal volumes respectively.
 */
static void ogfs_remove_associated_volumes(struct list_element *volumes)
{
	struct list_element *current = volumes;
	
	do {
		ogfs_volume_type_t type;
		void *saved_private_data;
		logical_volume_t *volume;
		
		volume = (logical_volume_t *)current->data;
		
		/*
		 * Ensure that the original private_data is being used as it
		 * is possible that a mkfs was done when we unclaimed the
		 * volume in the unmkfs_setup phase.
		 */
		saved_private_data = volume->private_data;
		volume->private_data = volume->original_fsim_private_data;
		
		/* Identify the type of volume to remove */
		type = ogfs_get_volume_type(volume);
		
		if (type == OGFS_CI_VOL)
			ogfs_remove_cidev(volume);
		else if (type == OGFS_JOURNAL_VOL)
			ogfs_remove_external_journal(volume);
		
		/* Restore the saved private_data pointer before continuing */
		volume->private_data = saved_private_data;
		
		current = current->next;
	} while (current != volumes);
}

/**
 *	ogfs_remove_filesystem - nuke filesystem and associated external journals
 *	@volume: the address of the logical volume structure
 *
 *	This routine removes the superblock from the given filesystem and wipes
 *	out the first log header for each external journal the filesystem was
 *	using as well as the volume that contains the associated cluster
 *	configuration.
 */
static int ogfs_remove_filesystem(logical_volume_t *volume)
{
	int fd, rc;
	
	LOG_ENTRY();
	
	fd = EngFncs->open_volume(volume, O_RDWR | O_EXCL, 0);
	
	if (fd >= 0) {
		char *zeroblock;
		
		zeroblock = EngFncs->engine_alloc(OGFS_BASIC_BLOCK);
		if (zeroblock) {
			memset((void *)zeroblock, 0, OGFS_BASIC_BLOCK);
			/* zero superblock */
			rc = fsim_write_bytes(volume, fd, OGFS_SB_OFFSET,
					      OGFS_BASIC_BLOCK, zeroblock);
			if (rc == 0) {
				struct ogfs_sb *ogfs_sb;
				struct list_element *assoc_vols;
				
				ogfs_sb = (struct ogfs_sb *)volume->private_data;
				assoc_vols = *((struct list_element **)(ogfs_sb->sb_reserved));
				
				/* clear any external journals and cidev */
				if (assoc_vols != NULL) {
					ogfs_remove_associated_volumes(assoc_vols);
					delete_all_list_elements(assoc_vols, FALSE);
				}
			
				/* free superblock in private data */
				EngFncs->engine_free(volume->private_data);
				volume->private_data = NULL;
			}
			EngFncs->engine_free(zeroblock);
		} else {
			rc = ENOMEM;
		}
		EngFncs->close_volume(volume, fd);
	} else {
		rc = -fd;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_unmkfs - remove a filesystem or ogfs meta-data from a volume
 *	@volume: the address of the logical volume structure
 *
 *	This routine zeroes out enough meta-data from the ogfs volume to avoid
 *	it being recognized.
 */
static int ogfs_unmkfs(logical_volume_t *volume)
{
	int rc;

	LOG_ENTRY();

	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If mounted, can't unmkfs. */
		rc = EBUSY;
	} else if (ogfs_get_volume_type(volume) == OGFS_FS_VOL) {
		rc = ogfs_remove_filesystem(volume);
	} else {
		rc = EPERM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_shrink - shrink a volume
 *	@volume: address of logical volume structure
 *	@requested_size: size in sectors to shrink by
 *	@new_size: address to update new volume size
 *
 *	OpenGFS currently does not support shrinking any of its volume types.
 */
static int ogfs_shrink(logical_volume_t *volume,
		       sector_count_t requested_size,
		       sector_count_t *new_size)
{
	int rc = ENOSYS;
	
	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_mkfs_setup(logical_volume_t *volume, option_array_t *options)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_mkfs(logical_volume_t *volume, option_array_t *options)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_fsck(logical_volume_t *volume, option_array_t *options)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_get_option_count(task_context_t *context)
{
	int count = 0;

	LOG_ENTRY();

	switch (context->action) {
	case EVMS_Task_mkfs:
		break;
	case EVMS_Task_fsck:
		break;
	default:
		break;
	}

	LOG_EXIT_INT(count);
	return count;
}

static int ogfs_get_volume_list(value_list_t **value_list,
				dlist_t selected_volumes)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_init_task(task_context_t *context)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_set_option(task_context_t *context,
			   u_int32_t       index,
			   value_t        *value,
			   task_effect_t  *effect)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_set_volumes(task_context_t *context,
			    dlist_t declined_volumes,
			    task_effect_t *effect)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_cluster_extended_info - return info from the cluster config header
 *	@volume: the volume to retrieve opengfs info from
 *	@info_name: name of field to retrieve additional info. NULL for all.
 *	@info: address to return pointer to fsim extended info
 *
 *	This routine is called to return additional information related to
 *	opengfs information about the existing volume. We allocate, initialize,
 *	and return an extended_info_array_t with selected information.
 */
static int ogfs_get_cluster_extended_info(logical_volume_t *volume,
					  char *info_name,
					  extended_info_array_t **info)
{
	int rc = EINVAL;
	cluster_global_t      *ogfs_cg;
	extended_info_array_t *einfo;
	
	LOG_ENTRY();

	ogfs_cg = (cluster_global_t *)volume->private_data;
	einfo = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
				     (OGFS_CG_INFO_COUNT * sizeof(extended_info_t)));
	
	if (einfo != NULL) {
		einfo->count = OGFS_CG_INFO_COUNT;

		/* version */
		SET_STRING(einfo->info[0].name,  "Version");
		SET_STRING(einfo->info[0].title, "Version Number");
		SET_STRING(einfo->info[0].desc,  "Version number of cluster configuration data");
		einfo->info[0].type              = EVMS_Type_Unsigned_Int32;
		einfo->info[0].unit              = EVMS_Unit_None;
		einfo->info[0].value.ui32        = ogfs_cg->cg_version;
		einfo->info[0].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[0].group, 0, sizeof(group_info_t));
		
		/* lock device */
		SET_STRING(einfo->info[1].name,  "LockDev");
		SET_STRING(einfo->info[1].title, "Lock Device");
		SET_STRING(einfo->info[1].desc,  "Name of lock device");
		einfo->info[1].type              = EVMS_Type_String;
		einfo->info[1].unit              = EVMS_Unit_None;
		SET_STRING(einfo->info[1].value.s, ogfs_cg->cg_lockdev);
		einfo->info[1].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[1].group, 0, sizeof(group_info_t));

		/* data device */
		SET_STRING(einfo->info[2].name,  "DataDev");
		SET_STRING(einfo->info[2].title, "Data Device");
		SET_STRING(einfo->info[2].desc,  "Name of data device");
		einfo->info[2].type              = EVMS_Type_String;
		einfo->info[2].unit              = EVMS_Unit_None;
		SET_STRING(einfo->info[2].value.s, ogfs_cg->cg_datadev);
		einfo->info[2].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[2].group, 0, sizeof(group_info_t));

		/* callback port */
		SET_STRING(einfo->info[3].name,  "Port");
		SET_STRING(einfo->info[3].title, "Callback Port");
		SET_STRING(einfo->info[3].desc,  "Port number used for callbacks between nodes");
		einfo->info[3].type              = EVMS_Type_Unsigned_Int32;
		einfo->info[3].unit              = EVMS_Unit_None;
		einfo->info[3].value.ui32        = ogfs_cg->cg_cbport;
		einfo->info[3].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[3].group, 0, sizeof(group_info_t));
		
		/* node timeout */
		SET_STRING(einfo->info[4].name,  "Timeout");
		SET_STRING(einfo->info[4].title, "Node Timeout");
		SET_STRING(einfo->info[4].desc,  "Time in seconds before node is considered 'dead'");
		einfo->info[4].type              = EVMS_Type_Unsigned_Int32;
		einfo->info[4].unit              = EVMS_Unit_None;
		einfo->info[4].value.ui32        = ogfs_cg->cg_node_timeout;
		einfo->info[4].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[4].group, 0, sizeof(group_info_t));

		*info = einfo;
		rc = 0;
	} else {
		rc = ENOMEM;
	}
		
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_sb_extended_info - return info from the filesystem superblock
 *	@volume: the volume to retrieve opengfs info from
 *	@info_name: name of field to retrieve additional info. NULL for all.
 *	@info: address to return pointer to fsim extended info
 *
 *	This routine is called to return additional information related to
 *	opengfs information about the existing volume. We allocate, initialize,
 *	and return an extended_info_array_t with selected information.
 */
static int ogfs_get_sb_extended_info(logical_volume_t *volume,
				     char *info_name,
				     extended_info_array_t **info)
{
	int rc = EINVAL;
	struct ogfs_sb        *ogfs_sb;
	extended_info_array_t *einfo;
	
	LOG_ENTRY();
	
	ogfs_sb = (struct ogfs_sb *)volume->private_data;
	einfo = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
				     (OGFS_SB_INFO_COUNT * sizeof(extended_info_t)));
	
	if (einfo != NULL) {
		einfo->count = OGFS_SB_INFO_COUNT;

		/* generation number */
		SET_STRING(einfo->info[0].name,  "Generation");
		SET_STRING(einfo->info[0].title, "Generation Number");
		SET_STRING(einfo->info[0].desc,  "Number incremented each time a change occurs");
		einfo->info[0].type              = EVMS_Type_Unsigned_Int64;
		einfo->info[0].unit              = EVMS_Unit_None;
		einfo->info[0].value.ui64        = ogfs_sb->sb_header.mh_generation;
		einfo->info[0].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[0].group, 0, sizeof(group_info_t));

		/* filesystem block size */
		SET_STRING(einfo->info[1].name,  "BlockSize");
		SET_STRING(einfo->info[1].title, "Block Size");
		SET_STRING(einfo->info[1].desc,  "File system block size in bytes");
		einfo->info[1].type              = EVMS_Type_Unsigned_Int32;
		einfo->info[1].unit              = EVMS_Unit_Bytes;
		einfo->info[1].value.ui32        = ogfs_sb->sb_bsize;
		einfo->info[1].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[1].group, 0, sizeof(group_info_t));
		
		/* locking protocol name */
		SET_STRING(einfo->info[2].name,  "LockProtocol");
		SET_STRING(einfo->info[2].title, "Locking Protocol");
		SET_STRING(einfo->info[2].desc,  "Name of locking protocol filesystem is using");
		einfo->info[2].type              = EVMS_Type_String;
		einfo->info[2].unit              = EVMS_Unit_None;
		SET_STRING(einfo->info[2].value.s, ogfs_sb->sb_lockproto);
		einfo->info[2].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[2].group, 0, sizeof(group_info_t));

		/* lock table name */
		SET_STRING(einfo->info[3].name,  "LockTable");
		SET_STRING(einfo->info[3].title, "Lock Table");
		SET_STRING(einfo->info[3].desc,  "Name of lock table for this filesystem");
		einfo->info[3].type              = EVMS_Type_String;
		einfo->info[3].unit              = EVMS_Unit_None;
		SET_STRING(einfo->info[3].value.s, ogfs_sb->sb_locktable);
		einfo->info[3].collection_type   = EVMS_Collection_None;
		memset(&einfo->info[3].group, 0, sizeof(group_info_t));
		
		*info = einfo;
		rc = 0;
	} else {
		rc = ENOMEM;
	}
	
	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_volume_info - return FSIM specific information about a volume
 *	@volume: the volume to retrieve opengfs info from
 *	@info_name: name of field to retrieve additional info. NULL for all.
 *	@info: address to return pointer to fsim extended info
 *
 *	This routine is called to return additional information related to
 *	opengfs information about the existing volume. We allocate, initialize,
 *	and return an extended_info_array_t with selected information.
 */
static int ogfs_get_volume_info(logical_volume_t *volume,
				char *info_name,
				extended_info_array_t **info)
{
	int rc;

	LOG_ENTRY();
	
	switch (ogfs_get_volume_type(volume)) {
	case OGFS_FS_VOL:
		rc = ogfs_get_sb_extended_info(volume, info_name, info);
		break;
	case OGFS_CI_VOL:
		rc = ogfs_get_cluster_extended_info(volume, info_name, info);
		break;
	default:
		rc = ENOSYS;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/**
 *	ogfs_get_plugin_info - return standard information on plugin
 *	@descriptor_name: return information on a particular field
 *	@info: address to return pointer to plugin information
 *
 *	This routine returns an extended_info_array_t contains some
 *	standard information on the plugin.
 */
static int ogfs_get_plugin_info(char *descriptor_name,
				extended_info_array_t **info)
{
	int rc = EINVAL;
	extended_info_array_t *einfo;
	char                  version_string[64];
	char                  required_engine_api_version_string[64];
	char                  required_fsim_api_version_string[64];

	LOG_ENTRY();

	if (info) {
		if (descriptor_name == NULL) {
			*info = NULL;

			einfo = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
						      (OGFS_PLUGIN_INFO_COUNT *
						      sizeof(extended_info_t)));
			if (einfo) {
				einfo->count = OGFS_PLUGIN_INFO_COUNT;

				sprintf(version_string, "%d.%d.%d",
					MAJOR_VERSION,
					MINOR_VERSION,
					PATCH_LEVEL);

				sprintf(required_engine_api_version_string, "%d.%d.%d",
					my_plugin_record->required_engine_api_version.major,
					my_plugin_record->required_engine_api_version.minor,
					my_plugin_record->required_engine_api_version.patchlevel);

				sprintf(required_fsim_api_version_string, "%d.%d.%d",
					my_plugin_record->required_plugin_api_version.fsim.major,
					my_plugin_record->required_plugin_api_version.fsim.minor,
					my_plugin_record->required_plugin_api_version.fsim.patchlevel);

				SET_STRING(einfo->info[0].name,  "Short Name");
				SET_STRING(einfo->info[0].title, "Short Name");
				SET_STRING(einfo->info[0].desc,  "A short name given to this plugin");
				einfo->info[0].type              = EVMS_Type_String;
				einfo->info[0].unit              = EVMS_Unit_None;
				SET_STRING(einfo->info[0].value.s, my_plugin_record->short_name);
				einfo->info[0].collection_type    = EVMS_Collection_None;
				memset(&einfo->info[0].group, 0, sizeof(group_info_t));

				SET_STRING(einfo->info[1].name,  "Long Name");
				SET_STRING(einfo->info[1].title, "Long Name");
				SET_STRING(einfo->info[1].desc,  "A long name given to this plugin");
				einfo->info[1].type              = EVMS_Type_String;
				einfo->info[1].unit              = EVMS_Unit_None;
				SET_STRING(einfo->info[1].value.s, my_plugin_record->long_name);
				einfo->info[1].collection_type   = EVMS_Collection_None;
				memset(&einfo->info[1].group, 0, sizeof(group_info_t));

				SET_STRING(einfo->info[2].name,  "Type");
				SET_STRING(einfo->info[2].title, "Plugin Type");
				SET_STRING(einfo->info[2].desc,  "There are various types of plugins; each responsible for some kind of storage object.");
				einfo->info[2].type              = EVMS_Type_String;
				einfo->info[2].unit              = EVMS_Unit_None;
				SET_STRING(einfo->info[2].value.s, "File System Interface Module");
				einfo->info[2].collection_type   = EVMS_Collection_None;
				memset(&einfo->info[2].group, 0, sizeof(group_info_t));

				SET_STRING(einfo->info[3].name,  "Version");
				SET_STRING(einfo->info[3].title, "Plugin Version");
				SET_STRING(einfo->info[3].desc,  "This is the version number of the plugin.");
				einfo->info[3].type              = EVMS_Type_String;
				einfo->info[3].unit              = EVMS_Unit_None;
				SET_STRING(einfo->info[3].value.s, version_string);
				einfo->info[3].collection_type   = EVMS_Collection_None;
				memset(&einfo->info[3].group, 0, sizeof(group_info_t));

				SET_STRING(einfo->info[4].name,  "Required Engine Services Version");
				SET_STRING(einfo->info[4].title, "Required Engine Services Version");
				SET_STRING(einfo->info[4].desc,  "This is the version of the Engine services that this plug-in requires.  It will not run on older versions of the Engine services.");
				einfo->info[4].type              = EVMS_Type_String;
				einfo->info[4].unit              = EVMS_Unit_None;
				SET_STRING(einfo->info[4].value.s, required_engine_api_version_string);
				einfo->info[4].collection_type   = EVMS_Collection_None;
				memset(&einfo->info[4].group, 0, sizeof(group_info_t));

				SET_STRING(einfo->info[5].name,  "Required Engine FSIM API Version");
				SET_STRING(einfo->info[5].title, "Required Engine FSIM API Version");
				SET_STRING(einfo->info[5].desc,  "This is the version of the Engine FSIM API that this plug-in requires.  It will not run on older versions of the Engine FSIM API.");
				einfo->info[5].type              = EVMS_Type_String;
				einfo->info[5].unit              = EVMS_Unit_None;
				SET_STRING(einfo->info[5].value.s, required_fsim_api_version_string);
				einfo->info[5].collection_type   = EVMS_Collection_None;
				memset(&einfo->info[5].group, 0, sizeof(group_info_t));

				*info = einfo;

				rc = 0;
			} else {
				rc = ENOMEM;
			}
		} else {
			/* There is no more information on any of the descriptors. */
			rc = EINVAL;
		}
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_can_expand_by(logical_volume_t *volume, sector_count_t *delta)
{
	int rc = ENOSYS;

	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static int ogfs_can_shrink_by(logical_volume_t *volume, sector_count_t *delta)
{
	int rc = ENOSYS;
	
	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static fsim_functions_t ogfs_fsim_ops = {
	.setup_evms_plugin	= ogfs_setup,
	.cleanup_evms_plugin	= ogfs_cleanup,
	.probe			= ogfs_probe,
	.can_mkfs		= ogfs_can_mkfs,
	.can_unmkfs		= ogfs_can_unmkfs,
	.can_fsck		= ogfs_can_fsck,
	.get_fs_size		= ogfs_get_fs_size,
	.get_fs_limits		= ogfs_get_fs_limits,
	.can_expand_by		= ogfs_can_expand_by,
	.can_shrink_by		= ogfs_can_shrink_by,
	.expand			= ogfs_expand,
	.shrink			= ogfs_shrink,
	.mkfs			= ogfs_mkfs,
	.mkfs_setup		= ogfs_mkfs_setup,
	.fsck			= ogfs_fsck,
	.unmkfs			= ogfs_unmkfs,
	.discard		= ogfs_discard,
	.unmkfs_setup		= ogfs_unmkfs_setup,
	.get_option_count	= ogfs_get_option_count,
	.init_task		= ogfs_init_task,
	.set_option		= ogfs_set_option,
	.set_volumes		= ogfs_set_volumes,
	.get_volume_info	= ogfs_get_volume_info,
	.get_plugin_info	= ogfs_get_plugin_info
};

plugin_record_t ogfs_plugin_record = {
	.id			= SetPluginID(EVMS_OEM_IBM,
				    EVMS_FILESYSTEM_INTERFACE_MODULE,
				    FS_TYPE_OGFS),
	.version = {
		.major		= MAJOR_VERSION,
		.minor		= MINOR_VERSION,
		.patchlevel	= PATCH_LEVEL
	},
	.required_engine_api_version = {
		.major		= 12,
		.minor		= 0,
		.patchlevel	= 0
	},
	.required_plugin_api_version = {
		.fsim = {
			.major	= 9,
			.minor	= 0, 
			.patchlevel = 0
		}
	},
	.short_name		= "OGFS",
	.long_name		= "OpenGFS File System Interface Module",
	.oem_name		= "IBM",
	.functions  = {
		.fsim = &ogfs_fsim_ops
	},
	.container_functions	= NULL
};

plugin_record_t *evms_plugin_records[] = {
	&ogfs_plugin_record,
	NULL
};
