/*
 *
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: sparsesegmgr.c
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <uuid/uuid.h>
#include <plugin.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h> //new
#include "sparsesegmgr.h"

/*---------------------------------------------------------------------------+
+                                                                            +
+                   PRIVATE DATA AREAS AND SUBROUTINES                       +
+                                                                            +
+----------------------------------------------------------------------------*/

static engine_functions_t      *EngFncs = NULL;

/*
 *  Returns the current object in the specified DLIST.
 *
 *  Note, this call will not modify anything in the DLIST, including
 *  the CURRENT_ITEM ptr and so it will not advance in the DLIST.
 */
static inline storage_object_t * get_object_from_list( dlist_t list )
{
    int rc;
    storage_object_t *obj;
    storage_object_t *object_from_list = NULL;
    TAG  tag;

    rc = BlindGetObject( list, &tag, NULL, FALSE, (void **)&obj );
    if (rc == DLIST_SUCCESS) {
        object_from_list = obj;
    }

    return  object_from_list;
}

static inline boolean i_can_modify( storage_object_t * object )
{
    sparse_volume_t * pdata;

    if ( object ) {
        pdata = (sparse_volume_t *) object->private_data;

        if ( pdata && pdata->signature == EVMS_SPARSE_SIGNATURE ) {
            return TRUE;
        }
    }

    return FALSE;
}

static inline storage_object_t * get_logical_disk ( storage_object_t * object )
{
    storage_object_t * ld = NULL;

    if ( object ) {
        if ( object->object_type == DISK ) {
            ld = object ;
        }
        else if ( object->object_type == SEGMENT &&
                  object->plugin == my_plugin_record ) {

            if ( object->private_data ) {

                if ( ((sparse_volume_t*)object->private_data )->signature ==
                     EVMS_SPARSE_SIGNATURE ) {

                    ld=((sparse_volume_t*)object->private_data)->child_object;
                }
            }
        }
    }
    return ld;
}

/* Major worker bee of assign and discover */
static int allocate_sparse_segment ( storage_object_t ** seg,
                                     storage_object_t ** meta,
                                     storage_object_t ** ld,
                                     sparse_metadata_t * md )
{

    sparse_volume_t * vol;
    sparse_metavolume_t * mvol;
    int rc;
    void *handle;
    TAG tag;

    rc = EngFncs->allocate_segment( NULL, meta );
    if ( rc ) {
        LOG_ERROR("Error allocating memory\n");
        goto out_all; /* no mem free */
    }

    rc = EngFncs->allocate_segment( NULL, seg );
    if ( rc ) {
        LOG_ERROR("Error allocating memory\n");
        goto out1; /* meta segment free */
    }

    (*meta)->private_data =
        EngFncs->engine_alloc( sizeof( sparse_metavolume_t ) );

    if ( ! (*meta)->private_data ) {
        LOG_ERROR("Error allocating memory\n");
        goto out2; /* segment frees */
    }

    mvol = (sparse_metavolume_t *)(*meta)->private_data;

    (*seg)->private_data =
        EngFncs->engine_alloc( sizeof( sparse_volume_t ) );

    if ( ! (*seg)->private_data ) {
        LOG_ERROR("Error allocating memory\n");
        goto out3; /* segment frees and metavolume */
    }

    vol = (sparse_volume_t *)(*seg)->private_data;

    tag = (*ld)->object_type;

    if ( isdigit( (*ld)->name[strlen((*ld)->name) - 1]) ) {
        sprintf( (*meta)->name, "%s.metadata", (*ld)->name );
        sprintf( (*seg)->name, "%s.1", (*ld)->name );
    }
    else {
        sprintf( (*meta)->name, "%s_metadata", (*ld)->name );
        sprintf( (*seg)->name, "%s1", (*ld)->name );
    }

    (*meta)->size = md->lba_of_first_chunk;
    (*meta)->start = 0;
    (*meta)->data_type = META_DATA_TYPE;
    (*meta)->plugin = my_plugin_record;

    (*seg)->size = md->virtual_size;
    (*seg)->start = md->lba_of_first_chunk;
    (*seg)->data_type = DATA_TYPE;
    (*seg)->plugin = my_plugin_record;
    strcpy( (*seg)->uuid, md->uuid);

    memcpy(&(*meta)->geometry, &(*ld)->geometry, sizeof(geometry_t));
    memcpy(&(*seg)->geometry, &(*ld)->geometry, sizeof(geometry_t));

    rc = EngFncs->register_name( (*meta)->name );
    if (rc) {
        LOG_ERROR("error, registering name for metadata segment failed\n");
        goto out4; /* everything sparse_map free */
    }

    rc = EngFncs->register_name( (*seg)->name );
    if (rc) {
        LOG_ERROR("error, registering name for data segment failed\n");
        goto out4; /* everything but sparse_map free*/
    }

    /* Init metadata segment Private Data */
    mvol->signature = EVMS_SPARSE_SIGNATURE;
    mvol->cflags = 0;
    mvol->meta_data = *md;
    mvol->child_object = *ld;
    mvol->object = *seg;
    mvol->chunk_size = md->chunk_size;
    mvol->chunk_shift = calc_log2(md->chunk_size);
    mvol->num_chunks = md->total_chunks;
    mvol->flags = 0;

    //hash table setup
    mvol->hash_table_size = md->total_chunks / MAX_HASH_CHAIN_ENTRIES + 1;
    mvol->sparse_map = EngFncs->engine_alloc
        ( mvol->hash_table_size * sizeof( sparse_hash_entry_t* ) );
    if ( ! mvol->sparse_map )
        goto out4; /* Everything but sparse_map free */

    memset( mvol->sparse_map, 0,
            mvol->hash_table_size * sizeof( sparse_hash_entry_t* ) );

    //These will need to be updated in discover path
    mvol->next_cow_entry = 0;
    mvol->current_cow_sector = md->lba_of_COW_table;
    mvol->next_free_chunk = 0;
    memset ( mvol->cow_table, 0xff, EVMS_VSECTOR_SIZE );

    LOG_DEBUG(" %ld sectors per chunk, %ld total chunks\n",
              mvol->chunk_size, mvol->num_chunks );

    rc = InsertObject ( (dlist_t)          (*meta)->child_objects,
                        (void *)           (*ld),
                        (TAG)              tag,
                        (void *)           NULL,
                        (Insertion_Modes)  AppendToList,
                        (boolean)          TRUE,
                        (void **)         &handle );

    if ( rc != DLIST_SUCCESS ) {
        LOG_ERROR("Error inserting parent into child_objects\n");
        goto out5;
    }

    rc = InsertObject ( (dlist_t)          (*ld)->parent_objects,
                        (void *)           (*meta),
                        (TAG)              SEGMENT_TAG,
                        (void *)           NULL,
                        (Insertion_Modes)  AppendToList,
                        (boolean)          TRUE,
                        (void **)         &handle );

    if ( rc != DLIST_SUCCESS ){
        LOG_ERROR("Error inserting child into parent_objects\n");
        goto out5;
    }


    /* Init Data segment private data */
    vol->signature = EVMS_SPARSE_SIGNATURE;
    vol->cflags = 0;
    vol->child_object = *ld;
    vol->object = *meta;

    rc = InsertObject ( (dlist_t)          (*seg)->child_objects,
                        (void *)           (*ld),
                        (TAG)              tag,
                        (void *)           NULL,
                        (Insertion_Modes)  AppendToList,
                        (boolean)          TRUE,
                        (void **)         &handle );

    if ( rc != DLIST_SUCCESS ) {
        LOG_ERROR("Error inserting parent into child_objects\n");
        goto out5;
    }

    rc = InsertObject ( (dlist_t)          (*ld)->parent_objects,
                        (void *)           (*seg),
                        (TAG)              SEGMENT_TAG,
                        (void *)           NULL,
                        (Insertion_Modes)  AppendToList,
                        (boolean)          TRUE,
                        (void **)         &handle );

    if ( rc != DLIST_SUCCESS ){
        LOG_ERROR("Error inserting child into parent_objects\n");
        goto out5;
    }

    goto out_all;

    /* Error path exits */
    /* might need to do some emptying of lists here */
 out5:
    EngFncs->engine_free( mvol->sparse_map );
 out4:
    EngFncs->engine_free( vol );
 out3:
    EngFncs->engine_free( mvol );
 out2:
    EngFncs->free_segment( *seg );
 out1:
    EngFncs->free_segment( *meta );

    /* Everybody's exit */
 out_all:
    return rc;

}

/********************  Expand option functions *********************/

static int allocate_expand_option_descriptors( task_context_t * context )
{
    u_int64_t max, min;
    storage_object_t * obj;
    sparse_metavolume_t *vol;
    sparse_volume_t *pdata;

    if ( ! context ) {
        return EINVAL;
    }

    context->option_descriptors->count = SPARSE_EXPAND_OPTIONS;
    obj = context->object;
    if ( ! obj ) {
        return EINVAL;
    }

    pdata = (sparse_volume_t *)obj->private_data;
    vol = (sparse_metavolume_t *)pdata->object->private_data;
    min = vol->chunk_size;
    max = SPARSE_MAX_VIRTUAL_SIZE - vol->meta_data.virtual_size;

    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].flags = 0;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.list = NULL;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range =
        EngFncs->engine_alloc( sizeof(value_range_t) );
    if ( context->option_descriptors->
         option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range==NULL ) {
        return ENOMEM;
    }
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint_type =
        EVMS_Collection_Range;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
        min.ui64 = min;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
        max.ui64 = max;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
        increment.ui64 = min;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].help = NULL;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].name =
        strdup(SPARSE_OPTION_VIRTUALSIZE_NAME);
    SET_STRING(context->option_descriptors->
                     option[SPARSE_OPTION_VIRTUALSIZE_INDEX].tip,
                     "Use this option to specifiy how much virtual space to add to the segment." );
    SET_STRING(context->option_descriptors->
                     option[SPARSE_OPTION_VIRTUALSIZE_INDEX].title,
                     "Additional Size" );
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].type =
        EVMS_Type_Unsigned_Int64;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].unit = EVMS_Unit_Sectors;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].value.ui64 = min;

    return 0;
}

static int get_expand_options ( option_array_t * options,
                                sector_count_t * size )
{
    int i;
    int rc = EINVAL;

    for ( i = 0; i < options->count; i++ ) {

        if ( options->option[i].is_number_based ) {

            if ( options->option[i].number == SPARSE_OPTION_VIRTUALSIZE_INDEX )
                {
                    *size = options->option[i].value.ui64;
                    rc = 0;
                }

        }
        else {

            if ( strcmp(options->option[i].name,
                        SPARSE_OPTION_VIRTUALSIZE_NAME ) == 0 ) {
                *size = options->option[i].value.ui64;
                rc = 0;
            }
        }
    }

    return rc;
}

static int set_expand_option( task_context_t * context,
                              u_int32_t index, value_t * value,
                              task_effect_t * effect )
{
    int rc = EINVAL;
    u_int64_t max, min;

    if ( context ) {

        switch ( index ) {
        case SPARSE_OPTION_VIRTUALSIZE_INDEX:
            max = context->option_descriptors->option[index].
                constraint.range->max.ui64;
            min = context->option_descriptors->option[index].
                constraint.range->min.ui64;

            if ( ( value->ui64 >= min ) &&
                 ( value->ui64 <= max ) ) {
                context->option_descriptors->
                    option[index].value.ui64 = value->ui64;
                rc = 0;
            }
            else if ( value->ui64 < min ) {
                context->option_descriptors->option[index].value.ui64 = min;
                rc =0;
            }
            else if ( value->ui64 > max ) {
                context->option_descriptors->option[index].value.ui64 = min;
                rc =0;
            }
            break;

        default:
            break;

        }
    }
    return rc;

}

/******************  Shrink option functions ************************/

static int allocate_shrink_option_descriptors( task_context_t * context )
{
    u_int64_t max, min;
    storage_object_t * obj;
    sparse_volume_t * pdata;
    sparse_metavolume_t *vol;

    if ( ! context ) {
        return EINVAL;
    }

    obj = context->object;
    if (! obj ) {
        return EINVAL;
    }

    pdata = (sparse_volume_t*)obj->private_data;
    vol = (sparse_metavolume_t *)pdata->object->private_data;
    min = vol->chunk_size;
    max = vol->meta_data.virtual_size -
        max( SPARSE_MIN_VIRTUAL_SIZE, vol->meta_data.child_object_size );

    context->option_descriptors->count = SPARSE_SHRINK_OPTIONS;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].flags = 0;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.list = NULL;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range =
        EngFncs->engine_alloc( sizeof(value_range_t) );
    if ( context->option_descriptors->
         option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range == NULL )
        {
            return ENOMEM;
        }
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint_type =
        EVMS_Collection_Range;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
        min.ui64 = min;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
        max.ui64 = max;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
        increment.ui64 = min;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].help = NULL;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].name =
        strdup(SPARSE_OPTION_VIRTUALSIZE_NAME);
    SET_STRING(context->option_descriptors->
                     option[SPARSE_OPTION_VIRTUALSIZE_INDEX].tip,
                     "Use this option to specifiy how much virtual space "
                     "to subtract from the segment. This will cause the "
                     "cow table to be rewritten. WARNING all data will "
                     "be lost. ");
    SET_STRING(context->option_descriptors->
                     option[SPARSE_OPTION_VIRTUALSIZE_INDEX].title,
                     "Shrink by Size" );
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].type =
        EVMS_Type_Unsigned_Int64;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].unit = EVMS_Unit_Sectors;
    context->option_descriptors->
        option[SPARSE_OPTION_VIRTUALSIZE_INDEX].value.ui64 = min;


    return 0;
}

static int get_shrink_options ( option_array_t * options,
                                sector_count_t * size )
{
    int i;
    int rc = EINVAL;

    for ( i = 0; i < options->count; i++ ) {

        if ( options->option[i].is_number_based ) {

            if ( options->option[i].number ==
                 SPARSE_OPTION_VIRTUALSIZE_INDEX ) {
                *size = options->option[i].value.ui64;
                rc = 0;
            }

        }
        else {

            if ( strcmp( options->option[i].name,
                         SPARSE_OPTION_VIRTUALSIZE_NAME ) == 0 ) {
                *size = options->option[i].value.ui64;
                rc = 0;
            }
        }
    }

   return rc;
}

static int set_shrink_option( task_context_t * context, u_int32_t index,
                              value_t * value, task_effect_t * effect )
{
    int rc = EINVAL;
    u_int64_t max, min;

    if ( context ) {
        max = context->option_descriptors->
            option[index].constraint.range->max.ui64;
        min = context->option_descriptors->
            option[index].constraint.range->min.ui64;

        switch ( index ) {
        case SPARSE_OPTION_VIRTUALSIZE_INDEX:
            if ( ( value->ui64 >= min ) &&
                 ( value->ui64 <= max ) ) {
                context->option_descriptors->
                    option[index].value.ui64 = value->ui64;
                rc = 0;
            }
            else if ( value->ui64 < min ) {
                context->option_descriptors->option[index].value.ui64 = min;
                rc = 0;
            }
            else if ( value->ui64 > max ) {
                context->option_descriptors->option[index].value.ui64 = max;
                rc = 0;
            }

            break;

        default:
            break;

        }
    }
    return rc;
}

/******************  Assign option functions ***********************/

static int allocate_assign_option_descriptors( task_context_t * context )
{
    int rc = EINVAL;

    if ( context ) {
        context->option_descriptors->count = SPARSE_ASSIGN_OPTIONS;

        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].flags = 0;
        SET_POWER2_LIST( context->option_descriptors->
                         option[SPARSE_OPTION_CHUNKSIZE_INDEX].constraint.list,
                         SPARSE_MIN_CHUNK_SIZE,
                         SPARSE_MAX_CHUNK_SIZE);
        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].constraint_type =
            EVMS_Collection_List;
        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].help = NULL;
        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].name =
            strdup(SPARSE_OPTION_CHUNKSIZE_NAME);
        SET_STRING(context->option_descriptors->
                         option[SPARSE_OPTION_CHUNKSIZE_INDEX].tip,
                         "Chunksize is the unit size of the data that will be"
                         "tracked by in the sparse." );
        SET_STRING(context->option_descriptors->
                         option[SPARSE_OPTION_CHUNKSIZE_INDEX].title,
                         "Chunk Size:" );
        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].type =
            EVMS_Type_Unsigned_Int32;
        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].unit = EVMS_Unit_Sectors;
        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].value.ui32 =
            SPARSE_DEFAULT_CHUNK_SIZE;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].flags = 0;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.list = NULL;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range =
            EngFncs->engine_alloc( sizeof(value_range_t) );
        if ( context->option_descriptors->
             option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range == NULL )
            {
                return ENOMEM;
            }
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint_type =
            EVMS_Collection_Range;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
            min.ui64 = SPARSE_MIN_VIRTUAL_SIZE;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
            max.ui64 = SPARSE_MAX_VIRTUAL_SIZE;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
            increment.ui64 = SPARSE_DEFAULT_CHUNK_SIZE;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].help = NULL;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].name =
            strdup(SPARSE_OPTION_VIRTUALSIZE_NAME);
        SET_STRING( context->option_descriptors->
                          option[SPARSE_OPTION_VIRTUALSIZE_INDEX].tip,
                          "Virtual Size is the size of the data that you want"
                          "the sparse to represent." );
        SET_STRING( context->option_descriptors->
                          option[SPARSE_OPTION_VIRTUALSIZE_INDEX].title,
                          "Virtual Size:" );
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].type =
            EVMS_Type_Unsigned_Int64;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].unit = EVMS_Unit_Sectors;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].value.ui64 =
            SPARSE_DEFAULT_VIRTUAL_SIZE;
        rc = 0;
    }

    return rc;
}

static int initialize_assign_option_descriptors( storage_object_t * ld,
                                                 task_context_t * context )
{
    int rc = EINVAL;

    if ( ld && context ) {

        if ( context->option_descriptors->count != SPARSE_ASSIGN_OPTIONS ) {
            LOG_ERROR("error, wrong number(%d) of assign options.\n",
                      context->option_descriptors->count );
            return rc;
        }

        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].flags = 0;
        context->option_descriptors->
            option[SPARSE_OPTION_CHUNKSIZE_INDEX].value.ui32 =
            SPARSE_DEFAULT_CHUNK_SIZE;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].flags = 0;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].value.ui64 =
            SPARSE_DEFAULT_VIRTUAL_SIZE;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
            min.ui64 = max( SPARSE_MIN_VIRTUAL_SIZE, ld->size );
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
            max.ui64 = SPARSE_MAX_VIRTUAL_SIZE;
        context->option_descriptors->
            option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
            increment.ui64 = SPARSE_DEFAULT_CHUNK_SIZE;
        rc = 0;
    }

    return rc;
}

static int get_assign_options( option_array_t * options,
                               u_int32_t * chunk_size,
                               u_int64_t * virtual_size )
{
    int i;
    int rc = 0;

        for ( i = 0; i < options->count && rc == 0; i++ ) {

        if ( options->option[i].is_number_based ) {

            switch ( options->option[i].number ) {

            case SPARSE_OPTION_CHUNKSIZE_INDEX:
                // option is in 512 byte sectors
                *chunk_size = options->option[i].value.ui32;
                break;

            case SPARSE_OPTION_VIRTUALSIZE_INDEX:

                *virtual_size = options->option[i].value.ui64;
                break;

            default:
                rc = EINVAL;
                break;
            }

        }
        else {

            if ( strcmp( options->option[i].name,
                         SPARSE_OPTION_CHUNKSIZE_NAME ) == 0 ) {

                *chunk_size = options->option[i].value.ui32;

            }
            else if ( strcmp( options->option[i].name,
                              SPARSE_OPTION_VIRTUALSIZE_NAME ) == 0 ) {

                *virtual_size = options->option[i].value.ui64;

            }
            else {
                rc = EINVAL;
            }
        }

    }

    return rc;
}

static int set_assign_option( task_context_t * context, u_int32_t index,
                              value_t * value, task_effect_t * effect )
{
    int rc = EINVAL;
    u_int64_t min, max;

    switch ( index ) {

    case SPARSE_OPTION_CHUNKSIZE_INDEX:
        if ( ( value->ui32 >= SPARSE_MIN_CHUNK_SIZE ) &&
             ( value->ui32 <= SPARSE_MAX_CHUNK_SIZE ) &&
             (value->ui32 & ( value->ui32 - 1 ) ) == 0 ) {
            context->option_descriptors->
                option[index].value.ui32 = value->ui32;
            context->option_descriptors->
                option[SPARSE_OPTION_VIRTUALSIZE_INDEX].constraint.range->
                increment.ui64 = value->ui32;
            *effect |= EVMS_Effect_Reload_Options;
            rc = 0;
        }
        break;

    case SPARSE_OPTION_VIRTUALSIZE_INDEX:
        min = context->option_descriptors->
            option[index].constraint.range->min.ui64;
        max = context->option_descriptors->
            option[index].constraint.range->max.ui64;

        if ( ( value->ui64 >= min ) &&
             ( value->ui64 <= max ) ) {
            context->option_descriptors->
                option[index].value.ui64 = value->ui64;
            rc = 0;
        }
        else if ( value->ui64 < min ) {
            context->option_descriptors->
                option[index].value.ui64 = min;
            rc = 0;
        }
        else if ( value->ui64 > max ) {
            context->option_descriptors->
                option[index].value.ui64 = max;
            rc = 0;
        }
        break;

    default:
        break;

   }
    return rc;
}

static boolean assign_prune_acceptable_disks ( ADDRESS   Object,
                                               TAG       ObjectTag,
                                               ADDRESS   ObjectHandle,
                                               ADDRESS   Parameters,
                                               boolean  *FreeMemory,
                                               uint     *Error )
{
    storage_object_t * obj = (storage_object_t *)Object;
    sparse_volume_t *pdata=NULL;
    *FreeMemory = FALSE;
    *Error = DLIST_SUCCESS;

    if ( obj ) {
        if ( obj->plugin != my_plugin_record ) {

            if ( obj->object_type == DISK ) {
                return FALSE;
            }
            else if ( obj->object_type == SEGMENT ) {
                pdata = (sparse_volume_t *)obj->private_data;

                if ( ! ( pdata->cflags & SEG_CFLAG_TOP_SEGMENT ) ) {
                    return FALSE;
                }
            }
        }
    }
    return TRUE;
}

static int get_acceptable_assign_objects( task_context_t * context )
{
    int   rc = EINVAL;
    uint  count;

    rc = GetListSize(context->acceptable_objects, &count);
    if (rc == 0) {

        if (count == 0) {

            rc = EngFncs->get_object_list(DISK | SEGMENT,
                                                 DATA_TYPE,
                                                 NULL,
                                                 NULL,
                                                 TOPMOST,
                                                 &context->acceptable_objects);

            if (rc == DLIST_SUCCESS) {

                rc = GetListSize(context->acceptable_objects, &count);
                if (rc) {
                    count = 0;
                }

                if (count > 0) {
                    PruneList( context->acceptable_objects,
                               assign_prune_acceptable_disks,
                               NULL );
                }

                rc = 0;

            }

        }
        else {
            LOG_ERROR("context already has acceptable objects\n");
        }

    }

    return rc;
}

static int set_assign_object( task_context_t * context,
                              dlist_t declined_objects,
                              task_effect_t * effect )
{
    int rc;
    storage_object_t *obj;
    void * handle;
    uint segment_count;
    declined_object_t *declined_object = NULL;

    rc = GetListSize( context->selected_objects, &segment_count );
    if ( rc || ( segment_count != 1 ) ) {
        LOG_ERROR("Sparse Assign only supports one selected object\n");
        return rc;
    }

    rc = GoToStartOfList( context->selected_objects );
    if ( rc != DLIST_SUCCESS ) {
        return rc;
    }

    obj = get_object_from_list( context->selected_objects );
    if ( !obj ) {
        LOG_ERROR("Couldn't get selected object\n");
        return DLIST_EMPTY;
    }
    rc = GetListSize( obj->parent_objects, &segment_count);
    if ( rc || ( segment_count != 0 ) ) {
        declined_object =
            EngFncs->engine_alloc( sizeof( declined_object_t ) );
        if ( !declined_object ) {
            LOG_ERROR("error, unable to malloc a declined object struct\n");
            return ENOMEM;
        }

        declined_object->object = obj;
        declined_object->reason = rc; //probably not what the UI really wants

        rc = InsertObject ( declined_objects,
                            declined_object,
                            DECLINED_OBJECT_TAG,
                            NULL,
                            AppendToList,
                            TRUE,
                            &handle );

        if ( rc != DLIST_SUCCESS ) {
                EngFncs->engine_free(declined_object);
                return rc;
        }
    }
    else {
        rc = initialize_assign_option_descriptors( obj, context );
    }
    if ( rc == DLIST_SUCCESS ) {
        *effect |= EVMS_Effect_Reload_Objects;
    }
    return rc;
}

/*
 *      Performs endian conversion on metadata sector.
 */
static int convert_metadata( sparse_metadata_t * metadata )
{
    metadata->signature = DISK_TO_CPU32(metadata->signature);
    metadata->CRC = DISK_TO_CPU32(metadata->CRC);
    metadata->version.major = DISK_TO_CPU32(metadata->version.major);
    metadata->version.minor = DISK_TO_CPU32(metadata->version.minor);
    metadata->version.patchlevel = DISK_TO_CPU32(metadata->version.patchlevel);
    metadata->flags = DISK_TO_CPU32(metadata->flags);
    metadata->virtual_size = DISK_TO_CPU64(metadata->virtual_size);
    metadata->child_object_size = DISK_TO_CPU64(metadata->child_object_size);
    metadata->lba_of_COW_table = DISK_TO_CPU64(metadata->lba_of_COW_table);
    metadata->lba_of_first_chunk = DISK_TO_CPU64(metadata->lba_of_first_chunk);
    metadata->chunk_size = DISK_TO_CPU32(metadata->chunk_size);
    metadata->total_chunks = DISK_TO_CPU32(metadata->total_chunks);

    return 0;
}

/******************** Hash Table Functions ********************/

/*
 *      Get a sparse_hash_entry_t from the pool and initialize.
 */
static sparse_hash_entry_t *
allocate_sparse_hash_entry( sparse_metavolume_t * volume,
                            u_int64_t   org_chunk,
                            u_int64_t   sparse_chunk )
{
    sparse_hash_entry_t * hash_entry;

    //YIKES - NOT doing anything about full

    hash_entry = (sparse_hash_entry_t *)
        EngFncs->engine_alloc( sizeof( sparse_hash_entry_t ) );

    if ( hash_entry ) {
        hash_entry->org_chunk   = org_chunk;
        hash_entry->sparse_chunk        = sparse_chunk;
        hash_entry->next        = NULL;
        hash_entry->prev        = NULL;
    }

    return hash_entry;
}

/*
 *      This function inserts a new entry into a sparse hash chain, immediately
 *      following the specified entry. This function should not be used to add
 *      an entry into an empty list, or as the first entry in an existing list.
 *      For that case, use insert_sparse_map_entry_at_head().
 */
static int insert_sparse_hash_entry( struct sparse_hash_entry * entry,
                                     struct sparse_hash_entry * base )
{
        entry->next = base->next;
        entry->prev = base;
        base->next = entry;
        if ( entry->next ) {
                entry->next->prev = entry;
        }
        return 0;
}

/*
 *      This function inserts a new entry into a sparse chain as the first
 *      entry in the chain.
 */
static int insert_sparse_hash_entry_at_head( sparse_hash_entry_t * entry,
                                             sparse_hash_entry_t ** head )
{
        entry->next = *head;
        entry->prev = NULL;
        *head = entry;
        if ( entry->next ) {
                entry->next->prev = entry;
        }
        return 0;
}

static void delete_sparse_hash_chain( sparse_hash_entry_t * head )
{
    sparse_hash_entry_t * next ;

    while ( head ) {
        next = head->next;
        EngFncs->engine_free( head );
        head = next;
    }
}

static void delete_sparse_hash_map ( sparse_metavolume_t* vol, int free_map )
{
    int i;

    if (  vol->sparse_map ) {
        for ( i = 0 ; i < vol->hash_table_size; i++ ) {
            delete_sparse_hash_chain( vol->sparse_map[i] );
            vol->sparse_map[i] = NULL;
        }
        if ( free_map ) {
            EngFncs->engine_free( vol->sparse_map );
            vol->sparse_map = NULL;
        }
    }
}

static int search_sparse_hash_chain( u_int64_t chunk,
                                     sparse_hash_entry_t * head,
                                     sparse_hash_entry_t ** result )
{
    sparse_hash_entry_t * curr = head;
    sparse_hash_entry_t * prev = head;

    while ( curr && curr->org_chunk < chunk ) {
        prev = curr;
        curr = curr->next;
    }

    if ( ! curr ) {
        *result = prev;
        return 0;
    }

    else if ( curr->org_chunk != chunk ) {
        *result = curr->prev;
        return 0;
    }

    else {
        *result = curr;
        return 1;
    }
}

static int add_cow_entry_to_sparse_map( u_int64_t org_chunk,
                                        u_int64_t sparse_chunk,
                                        sparse_metavolume_t * volume )
{
    sparse_hash_entry_t * new_entry;
    sparse_hash_entry_t * target_entry;
    unsigned long hash_value;

    new_entry = allocate_sparse_hash_entry( volume, org_chunk, sparse_chunk );
    if ( ! new_entry ) {
        return ENOMEM;
    }

    hash_value = (long)org_chunk % volume->hash_table_size;
    if ( ! search_sparse_hash_chain( org_chunk,
                                     volume->sparse_map[hash_value],
                                     &target_entry ) ) {
        if ( target_entry ) {
            insert_sparse_hash_entry( new_entry, target_entry );
        }
        else {
            insert_sparse_hash_entry_at_head
                ( new_entry, &(volume->sparse_map[hash_value]));
        }
    }

    return 0;
}

static int build_sparse_maps( sparse_metavolume_t * volume, int del_map )
{

    int rc = 0;
    int done = 0;

    if ( del_map ) {
        delete_sparse_hash_map( volume, FALSE );
        memset( volume->sparse_map, 0,
                volume->hash_table_size * sizeof( sparse_hash_entry_t* ) );
    }

    while ( ! done ) {
        // Read in one sector's worth of COW tables.
        if ( READ( volume->child_object, volume->current_cow_sector, 1,
                   volume->cow_table ) ) {
            rc = EIO;
            goto out;
        }

        //Translate valid into map entries
        for ( volume->next_cow_entry = 0;

              volume->next_cow_entry <
                  ( EVMS_VSECTOR_SIZE / sizeof( u_int64_t ) ) &&
                  volume->cow_table[volume->next_cow_entry] !=
                  0xffffffffffffffffULL;

              volume->next_cow_entry++, volume->next_free_chunk++ ) {

            if ( ( rc = add_cow_entry_to_sparse_map
                   ( volume->cow_table[volume->next_cow_entry],
                     volume->next_free_chunk, volume ) ) ) {
                goto out;
            }
        }

        if ( volume->next_cow_entry ==
             ( EVMS_VSECTOR_SIZE / sizeof( u_int64_t ) ) ) {
            volume->current_cow_sector++;
        }
        else {
            done = 1;
        }
    }

 out:
    if ( rc )
        delete_sparse_hash_map( volume, TRUE );

    return rc;

}

static int sparse_segment_discovery(  void *       Object,
                                      TAG          ObjectTag,
                                      void *       ObjectHandle,
                                      void *       Parameters )
{
    storage_object_t * obj;
    storage_object_t * seg = NULL;
    storage_object_t * meta = NULL;
    int rc;
    void *handle;
    dlist_t output_objects;
    sparse_volume_t *pdata = NULL;
    sparse_metavolume_t *vol = NULL;
    int org_crc, new_crc;
    char data[EVMS_VSECTOR_SIZE];
    sparse_metadata_t *metadata = (sparse_metadata_t *)data;
    dm_target_t * trgts;

    LOG_DEBUG("examining object %s\n", ((storage_object_t *)Object)->name);

    //pickup the parameters
    obj = (storage_object_t *) Object;
    output_objects = ((discovery_parm_block_t *)Parameters)->output_objects;
    pdata = (sparse_volume_t *) obj->private_data;

    //test if the object is acceptable
    if ( ( obj->plugin == my_plugin_record ) ||
         ( obj->data_type != DATA_TYPE ) ||
         ( obj->object_type == SEGMENT && ( pdata->cflags&SEG_CFLAG_TOP_SEGMENT ) ) ) {
        LOG_DEBUG("object is not acceptable\n" );
        goto out;
    }

    //get metadata
    rc = READ( obj, 0, 1, (void *)data);
    if ( rc ) {
        LOG_ERROR("Error reading object %s rc=%d\n", obj->name, rc);
        goto out;
    }
    if ( DISK_TO_CPU32( metadata->signature ) != EVMS_SPARSE_SIGNATURE ) {
        rc = EINVAL;
        LOG_DEBUG("Object not mine, no Sparse Metadata found on %s rc=%d\n",
                  obj->name, rc);
        goto out;
    }
    LOG_DEBUG("Found a Sparse on %s\n", obj->name);
    org_crc = DISK_TO_CPU32( metadata->CRC );
    metadata->CRC = 0;
    new_crc = EngFncs->
        calculate_CRC( EVMS_INITIAL_CRC, metadata, sizeof(sparse_metadata_t) );

    if ( new_crc != org_crc ) {
        LOG_ERROR("Sparse:Error in crc on object %s \n", obj->name);
        rc = EVMS_FEATURE_FATAL_ERROR;
        goto out;
    }
    convert_metadata( metadata );
    if ( metadata->version.major > my_plugin_record->version.major ) {
        LOG_ERROR("Sparse:Error Metadata at higher version than plugin %s \n",
                  obj->name);
        rc = EVMS_FEATURE_FATAL_ERROR;
        goto out;
    }

    rc = allocate_sparse_segment( &seg, &meta, &obj, metadata );
    if ( rc ) {
        goto out;
    }
    meta->flags &= ~SOFLAG_ACTIVE;
    meta->dev_major = 0;
    meta->dev_minor = 0;


    rc = EngFncs->dm_update_status( seg );
    if (!rc) {

	if (seg->flags & SOFLAG_ACTIVE) {
	
		rc = EngFncs->dm_get_targets( seg, &trgts);
		if (!rc) {
		
			if ( trgts->start != 0 ||
			     trgts->length != metadata->virtual_size ||
			     trgts->type != DM_TARGET_SPARSE ||
			     trgts->next != NULL ||
			     trgts->data.sparse->device.major != obj->dev_major ||
			     trgts->data.sparse->device.minor != obj->dev_minor ||
			     trgts->data.sparse->device.start != metadata->lba_of_first_chunk ||
			     trgts->data.sparse->num_chunks != metadata->total_chunks ||
			     trgts->data.sparse->chunk_size != metadata->chunk_size ) {

			     LOG_DEBUG("kernel sparse seg metadata is incorrect\n");
			     seg->flags |= SOFLAG_NEEDS_ACTIVATE;

			}
		
			EngFncs->dm_deallocate_targets( trgts );
		}
	}
	else {
		seg->flags |= SOFLAG_NEEDS_ACTIVATE;
	}

    }

    vol = (sparse_metavolume_t *)meta->private_data;

    rc = build_sparse_maps( vol, FALSE );
    if ( rc ) {
        LOG_ERROR("not able to build all of sparse_map");
        goto out1;
    }

    rc = CopyList( output_objects, obj->parent_objects, AppendToList );

    if ( rc != DLIST_SUCCESS ) {
        goto out2;
    }
    else {
        //metadata and data segment added
        *((discovery_parm_block_t *)Parameters)->object_count += 2;
    }

    goto out_all;

    /****** error conditions clean-up ******/
 out2:
    delete_sparse_hash_map( vol, TRUE );
 out1:
    EngFncs->engine_free( meta->private_data );
    EngFncs->engine_free( seg->private_data );
    EngFncs->free_segment( seg );
    EngFncs->free_segment( meta );
 out:
    // place the disk storage object back onto the discovery dlist
    InsertObject ( (dlist_t)          output_objects,
                   (void *)           obj,
                   (TAG)              ObjectTag,
                   (void *)           NULL,
                   (Insertion_Modes)  AppendToList,
                   (boolean)          TRUE,
                   (void **)         &handle );
 out_all:
    return 0;
}

/*------------------------------------------------------------------+
+                                                                   +
+                   Start Of EVMS Plugin Functions                  +
+                (exported to engine via function table)            +
+                                                                   +
+-------------------------------------------------------------------*/

/**************************** Setup and Cleanup *********************/

static int SPARSE_SetupEVMSPlugin( engine_functions_t * engine_functions )
{
    EngFncs = engine_functions;
    LOG_ENTRY();
    LOG_EXIT_INT(0);
    return 0;
}

static void SPARSE_Cleanup(void)
{
    int rc = 0;
    dlist_t tmp_list = CreateList();
    storage_object_t * object;

    LOG_ENTRY();

    if ( tmp_list != NULL ) {
        rc = EngFncs->get_object_list( SEGMENT,
                                              0,
                                              my_plugin_record,
                                              NULL,
                                              0,
                                              &tmp_list );
        rc = GoToStartOfList( tmp_list );
        if ( rc == DLIST_SUCCESS ) {

            rc = GetObject( tmp_list, SEGMENT_TAG,
                            NULL, TRUE, (void**) &object );

            while ( rc == DLIST_SUCCESS ) {
                if ( object->data_type == META_DATA_TYPE )
                    delete_sparse_hash_map
                        ( (sparse_metavolume_t*)object->private_data, TRUE );

                if ( object->private_data )
                    EngFncs->engine_free(object->private_data );

                rc = GetNextObject( tmp_list, SEGMENT_TAG, (void**) &object );
            }
        }
        DestroyList(&tmp_list, FALSE);
    }

    LOG_EXIT_VOID();
}

/****************** The CAN'T Functions **********************/

/* This plugin supports Assign/Unassign instead */
static int SPARSE_CanDestroy( storage_object_t * seg )
{

    LOG_ENTRY();
    LOG_EXIT_INT(ENOSYS);
    return ENOSYS;
}

/* Move not implemented yet */
static int SPARSE_CanReplaceChild( storage_object_t * object,
                                   storage_object_t * child,
                                   storage_object_t * new_child )
{

    LOG_ENTRY();
    LOG_EXIT_INT(ENOSYS);
    return ENOSYS;
}

/****************** The CAN Functions **************************/

static int SPARSE_CanSetVolume( storage_object_t * seg, boolean flag )
{
    int rc = EINVAL;

    LOG_ENTRY();

    if ( i_can_modify ( seg ) == TRUE ) {

        rc = 0;

    }

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_CanUnassign( storage_object_t *ld )
{
    int rc = EINVAL;
    storage_object_t *seg=NULL;

    LOG_ENTRY();

    rc = GetObject( ld->parent_objects,
                    SEGMENT_TAG, NULL, FALSE, (void **) &seg );

    LOG_DEBUG( "seg: %s\n", seg->name );

    if ( ( seg ) &&
         ( seg->object_type == SEGMENT ) &&
         ( i_can_modify( seg ) == TRUE ) ) {

        rc = 0;

    }
    LOG_EXIT_INT(rc);
    return rc;
}

/* basically allow anything to expand */
static int SPARSE_CanExpand( storage_object_t *seg,
                             sector_count_t   *expand_limit,
                             dlist_t           expansion_points )
{
    int rc = EINVAL;
    expand_object_info_t *expand_object;
    sparse_volume_t * pdata = NULL;
    sparse_metavolume_t * vol = NULL;
    void *handle;

    LOG_ENTRY();

    if ( ( expansion_points ) &&
         ( seg ) &&
         ( seg->object_type == SEGMENT ) &&
         ( seg->data_type == DATA_TYPE ) &&
         ( i_can_modify( seg ) == TRUE ) &&
         /* ( seg->volume == NULL ) && */
         ( seg->consuming_container == NULL ) ) {

        expand_object = (expand_object_info_t *)
            EngFncs->engine_alloc( sizeof( expand_object_info_t ) );

        if ( ! expand_object ) {
            LOG_ERROR("error, alloc of expand object failed\n");
            LOG_EXIT_INT(ENOMEM);
            return ENOMEM;
        }
        pdata = (sparse_volume_t*)seg->private_data;
        vol = (sparse_metavolume_t*)pdata->object->private_data;
        expand_object->object = seg;
        expand_object->max_expand_size = SPARSE_MAX_VIRTUAL_SIZE -
            ( vol->meta_data.virtual_size );

        rc = InsertObject( (dlist_t) expansion_points,
                           (void * ) expand_object,
                           (TAG) EXPAND_OBJECT_TAG,
                           (void *) NULL,
                           (Insertion_Modes) AppendToList,
                           (boolean) TRUE,
                           (void **) &handle);
        if ( rc ) {
            EngFncs->engine_free( expand_object );
            LOG_EXIT_INT(EPERM);
            return EPERM;
        }
    }
    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_CanExpandBy(storage_object_t * seg, sector_count_t *size)
{
    int rc = EINVAL;
    sparse_metavolume_t *vol = NULL;
    sparse_volume_t *pdata = NULL;

    LOG_ENTRY();

    if ( ( seg ) &&
         ( seg->object_type == SEGMENT ) &&
         ( seg->data_type == DATA_TYPE ) &&
         ( i_can_modify( seg ) == TRUE ) &&
         /* ( seg->volume == NULL ) && */
         ( seg->consuming_container == NULL ) ) {

        pdata = (sparse_volume_t*)seg->private_data;
        vol = (sparse_metavolume_t *)pdata->object->private_data;
        if ( *size >
             ( SPARSE_MAX_VIRTUAL_SIZE - vol->meta_data.virtual_size ) ) {
            *size = SPARSE_MAX_VIRTUAL_SIZE - vol->meta_data.virtual_size;
            rc = EINVAL;
        }
    }
    LOG_EXIT_INT(rc);
    return rc;
}

/* only shrink un-claimed segments b/c must reset all of COW table */
static int SPARSE_CanShrink( storage_object_t * seg,
                             sector_count_t   * shrink_limit,
                             dlist_t            shrink_points )
{
    uint count ;
    int rc = EINVAL;
    shrink_object_info_t *shrink_object;
    sparse_metavolume_t * vol = NULL;
    sparse_volume_t * pdata = NULL;
    void *handle;

    LOG_ENTRY();

    if ( ( shrink_points ) &&
         ( seg ) &&
         ( seg->object_type == SEGMENT ) &&
         ( seg->data_type == DATA_TYPE ) &&
         ( i_can_modify( seg ) == TRUE ) &&
         ( seg->volume == NULL ) &&
         ( seg->consuming_container == NULL ) ) {

        rc = GetListSize(seg->parent_objects, &count);
        if ( rc ) {
            LOG_EXIT_INT(rc);
            return rc;
        }
        if ( count ) {
            LOG_EXIT_INT(EINVAL);
            return EINVAL;
        }

        shrink_object = (shrink_object_info_t *)
            EngFncs->engine_alloc( sizeof( shrink_object_info_t ) );

        if ( ! shrink_object ) {
            LOG_ERROR("error, alloc of shrink object failed\n");
            LOG_EXIT_INT(ENOMEM);
            return ENOMEM;
        }
        pdata = (sparse_volume_t*)seg->private_data;
        vol = (sparse_metavolume_t*)pdata->object->private_data;
        shrink_object->object = seg;
        shrink_object->max_shrink_size = vol->meta_data.virtual_size -
            SPARSE_MIN_VIRTUAL_SIZE;

        rc = InsertObject( (dlist_t) shrink_points,
                           (void * ) shrink_object,
                           (TAG) SHRINK_OBJECT_TAG,
                           (void *) NULL,
                           (Insertion_Modes) AppendToList,
                           (boolean) TRUE,
                           (void **) &handle);
        if ( rc ) {
            EngFncs->engine_free( shrink_object );
            LOG_EXIT_INT(EPERM);
            return EPERM;
        }
    }
    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_CanShrinkBy( storage_object_t * seg,
                               sector_count_t   * size )
{
    int rc = EINVAL;
    uint count;
    sparse_metavolume_t *vol = NULL;
    sparse_volume_t *pdata = NULL;

    LOG_ENTRY();

    if ( ( i_can_modify( seg ) == TRUE ) &&
         ( seg->volume == NULL ) &&
         ( seg->consuming_container == NULL ) ) {

        rc = GetListSize( seg->parent_objects, &count );
        if ( rc ) {
            LOG_EXIT_INT(rc);
            return rc;
        }

        if ( count ) {
            LOG_EXIT_INT(EINVAL);
            return EINVAL;
        }

        pdata = (sparse_volume_t*)seg->private_data;
        vol = (sparse_metavolume_t *)pdata->object->private_data;
        if ( *size >
             ( vol->meta_data.virtual_size - SPARSE_MIN_VIRTUAL_SIZE ) ) {
            *size = vol->meta_data.virtual_size - SPARSE_MIN_VIRTUAL_SIZE;
            rc = EINVAL;
        }
    }
    LOG_EXIT_INT(rc);
    return rc;
}

/***************** Unsupported Functions *************************/

/* This segment manager does not support create */
static int SPARSE_CreateSegment( dlist_t          input_objects,
                                 option_array_t * options,
                                 dlist_t          new_objects)
{
    LOG_ENTRY();
    LOG_EXIT_INT(ENOSYS);
    return ENOSYS;
}

/* This segment mgr. doesn't support destroy */
static int SPARSE_DestroySegment( storage_object_t * seg,
                                  dlist_t child_objects )
{
    LOG_ENTRY();
    LOG_EXIT_INT(ENOSYS);
    return ENOSYS;
}

/* Move not yet implemented */
static int SPARSE_ReplaceChild( storage_object_t * object,
                                storage_object_t * child,
                                storage_object_t * new_child )
{
    LOG_ENTRY();
    LOG_EXIT_INT(ENOSYS);
    return ENOSYS;
}

/* Not supported */
static int SPARSE_DirectPluginCommunication(void * thing,
                                            boolean target_kernel_plugin,
                                            void * arg)
{
    LOG_ENTRY();
    LOG_EXIT_INT(ENOSYS);
    return ENOSYS;
}

/*********** Supported ( but equal No-Op )Functions *********************/

/*
 * This call notifies you that your object is being made into (or part of)
 * a volume or that your object is no longer part of a volume.  The "flag"
 * parameter indicates whether the volume is being created (TRUE) or
 * removed (FALSE).
 */
static void SPARSE_SetVolume(storage_object_t * object, boolean flag)
{
    /* Don't Care */
    return;
}


/****************** Supported Functions **********************/

/*
 *  Called to run discovery code on dlist of evms objects that
 *  the engine has found so far.  We essentially need to walk the
 *  list of objects, looking for Logical Disks, and see if we
 *  recognize partitioning schemes.  If so, consume the logical
 *  disk by removing it from the dlist, and place all new segment
 *  objects on the output_object dlist. Any object we dont like in
 *  the input_object dlist must be copied to the ouput_object dlist.
 *
 */
static int SPARSE_Discover( dlist_t input_objects,
                            dlist_t output_objects,
                            boolean final_call )
{
    int rc = 0;
    int count = 0;
    discovery_parm_block_t dparms = {output_objects, &count};

    LOG_ENTRY();
    ForEachItem( input_objects, sparse_segment_discovery,
                 (void *) &dparms, TRUE );

    rc = count;

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_Assign( storage_object_t * input_object,
                          option_array_t * options)
{
    int rc=EINVAL;
    u_int32_t chunk_size = SPARSE_DEFAULT_CHUNK_SIZE;
    u_int64_t virtual_size = SPARSE_DEFAULT_VIRTUAL_SIZE;
    u_int32_t chunks, cow_size;
    sparse_metadata_t * md;
    sector_count_t offset, start = 0;
    storage_object_t * child;
    uint child_count;
    storage_object_t *seg = NULL;
    storage_object_t *meta = NULL;
    guid_t id;

    LOG_ENTRY();

    rc = get_assign_options ( options, &chunk_size, &virtual_size );
    if (rc) {
        LOG_ERROR("Getting assign options failed\n");
        LOG_EXIT_INT(rc);
        return rc;
    }

    // # of chunks in the input_object - metadata
    chunks = ( input_object->size - 1 ) / chunk_size;

    //cow size for those chunks
    cow_size = ( ( chunks * 8 ) + 511 ) / EVMS_VSECTOR_SIZE;

    //subtract chunks which make up cow table
    chunks -= cow_size / chunk_size ;

    //add chunks which are gained from not needing cow entries
    //for the cow table itself
    chunks += cow_size / chunk_size / 64;

    //refigure cow size based on these fixups
    cow_size = ( ( chunks * 8 ) + 511 ) / EVMS_VSECTOR_SIZE;

    //find out what disk sector we start on for alignment purposes
    //we want to chunksize align
    if ( input_object->object_type != DISK ) {
        child = input_object;
        rc = GetListSize( child->child_objects, &child_count );
        if ( rc ) {
            LOG_EXIT_INT(rc);
            return rc;
        }
        while ( child_count > 1 ) {

            start += child->start;

            rc = GoToStartOfList( child->child_objects );
            if ( rc ) {
                LOG_EXIT_INT(rc);
                return rc;
            }
            rc = GetObject( child->child_objects,
                                SEGMENT_TAG, NULL, FALSE, (void **)&child );
            if ( rc ) {
                LOG_EXIT_INT(rc);
                return rc;
            }

            rc = GetListSize( child->child_objects, &child_count );
            if ( rc ) {
                LOG_EXIT_INT(rc);
                return rc;
            }

        }
        start += child->start;
    }

    //first guess at lba of first chunk
    start += ( 1 + cow_size );

    //dead sectors for chunk_size alignment
    offset = chunk_size - ( start % chunk_size );

    //actual data chunks
    chunks = ( input_object->size - ( 1 + cow_size + offset ) ) / chunk_size;

    LOG_DEBUG( "input_size=%Ld, chunk_size=%d, chunks=%d, cow_size=%d, "
               "offset=%Ld\n",
               input_object->size, chunk_size, chunks, cow_size, offset );

    /*

    //DEBUGGING None of these should ever occur

    //sanity check cow table will fit
    if ( ( ( chunks + 63 ) / 64 ) > ( cow_size + offset ) )  {
        LOG_ERROR( "Error occured in calculation cow table will not fit in "
        "allotted space\n" );
        LOG_EXIT_INT(rc);
        return rc;
        }

    //sanity check everything will fit
    if ( ( 1 + cow_size + offset + chunks * chunk_size ) >
         input_object->size )
        {
            LOG_ERROR( "Error occured in calculation everything will not "
            "fit in allotted space\n" );
            LOG_EXIT_INT(rc);
            return rc;
    }

    //sanity check chunksize aligned
    if ( ( ( start + offset ) % chunk_size ) != 0 ) {
        LOG_DEBUG( "DEBUG WARNING: segment is not aligned\n" );
    }

    //sanity check cowtable right size
    if ( ( cow_size - ( chunks / 64 ) ) > 1 ) {
        LOG_DEBUG( "DEBUG WARNING: cow table is too big\n" );
    }

    */

    md = (sparse_metadata_t *)EngFncs->engine_alloc(EVMS_VSECTOR_SIZE);

    md->signature = EVMS_SPARSE_SIGNATURE;
    md->version.major = MAJOR_VERSION;
    md->version.minor = MINOR_VERSION;
    md->version.patchlevel = PATCH_LEVEL;
    md->flags = 0;
    md->virtual_size = ( virtual_size / chunk_size ) * chunk_size;
    md->child_object_size = input_object->size;
    md->lba_of_COW_table = 1;
    md->lba_of_first_chunk = 1 + cow_size + offset;
    md->chunk_size = chunk_size;
    md->total_chunks = chunks;

    uuid_generate_random( (char*) &id );
    sprintf( md->uuid, "%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X",
                     id.time_low,
                     id.time_mid,
                     id.time_high,
                     id.clock_seq_high,
                     id.clock_seq_low,
                     id.node[0],
                     id.node[1],
                     id.node[2],
                     id.node[3],
                     id.node[4],
                     id.node[5] );
    rc = EngFncs->register_name( md->uuid );

    DeleteAllItems(input_object->parent_objects, FALSE);

    rc = allocate_sparse_segment( &seg, &meta, &input_object, md );

    ((sparse_metavolume_t*)meta->private_data)->flags |=
        SPARSE_INITIALIZE_COW;

    if ( !rc ) {
        seg->flags |= SOFLAG_DIRTY | SOFLAG_NEEDS_ACTIVATE;
        meta->flags |= SOFLAG_DIRTY;
    }

    EngFncs->engine_free( md );

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_UnAssign( storage_object_t *ld )
{
    sparse_metavolume_t *vol = NULL;
    sparse_volume_t *pdata = NULL;
    int rc = EINVAL;
    storage_object_t *seg = NULL;

    LOG_ENTRY();

    rc = ExtractObject( ld->parent_objects,
                        SEGMENT_TAG,
                        NULL,
                        (void **) &seg );

    LOG_DEBUG( "seg: %s\n", seg->name );

    if ( ( seg ) &&
         ( seg->object_type == SEGMENT ) &&
         ( i_can_modify( seg ) == TRUE ) ) {

        pdata = (sparse_volume_t*)seg->private_data;
        vol = (sparse_metavolume_t*)pdata->object->private_data;

        // blast first cow sector for safety
        KILL_SECTORS(ld, vol->meta_data.lba_of_COW_table, 1);

        // blast metadata
        KILL_SECTORS(ld, 0, 1);

        delete_sparse_hash_map( vol, TRUE );

        // free meta data
        EngFncs->engine_free( vol );
        EngFncs->free_segment( pdata->object );

        // free segment
        EngFncs->engine_free( seg->private_data );
        EngFncs->free_segment( seg );

        rc = 0;
    }

    DeleteAllItems(ld->parent_objects, FALSE);

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_Expand( storage_object_t *seg,
                          storage_object_t *expand_object,
                          dlist_t           objects,
                          option_array_t   *options )
{
    int rc = EINVAL;
    sector_count_t expand_sectors = 0;
    sparse_metavolume_t * vol = NULL;
    sparse_volume_t * pdata = NULL;

    LOG_ENTRY();

    if ( ( seg ) &&
         ( seg == expand_object ) &&
         ( seg->object_type == SEGMENT ) &&
         ( seg->data_type == DATA_TYPE ) &&
         ( i_can_modify( seg ) == TRUE ) &&
         /* ( seg->volume == NULL ) && */
         ( seg->consuming_container == NULL ) ) {

        pdata = (sparse_volume_t*)seg->private_data;
        vol = (sparse_metavolume_t *)pdata->object->private_data;
        get_expand_options( options, &expand_sectors );

        if ( expand_sectors >
             ( SPARSE_MAX_VIRTUAL_SIZE - vol->meta_data.virtual_size ) ) {
            expand_sectors = SPARSE_MAX_VIRTUAL_SIZE -
                vol->meta_data.virtual_size;
        }

        vol->meta_data.virtual_size += expand_sectors;
        vol->meta_data.virtual_size = ( vol->meta_data.virtual_size /
                                        vol->chunk_size ) * vol->chunk_size;
        seg->size = vol->meta_data.virtual_size;
        seg->flags |= SOFLAG_DIRTY; //customary
        seg->flags |= SOFLAG_NEEDS_ACTIVATE;
        pdata->object->flags |= SOFLAG_DIRTY; //where real work will get done
        rc = 0;
    }

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_Shrink( storage_object_t * seg,
                          storage_object_t * shrink_object,
                          dlist_t            objects,
                          option_array_t   * options )
{
    int rc = EINVAL;
    uint count;
    sector_count_t shrink_sectors = 0;
    sparse_metavolume_t * vol = NULL;
    sparse_volume_t * pdata = NULL;
    int answer = 0;
    char * choice_text[3] = {"Don't Shrink", "Shrink", NULL};

    LOG_ENTRY();

    if ( ( seg ) &&
         ( seg->object_type = SEGMENT ) &&
         ( seg->data_type = DATA_TYPE ) &&
         ( i_can_modify( seg ) == TRUE ) &&
         ( seg->volume == NULL ) &&
         ( seg->consuming_container == NULL ) ) {

        rc = GetListSize( seg->parent_objects, &count );
        if ( rc ) {
            LOG_EXIT_INT(rc);
            return rc;
        }

        if ( count ) {
            LOG_EXIT_INT(EINVAL);
            return EINVAL;
        }

        rc = E_CANCELED;
        QUESTION( &answer, choice_text,
                  "***WARNING*** Selecting Shrink will cause the Sparse to "
                  "Clear its Copy-On-Write Tables.  All data on %s will be "
                  "lost. \n\n", seg->name );

        if ( answer ) {
            pdata = (sparse_volume_t*)seg->private_data;
            vol = (sparse_metavolume_t *)pdata->object->private_data;
            get_shrink_options( options, &shrink_sectors );

            if ( shrink_sectors >
                 ( vol->meta_data.virtual_size - SPARSE_MIN_VIRTUAL_SIZE ) ) {
                shrink_sectors = vol->meta_data.virtual_size -
                    SPARSE_MIN_VIRTUAL_SIZE;
            }

            vol->next_cow_entry = 0;
            vol->current_cow_sector = vol->meta_data.lba_of_COW_table;
            vol->next_free_chunk = 0;
            memset( vol->cow_table, 0xff, EVMS_VSECTOR_SIZE);
            vol->next_free_chunk = 0;

            delete_sparse_hash_map( vol, FALSE );

            vol->meta_data.virtual_size -= shrink_sectors;
            vol->meta_data.virtual_size =
                ( vol->meta_data.virtual_size / vol->chunk_size ) *
                vol->chunk_size;
            seg->size = vol->meta_data.virtual_size;

            vol->flags |= SPARSE_INITIALIZE_COW;
            vol->flags &= ~EVMS_SPARSE_FULL;
            vol->meta_data.flags &= ~EVMS_SPARSE_FULL;

            pdata->object->flags |= SOFLAG_DIRTY; //real work
            seg->flags |= SOFLAG_DIRTY; //customary
            seg->flags |= SOFLAG_NEEDS_ACTIVATE;

            rc = 0;
        }
    }
    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_AddSectorsToKillList( storage_object_t *seg,
                                        lsn_t sector,
                                        sector_count_t count)
{
    int rc = 0;
    sparse_hash_entry_t * result;
    sparse_volume_t * pdata = NULL;
    sparse_metavolume_t *vol = NULL;
    u_int64_t chunk, chunk_stop;
    unsigned long remainder, hash_value;
    lsn_t remapped_sector;
    sector_count_t actual_count;

    LOG_ENTRY();
    LOG_DEBUG("original request: @sector %Ld, for %Ld sectors\n",
              sector, count);

    if ( i_can_modify( seg ) != TRUE ) {
        LOG_ERROR("Not allowed to modify segment\n");
        LOG_EXIT_INT(EINVAL);
        return EINVAL;
    }

    if ( seg->data_type != DATA_TYPE ) {
        LOG_ERROR("Not allowed to kill my metadata\n");
        LOG_EXIT_INT(EINVAL);
        return EINVAL;
    }

    pdata = (sparse_volume_t*)seg->private_data;
    vol = (sparse_metavolume_t*)pdata->object->private_data;

    chunk = sector >> vol->chunk_shift;
    chunk_stop = ( sector + count - 1) >> vol->chunk_shift;

    for ( ; chunk <= chunk_stop && count > 0; chunk++ ) {

        remainder = sector & (u_int64_t)(vol->chunk_size-1);

        actual_count = min( vol->chunk_size - remainder, count );

        hash_value = ((unsigned long) chunk ) % vol->hash_table_size;

        if ( search_sparse_hash_chain( chunk,
                                       vol->sparse_map[hash_value],
                                       &result ) ) {
            //Engine found a remap queue the kill
            remapped_sector =
                ( result->sparse_chunk << vol->chunk_shift ) + remainder;
            KILL_SECTORS( vol->child_object,
                          seg->start + remapped_sector,
                          actual_count );
        }
        // else ignoring b/c not remapped

        //Update counts -- in all cases
        rc = 0;
        sector += actual_count;
        count -= actual_count;
    }

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_CommitChanges( storage_object_t *seg, uint phase )
{
    int rc = 0;
    lba_t sector;
    int size;
    char * buffer;
    sparse_metadata_t * md;
    sparse_metavolume_t * vol = NULL;
    storage_object_t * ld = NULL;
    int use_progress;
    progress_t progress;

    LOG_ENTRY();

    switch(phase) {

    case SETUP:
        break;

    case FIRST_METADATA_WRITE:
        if ( seg->data_type == DATA_TYPE ) { //meat of commit on metadata
            seg->flags &= ~SOFLAG_DIRTY;
            LOG_EXIT_INT(0);
            return 0;
        }

        ld = get_logical_disk( seg );
        if ( !ld ) {
            LOG_EXIT_INT(EINVAL);
            return EINVAL;
        }

        vol = (sparse_metavolume_t *)seg->private_data;

        if ( vol->flags & SPARSE_INITIALIZE_COW &&
             seg->flags & SOFLAG_ACTIVE ) {
            LOG_ERROR("Cannot clear cow table on active kernel object\n");
            LOG_EXIT_INT(EINVAL);
            return EINVAL;
        }

        if ( vol->flags & SPARSE_INITIALIZE_COW ) {
            //alloc 64k buffer
            buffer = EngFncs->engine_alloc( 64 * 1024 );
        }
        else {
            //alloc 1 sector buffer
            buffer = EngFncs->engine_alloc( 512 );
        }

        if ( !buffer ) {
            LOG_EXIT_INT(ENOMEM);
            return ENOMEM;
        }

        md = (sparse_metadata_t *)buffer;
        memcpy( md, &(vol->meta_data), sizeof(sparse_metadata_t) );
        convert_metadata( md );
        md->CRC = 0;
        md->CRC = DISK_TO_CPU32( EngFncs->
                                 calculate_CRC( EVMS_INITIAL_CRC,
                                                md,
                                                sizeof(sparse_metadata_t) ) );
        rc = WRITE( ld, 0, 1, buffer );

        //only on Assign and Shrink and Reset
        if ( vol->flags & SPARSE_INITIALIZE_COW ) {
            progress.id = 0;
            progress.title = "Initializing CopyOnWrite table for Sparse...";
            progress.description = "";
            progress.type = DISPLAY_PERCENT;
            progress.count = 0;
            progress.total_count = vol->meta_data.lba_of_first_chunk - 1;
            progress.remaining_seconds = 0;
            progress.plugin_private_data = NULL;
            progress.ui_private_data = NULL;

            use_progress = EngFncs->progress( &progress ) == 0;

            if (!use_progress) {
                MESSAGE("Initializing CopyOnWrite table for Sparse %s. "
                        "This will take a moment or two.\n",
                        seg->name );
            }

            //Initialize the entire range of COW tables
            memset( buffer, 0xff, 64*1024 );

            for ( sector = vol->meta_data.lba_of_COW_table;
                  sector < vol->meta_data.lba_of_first_chunk;
                  sector += 128 ) {
                size = min( 128, vol->meta_data.lba_of_first_chunk - sector );
                if ( use_progress ) {
                    progress.count += size;
                    EngFncs->progress( &progress );
                }
                rc = WRITE( ld, sector, size, buffer );
            }
            vol->flags &= ~SPARSE_INITIALIZE_COW;
            vol->object->flags &= ~SOFLAG_HAS_STOP_DATA;
        }

        if ( buffer ) {
            EngFncs->engine_free( buffer );
        }

        break;

    case SECOND_METADATA_WRITE:

        break;

    case POST_ACTIVATE:

        if ( seg->data_type == META_DATA_TYPE ) {
            rc = build_sparse_maps( (sparse_metavolume_t*)seg->private_data,
                                    TRUE );
            seg->flags &= ~SOFLAG_DIRTY;
        }

        break;
    }
    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_Read( storage_object_t *seg,
                        lsn_t             sector,
                        sector_count_t    count,
                        void             *buffer )
{
    char * io_buffer_ptr = (char *)buffer;
    sparse_hash_entry_t * result;
    sparse_volume_t * pdata = NULL;
    sparse_metavolume_t *vol = NULL;
    u_int64_t chunk, chunk_stop;
    unsigned long remainder, hash_value;
    lsn_t remapped_sector;
    sector_count_t actual_count;
    int fd, rc = 0;

    LOG_ENTRY();
    LOG_DEBUG("original request: %s @sector %Ld, for %Ld sector(s)\n",
              seg->name, sector, count);

    if ( ! buffer ) {
        LOG_EXIT_INT(EFAULT);
        return EFAULT;
    }

    if ( seg->data_type != DATA_TYPE ) {
        LOG_ERROR("You can't read my metadata\n");
        LOG_EXIT_INT(EINVAL);
        return EINVAL;
    }

    if ( seg->flags & SOFLAG_ACTIVE ) {
        LOG_DEBUG("Object active in the kernel reading through the node\n");
        fd = EngFncs->open_object( seg, O_SYNC | O_RDONLY );
        rc = EngFncs->read_object( seg, fd, buffer,
                                         count<<EVMS_VSECTOR_SIZE_SHIFT,
                                         sector<<EVMS_VSECTOR_SIZE_SHIFT );
        EngFncs->close_object( seg, fd );
        if ( rc < 0 )
            return EIO;

        return 0;
    }


    pdata = (sparse_volume_t*)seg->private_data;
    vol = (sparse_metavolume_t*)pdata->object->private_data;

    chunk = sector >> vol->chunk_shift;
    chunk_stop = ( sector + count - 1 ) >> vol->chunk_shift;

    for ( ; chunk<=chunk_stop && count > 0; chunk++ ) {

        remainder = sector & (u_int64_t)(vol->chunk_size-1);
        hash_value = ((unsigned long) chunk ) % vol->hash_table_size;

        actual_count = min( vol->chunk_size - remainder, count );

        if ( search_sparse_hash_chain( chunk, vol->sparse_map[hash_value],
                                       &result ) ) {

            remapped_sector =
                ( result->sparse_chunk << vol->chunk_shift ) + remainder;
            LOG_DEBUG("Reading at original chunk: %Ld, which is sparse "
                      "chunk: %Ld, at sparse sector: %Ld for %Ld sectors\n",
                      chunk, result->sparse_chunk,
                      seg->start + remapped_sector, actual_count );

            READ( vol->child_object, seg->start + remapped_sector,
                  actual_count, (void *)io_buffer_ptr );
        }
        else {
            LOG_DEBUG("Reading at original chunk: %Ld, which is unmapped "
                      "to a sparse chunk for %Ld sectors yields all zeros\n",
                      chunk, actual_count );

            memset ( (void *)io_buffer_ptr, 0x0,
                     actual_count << EVMS_VSECTOR_SIZE_SHIFT );
        }

        sector += actual_count;
        count -= actual_count;
        io_buffer_ptr += ( actual_count << EVMS_VSECTOR_SIZE_SHIFT );
    }

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_Write( storage_object_t *seg,
                         lsn_t             sector,
                         sector_count_t    count,
                         void             *buffer )
{
    char * io_buffer_ptr = (char *)buffer;
    sparse_hash_entry_t * result;
    sparse_volume_t* pdata = NULL;
    sparse_metavolume_t *vol = NULL;
    u_int64_t chunk, chunk_stop, new_sparse_chunk;
    unsigned long remainder, hash_value;
    lsn_t remapped_sector;
    sector_count_t actual_count;
    int fd, rc = EINVAL;

    LOG_ENTRY();
    LOG_DEBUG("original request: %s @sector %Ld, for %Ld sectors\n",
              seg->name, sector, count);

    if ( ! buffer ) {
        LOG_EXIT_INT(EFAULT);
        return EFAULT;
    }

    if ( seg->data_type != DATA_TYPE ) {
        LOG_ERROR("You can't read my metadata\n");
        LOG_EXIT_INT(EINVAL);
        return EINVAL;
    }

    if ( seg->flags & SOFLAG_ACTIVE ) {
        LOG_DEBUG("Object active in the kernel writing through the node\n");
        fd = EngFncs->open_object( seg, O_SYNC | O_RDWR );
        rc = EngFncs->write_object( seg, fd, buffer,
                                          count<<EVMS_VSECTOR_SIZE_SHIFT,
                                          sector<<EVMS_VSECTOR_SIZE_SHIFT );
        EngFncs->close_object( seg, fd );
        if ( rc < 0 )
            return EIO;

        ((sparse_volume_t*)seg->private_data)->object->flags |= SOFLAG_DIRTY;
        return 0;
    }

    pdata = (sparse_volume_t*)seg->private_data;
    vol = (sparse_metavolume_t*)pdata->object->private_data;

    chunk = sector >> vol->chunk_shift;
    chunk_stop = ( sector + count - 1 ) >> vol->chunk_shift;

    for ( ; chunk<=chunk_stop && count > 0; chunk++ ) {

        remainder = sector & (u_int64_t)(vol->chunk_size-1);
        hash_value = ((unsigned long) chunk ) % vol->hash_table_size;

        actual_count = min (vol->chunk_size - remainder, count );

        if ( search_sparse_hash_chain( chunk, vol->sparse_map[hash_value],
                                       &result ) ) {

            remapped_sector =
                ( result->sparse_chunk << vol->chunk_shift ) + remainder;

            LOG_DEBUG("Writing at original chunk: %Ld, which is sparse "
                      "chunk: %Ld, at sparse sector: %Ld for %Ld sectors\n",
                      chunk, result->sparse_chunk,
                      seg->start + remapped_sector, actual_count );

            WRITE( vol->child_object, seg->start + remapped_sector,
                   actual_count, (void *)io_buffer_ptr );

        }

        else {
            //add new chunk to sparse_map
            new_sparse_chunk = vol->next_free_chunk++;
            if ( new_sparse_chunk >= vol->num_chunks ) {
                LOG_ERROR("Sparse is fully remapped not able to map any "
                          "new chunks");
                vol->flags |= EVMS_SPARSE_FULL;
                vol->meta_data.flags |= EVMS_SPARSE_FULL;
                LOG_EXIT_INT(ENOSPC);
                return ENOSPC;
            }
            rc = add_cow_entry_to_sparse_map ( chunk, new_sparse_chunk, vol );

            remapped_sector =
                ( new_sparse_chunk << vol->chunk_shift ) + remainder;

            //write the data requested
            LOG_DEBUG("Writing at original chunk: %Ld, which is sparse "
                      "chunk: %Ld, at sparse sector: %Ld for %Ld sectors\n",
                      chunk, new_sparse_chunk,
                      seg->start + remapped_sector, actual_count);

            WRITE( vol->child_object, seg->start + remapped_sector,
                   actual_count, (void *)io_buffer_ptr );

            //update the cow_table
            vol->cow_table[vol->next_cow_entry++] = chunk;

            //write the current cow table
            WRITE( vol->child_object, vol->current_cow_sector, 1,
                   (void *)vol->cow_table );

            if (vol->next_cow_entry ==
                ( EVMS_VSECTOR_SIZE / sizeof( u_int64_t ) ) ) {
                //use the next sector of cow table
                vol->current_cow_sector++;
                memset ( vol->cow_table, 0xff, EVMS_VSECTOR_SIZE );
            }

        }
        sector += actual_count;
        count -= actual_count;
        io_buffer_ptr += ( actual_count << EVMS_VSECTOR_SIZE_SHIFT );


    }

    rc = 0;
    LOG_EXIT_INT(rc);
    return rc;
}

/*
 *  Function:  get_DM_info
 *
 *  Called when a DM activate fails.  The job here is to see if the 
 *  segment is active in the kernel.  If so ... compare the target 
 *  mapping against the engine segment object.  If the segments starting
 *  LBA and SIZE match ... then we dont need to activate the segment
 *  because it is already active and has a correct mapping ... return 0.
 */
static int get_DM_info( storage_object_t *seg )
{	
	dm_target_t *targets=NULL;
	dm_device_t *dev = NULL;
	int rc;

	LOG_ENTRY();
	LOG_DEBUG("seg= %s\n", seg->name);
		
        rc = EngFncs->dm_update_status(seg);
	if (!rc) {
		if (seg->flags & SOFLAG_ACTIVE) {  
			LOG_DEBUG("segment IS active in the kernel\n");
			rc = EngFncs->dm_get_targets(seg, &targets);
			if ( !rc && targets != NULL) {
				dev = (dm_device_t *) targets->data.linear;
				if ( seg->start == dev->start &&
				     seg->size  == targets->length) {
					LOG_DEBUG("kernel object matches ... marking segment active\n");
					rc = 0;
				}
				else {
					LOG_ERROR("error, got a DM object using our segment name but the metadata differs. dont know what to do!\n");
					rc = ENODEV; // dont know how to handle this.
				}
 			}
			else {
				rc = ENODEV;
			}
			if (targets) EngFncs->dm_deallocate_targets(targets);
		}
		else {
			LOG_DEBUG("segment is NOT active in the kernel\n");
			rc=ENODEV;
		}
	}

	if (rc) {
		seg->flags &= ~SOFLAG_ACTIVE;
		seg->flags |= SOFLAG_NEEDS_ACTIVATE;
	}
 		
	LOG_EXIT_INT(rc);
	return rc;
}

static int SPARSE_Activate( storage_object_t * seg )
{
    int rc = EINVAL;
    dm_target_t* trgt;
    dm_target_sparse_t *sprs;
    storage_object_t *ld = NULL;
    sparse_metavolume_t * vol;
    sparse_volume_t * pdata;

    LOG_ENTRY();

    ld = get_logical_disk( seg );
    if ( !ld ) {
        LOG_EXIT_INT(rc);
        return rc;
    }

    trgt = EngFncs->dm_allocate_target( DM_TARGET_SPARSE, 0,
                                               seg->size, 0 );
    if ( !trgt ) {
        LOG_EXIT_INT(rc);
        return rc;
    }

    sprs = trgt->data.sparse;
    if ( !sprs ) {
        LOG_EXIT_INT(rc);
        return rc;
    }

    pdata = (sparse_volume_t*)seg->private_data;
    vol = (sparse_metavolume_t*)pdata->object->private_data;

    sprs->device.major = ld->dev_major;
    sprs->device.minor = ld->dev_minor;
    sprs->device.start = seg->start;
    sprs->num_chunks = vol->num_chunks;
    sprs->chunk_size = vol->chunk_size;

    rc = EngFncs->dm_activate( seg, trgt );
    EngFncs->dm_deallocate_targets(trgt);

    if (rc) {
        rc = get_DM_info(seg);  // FAILED ... test if already active
    }

    if ( !rc )
        seg->flags &= ~SOFLAG_NEEDS_ACTIVATE;

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_DeActivate( storage_object_t * seg )
{
    int rc;
    LOG_ENTRY();
    rc = EngFncs->dm_deactivate( seg );
    LOG_EXIT_INT(rc);
    return rc;
}

/************************* Task Functions **************************/

static int SPARSE_GetOptionCount( task_context_t * task )
{
    int count = 0;
    LOG_ENTRY();
    switch( task->action ) {
    case EVMS_Task_Assign_Plugin:
        count = SPARSE_ASSIGN_OPTIONS;
        break;

    case EVMS_Task_Expand:
        count = SPARSE_EXPAND_OPTIONS;
        break;

    case EVMS_Task_Shrink:
        count = SPARSE_SHRINK_OPTIONS;
        break;

    default:
        count = -1;
    }
    LOG_EXIT_INT(count);
    return (count);
}

static int SPARSE_InitTask( task_context_t * context )
{
    int rc = EINVAL;

    LOG_ENTRY();

    if ( context ) {

        switch ( context->action ) {

        case EVMS_Task_Assign_Plugin:

            context->min_selected_objects = 1;
            context->max_selected_objects = 1;

            rc = allocate_assign_option_descriptors( context );
            if (rc == 0) {
                rc = get_acceptable_assign_objects( context );
            }
            break;

        case EVMS_Task_Expand:
            context->min_selected_objects = 0;
            context->max_selected_objects = 0;

            rc = allocate_expand_option_descriptors( context );

            break;

        case EVMS_Task_Shrink:
            context->min_selected_objects = 0;
            context->max_selected_objects = 0;

            rc = allocate_shrink_option_descriptors( context );

            break;

        case SPARSE_FUNCTION_RESET:

            context->min_selected_objects = 0;
            context->max_selected_objects = 0;
            context->option_descriptors->count = 0;
            rc = 0;

            break;

        default:
            LOG_ERROR("error, context->action is unknown or unsupported\n");
            break;
        }
    }

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_SetOption( task_context_t * context, u_int32_t index,
                             value_t * value, task_effect_t * effect )
{
    int rc = EINVAL;

    LOG_ENTRY();

    if (! context ) {
        LOG_EXIT_INT(rc);
        return rc;
    }

    switch ( context->action ) {
    case EVMS_Task_Assign_Plugin:
        rc = set_assign_option( context, index, value, effect );
        break;

    case EVMS_Task_Expand:
        rc = set_expand_option( context, index, value, effect );
        break;

    case EVMS_Task_Shrink:
        rc = set_shrink_option( context, index, value, effect );
        break;

    default:
        LOG_ERROR("error, context->action is unknown or unsupported\n");
        break;
    }
    LOG_EXIT_INT(rc);
    return rc;
}


static int SPARSE_SetObjects ( task_context_t * context,
                               dlist_t declined_objects,
                               task_effect_t * effect )
{
    int rc=EINVAL;

    LOG_ENTRY();
    if ( context ) {
        switch ( context->action ) {
        case EVMS_Task_Assign_Plugin:
            rc = set_assign_object( context, declined_objects, effect );
            break;

        case EVMS_Task_Expand:
        case EVMS_Task_Shrink:
        case SPARSE_FUNCTION_RESET:
            break;

        default:
            LOG_ERROR("context->action is unknown or unsupported\n");
            break;
        }
    }
    LOG_EXIT_INT(rc);
    return rc;
}

/********************** Plugin Functions ********************/

static int SPARSE_GetPluginFunctions( storage_object_t * object,
                                      function_info_array_t ** functions )
{
    int rc = 0;
    function_info_array_t * fia;
    int function_count = 0;
    uint count;

    LOG_ENTRY();

    *functions = NULL; //in case of early termination

    if ( object == NULL ) {
        LOG_EXIT_INT(ENOSYS);
        return ENOSYS;
    }
    if ( object->data_type != DATA_TYPE ) {
        LOG_EXIT_INT(ENOSYS); //not implemented for non-DATA_TYPE segments
        return ENOSYS;
    }

    fia = EngFncs->engine_alloc( sizeof( function_info_array_t ) +
                                        sizeof( function_info_t ) *
                                        ( SPARSE_FUNCTION_COUNT - 1 ) );
    if ( ! fia ) {
        LOG_CRITICAL("Error allocating memory for an action info array.\n");
        LOG_EXIT_INT(ENOMEM);
        return ENOMEM;
    }

    rc = GetListSize( object->parent_objects, &count );
    if ( rc ) {
        EngFncs->engine_free( fia );
        LOG_ERROR("could not get the number of parent objects\n");
        LOG_EXIT_INT(rc);
        return rc;
    }

    if ( count  || object->volume || object->consuming_container ) {
        goto out;
    }

    fia->info[function_count].function = SPARSE_FUNCTION_RESET;
    SET_STRING(fia->info[function_count].name, "reset");
    SET_STRING(fia->info[function_count].title, "Reset");
    SET_STRING(fia->info[function_count].verb, "Reset");
    SET_STRING(fia->info[function_count].help,
               "Re-initialize COW table all previous data written is lost");
    function_count++;

 out:
    fia->count = function_count;
    *functions = fia;
    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_PluginFunction( storage_object_t * object,
                                  task_action_t action,
                                  dlist_t objects,
                                  option_array_t * options )
{
    int rc = EINVAL;
    sparse_volume_t * pdata = NULL;
    sparse_metavolume_t * vol = NULL;
    uint count;
    int answer = 0;
    char * choice_text[3] = {"Don't Reset", "Reset", NULL};

    LOG_ENTRY();

    if ( ( action < EVMS_Task_Plugin_Function ) ||
         ( action >= EVMS_Task_Plugin_Function + SPARSE_FUNCTION_COUNT ) ) {
        LOG_ERROR("Action code 0x%x is out of range.\n", action );
        LOG_EXIT_INT(rc);
        return rc;
    }

    pdata = (sparse_volume_t *) object->private_data;
    vol = (sparse_metavolume_t *) pdata->object->private_data;

    if ( ! vol ) {
        LOG_ERROR("Object %s does not have sparse private data. \n",
                  object->name );
        LOG_EXIT_INT(rc);
        return rc;
    }

    switch ( action ) {
    case SPARSE_FUNCTION_RESET:
        rc = GetListSize( object->parent_objects, &count );
        if ( rc ) {
            LOG_ERROR("could not get the number of parent objects\n");
            break;
        }
        if ( count ) {
            LOG_ERROR("Cannot reset because sparse(%s) has parent objects\n",
                      object->name);
            rc = EINVAL;
            break;
        }

        QUESTION( &answer, choice_text,
                 "***WARNING*** Selecting Reset will cause the Sparse to "
                 "Clear its Copy-On-Write Tables.  All data on %s will be "
                 "lost. \n\n", object->name );

        if ( answer ) {
            vol->next_cow_entry = 0;
            vol->current_cow_sector = vol->meta_data.lba_of_COW_table;
            vol->next_free_chunk = 0;
            memset( vol->cow_table, 0xff, EVMS_VSECTOR_SIZE);
            vol->next_free_chunk = 0;

            delete_sparse_hash_map( vol, FALSE );

            vol->flags |= SPARSE_INITIALIZE_COW;
            vol->flags &= ~EVMS_SPARSE_FULL;
            vol->meta_data.flags &= ~EVMS_SPARSE_FULL;
            object->flags |= SOFLAG_DIRTY;
            object->flags |= SOFLAG_NEEDS_ACTIVATE;
            pdata->object->flags |= SOFLAG_DIRTY;
            rc = 0;
        }
        else
            rc = E_CANCELED;
        break;

    default:
        rc = EINVAL;
        break;
    }

    LOG_EXIT_INT(rc);
    return rc;
}

/********************** Info Functions **************************/

static int SPARSE_GetInfo( storage_object_t * object,
                           char * name,
                           extended_info_array_t ** info_array )
{
    extended_info_array_t * info = NULL;
    sparse_metavolume_t * volume = NULL;
    sparse_volume_t * pdata = NULL;
    int rc = 0;

    LOG_ENTRY();

    if ( object->data_type == DATA_TYPE ) {
        pdata = (sparse_volume_t*)object->private_data;
        volume = (sparse_metavolume_t*)pdata->object->private_data;
        pdata = (sparse_volume_t *)pdata->object->private_data;
    }
    else {
        volume = (sparse_metavolume_t*)object->private_data;
        pdata = (sparse_volume_t *)object->private_data;
    }

    info = EngFncs->engine_alloc
        ( sizeof( extended_info_array_t ) + sizeof( extended_info_t ) * 4 );
    if ( ! info ) {
        LOG_EXIT_INT(ENOMEM);
        return ENOMEM;
    }

    SET_STRING( info->info[0].name, "ActualSize" );
    SET_STRING( info->info[0].title, "Actual Size" );
    SET_STRING( info->info[0].desc,
                      "The size in Kilobytes of the underlying object." );
    info->info[0].type = EVMS_Type_Unsigned_Int64;
    info->info[0].unit = EVMS_Unit_Kilobytes;
    info->info[0].value.ui64 = volume->meta_data.child_object_size / 2;
    info->info[0].collection_type = EVMS_Collection_None;
    memset( &info->info[0].group, 0, sizeof( group_info_t ) );

    SET_STRING( info->info[1].name, "VirtualSize" );
    SET_STRING( info->info[1].title, "Virtual Size" );
    SET_STRING( info->info[1].desc,
                      "The exported size in Kilobytes of the sparse object." );
    info->info[1].type = EVMS_Type_Unsigned_Int64;
    info->info[1].unit = EVMS_Unit_Kilobytes;
    info->info[1].value.ui64 = volume->meta_data.virtual_size / 2;
    info->info[1].collection_type = EVMS_Collection_None;
    memset( &info->info[1].group, 0, sizeof( group_info_t ) );

    SET_STRING( info->info[2].name, "ChunkSize" );
    SET_STRING( info->info[2].title, "Chunk Size" );
    SET_STRING( info->info[2].desc,
                      "The size in Kilobytes of chunks mapped by the sparse.");
    info->info[2].type = EVMS_Type_Unsigned_Int32;
    info->info[2].unit = EVMS_Unit_Kilobytes;
    info->info[2].value.ui64 = volume->meta_data.chunk_size / 2;
    info->info[2].collection_type = EVMS_Collection_None;
    memset( &info->info[2].group, 0, sizeof( group_info_t ) );

    SET_STRING( info->info[3].name, "Usage" );
    SET_STRING( info->info[3].title, "Usage" );
    SET_STRING( info->info[3].desc,
                      "The percentage of the underlying object already full.");
    info->info[3].type = EVMS_Type_Unsigned_Int32;
    info->info[3].unit = EVMS_Unit_Percent;
    info->info[3].value.ui32 =
        (volume->next_free_chunk * 100) / volume->num_chunks;
    info->info[3].collection_type = EVMS_Collection_None;
    memset( &info->info[3].group, 0, sizeof( group_info_t ) );

    SET_STRING( info->info[4].name, "Status" );
    SET_STRING( info->info[4].title, "Sparse Status" );
    SET_STRING( info->info[4].desc,
                      "The current status of the Sparse." );
    info->info[4].type = EVMS_Type_String;
    info->info[4].unit = EVMS_Unit_None;

    if ( object->flags & SOFLAG_ACTIVE ) {
        if ( volume->flags & EVMS_SPARSE_FULL ) {
            SET_STRING(info->info[4].value.s, "Active, Fully Remapped");
        }
        else {
            SET_STRING(info->info[4].value.s, "Active");
        }
    }
    else if ( volume->flags & EVMS_SPARSE_FULL ) {
        SET_STRING( info->info[4].value.s, "Inactive, Full" );
    }
    else {
        SET_STRING( info->info[4].value.s, "Inactive" );
    }

    info->info[4].collection_type = EVMS_Collection_None;
    memset( &info->info[4].group, 0, sizeof( group_info_t ) );

    info->count = 5;
    *info_array = info;

    LOG_EXIT_INT(rc);
    return rc;
}

static int SPARSE_GetPluginInfo( char * name,
                                 extended_info_array_t * * info_array )
{
    extended_info_array_t   * info = NULL;
    char                    buffer[50] = {0};
    int                     i = 0;

    LOG_ENTRY();

    if ( ! name ) {
        // Get memory for the info array
        if ( ! (info =
                EngFncs->engine_alloc
                (sizeof(extended_info_array_t) + sizeof(extended_info_t)*5))) {
            LOG_ERROR("Error allocating memory for info array\n");
            LOG_EXIT_INT(ENOMEM);
            return ENOMEM;
        }

        // Short Name
        SET_STRING(info->info[i].name, "ShortName");
        SET_STRING(info->info[i].title, "Short Name");
        SET_STRING(info->info[i].desc, "A short name given to this plugin");
        info->info[i].type = EVMS_Type_String;
        SET_STRING(info->info[i].value.s, my_plugin_record->short_name);
        i++;

                // Long Name
        SET_STRING(info->info[i].name, "LongName");
        SET_STRING(info->info[i].title, "Long Name");
        SET_STRING(info->info[i].desc, "A long name given to this plugin");
        info->info[i].type = EVMS_Type_String;
        SET_STRING(info->info[i].value.s, my_plugin_record->long_name);
        i++;

        // Plugin Type
        SET_STRING(info->info[i].name, "Type");
        SET_STRING(info->info[i].title, "Plugin Type");
        SET_STRING(info->info[i].desc,
                   "There are various types of plugins; each responsible for"
                   "some kind of storage object.");
        info->info[i].type = EVMS_Type_String;
        SET_STRING(info->info[i].value.s, "Segment Manager");
        i++;

        // Plugin Version
        SET_STRING(info->info[i].name, "Version");
        SET_STRING(info->info[i].title, "Plugin Version");
        SET_STRING(info->info[i].desc,
                   "This is the version number of the plugin.");
        info->info[i].type = EVMS_Type_String;
        snprintf(buffer, 50, "%d.%d.%d", MAJOR_VERSION,
                 MINOR_VERSION, PATCH_LEVEL);
        SET_STRING(info->info[i].value.s, buffer);
        i++;

        // Required Engine Services Version
        SET_STRING(info->info[i].name, "Required_Engine_Version");
        SET_STRING(info->info[i].title, "Required Engine Services Version");
        SET_STRING(info->info[i].desc,
                   "This is the version of the Engine services that this "
                   "plug-in requires.  It will not run on older versions "
                   "of the Engine services.");
        info->info[i].type = EVMS_Type_String;
        snprintf(buffer, 50, "%d.%d.%d",
                 my_plugin_record->required_engine_api_version.major,
                 my_plugin_record->required_engine_api_version.minor,
                 my_plugin_record->required_engine_api_version.patchlevel);
        SET_STRING(info->info[i].value.s, buffer);
        i++;

        // Required Plug-in API Version
        SET_STRING(info->info[i].name, "Required_Plugin_Version");
        SET_STRING(info->info[i].title, "Required Plug-in API Version");
        SET_STRING(info->info[i].desc,
                   "This is the version of the Engine plug-in API that this "
                   "plug-in requires.  It will not run on older versions of "
                   "the Engine plug-in API.");
        info->info[i].type = EVMS_Type_String;
        snprintf(buffer, 50, "%d.%d.%d",
                 my_plugin_record->required_plugin_api_version.plugin.major,
                 my_plugin_record->required_plugin_api_version.plugin.minor,
                 my_plugin_record->
                 required_plugin_api_version.plugin.patchlevel);
        SET_STRING(info->info[i].value.s, buffer);
        i++;
    } else {
        LOG_ERROR("No support for extra plugin information about \"%s\"\n",
                  name);

        LOG_EXIT_INT(EINVAL);
        return EINVAL;
    }

    info->count = i;
    *info_array = info;

    LOG_EXIT_INT(0);
    return 0;
}

/*--------------------------------------------------------------------+
+                                                                     +
+                    PLUGIN FUNCTION TABLE                            +
+                                                                     +
+---------------------------------------------------------------------*/
static struct plugin_functions_s sft={

    // the following routines are found above
    setup_evms_plugin:                   SPARSE_SetupEVMSPlugin,
    cleanup_evms_plugin:                 SPARSE_Cleanup,
    can_set_volume:                      SPARSE_CanSetVolume,
    can_delete:                          SPARSE_CanDestroy,
    can_expand:                          SPARSE_CanExpand,
    can_expand_by:                       SPARSE_CanExpandBy,
    can_shrink:                          SPARSE_CanShrink,
    can_shrink_by:                       SPARSE_CanShrinkBy,
    can_replace_child:                   SPARSE_CanReplaceChild,
    can_unassign:                        SPARSE_CanUnassign,
    discover:                            SPARSE_Discover,
    create:                              SPARSE_CreateSegment,
    assign:                              SPARSE_Assign,
    unassign:                            SPARSE_UnAssign,
    delete:                              SPARSE_DestroySegment,
    expand:                              SPARSE_Expand,
    shrink:                              SPARSE_Shrink,
    replace_child:                       SPARSE_ReplaceChild,
    add_sectors_to_kill_list:            SPARSE_AddSectorsToKillList,
    commit_changes:                      SPARSE_CommitChanges,
    read:                                SPARSE_Read,
    write:                               SPARSE_Write,
    activate:                            SPARSE_Activate,
    deactivate:                          SPARSE_DeActivate,
    set_volume:                          SPARSE_SetVolume,
    get_option_count:                    SPARSE_GetOptionCount,
    init_task:                           SPARSE_InitTask,
    set_option:                          SPARSE_SetOption,
    set_objects:                         SPARSE_SetObjects,
    get_info:                            SPARSE_GetInfo,
    get_plugin_info:                     SPARSE_GetPluginInfo,
    get_plugin_functions:                SPARSE_GetPluginFunctions,
    plugin_function:                     SPARSE_PluginFunction,
    direct_plugin_communication:         SPARSE_DirectPluginCommunication
};



/*--------------------------------------------------------------------+
+                                                                     +
+               BUILD AND EXPORT AN EVMS PLUGIN RECORD                +
+                                                                     +
+---------------------------------------------------------------------*/

static plugin_record_t sparse_plugin_record = {

    id: SetPluginID( EVMS_OEM_IBM,
                     EVMS_SEGMENT_MANAGER,
                     EVMS_SPARSE_FEATURE_ID ),

    version: {MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL},

    required_engine_api_version: {12,0,0},
    required_plugin_api_version: {plugin: {11,0,0} },

    short_name: "SparseSegMgr",
    long_name: "Sparse Storage Segment Manager",
    oem_name: "IBM",

    functions: {plugin: &sft},

    container_functions: NULL

};

// Vector of plugin record ptrs that we export for the EVMS Engine.
plugin_record_t *my_plugin_record = &sparse_plugin_record;
plugin_record_t *evms_plugin_records[] = {
    &sparse_plugin_record,
    NULL
};
